#include "judge.h"

#define T_INT_32 (m->int32_ty_)
#define T_FLOAT_ (m->float32_ty_)
#define T_INT_1 (m->int1_ty_)

bool is_void(Instruction * t) 
{ return ((t->op_id_ == Instruction::Ret) || (t->op_id_ == Instruction::Br) || (t->op_id_ == Instruction::Store) || (t->op_id_ == Instruction::Call && t->type->tid_==Type::VoidTyID)); }

bool is_phi(Instruction * t)  

{ return t->op_id_ == Instruction::PHI; }

bool is_store(Instruction * t) 
{ return t->op_id_ == Instruction::Store; }

bool is_alloca(Instruction * t) 
{ return t->op_id_ == Instruction::Alloca; }

bool is_ret(Instruction * t) 
{ return t->op_id_ == Instruction::Ret; }
bool is_load(Instruction * t) { return t->op_id_ == Instruction::Load; }
bool is_br(Instruction * t) { return t->op_id_ == Instruction::Br; }

bool is_add(Instruction * t) { return t->op_id_ == Instruction::Add; }
bool is_sub(Instruction * t) { return t->op_id_ == Instruction::Sub; }
bool is_mul(Instruction * t) { return t->op_id_ == Instruction::Mul; }
bool is_div(Instruction * t) { return t->op_id_ == Instruction::SDiv; }
bool is_rem(Instruction * t) { return t->op_id_ == Instruction::SRem; }

bool is_fadd(Instruction * t) { return t->op_id_ == Instruction::FAdd; }
bool is_fsub(Instruction * t) { return t->op_id_ == Instruction::FSub; }
bool is_fmul(Instruction * t) { return t->op_id_ == Instruction::FMul; }
bool is_fdiv(Instruction * t) { return t->op_id_ == Instruction::FDiv; }

bool is_cmp(Instruction * t) { return t->op_id_ == Instruction::ICmp; }
bool is_fcmp(Instruction * t) { return t->op_id_ == Instruction::FCmp; }

bool is_call(Instruction * t) { return t->op_id_ == Instruction::Call; }
bool is_gep(Instruction * t) { return t->op_id_ == Instruction::GetElementPtr; }
bool is_zext(Instruction * t) { return t->op_id_ == Instruction::ZExt; }
bool is_fptosi(Instruction * t) { return t->op_id_ == Instruction::FPtoSI; }
bool is_sitofp(Instruction * t) { return t->op_id_ == Instruction::SItoFP; }

bool is_int_binary(Instruction * t)
{
    return (is_add(t) || is_sub(t) || is_mul(t) || is_div(t) || is_rem(t)) &&
            (t->num_ops_ == 2);
}

bool is_float_binary(Instruction * t)
{
    return (is_fadd(t) || is_fsub(t) || is_fmul(t) || is_fdiv(t)) &&
            (t->num_ops_ == 2);
}

bool is_binary(Instruction * t) { return is_int_binary(t) || is_float_binary(t);}

bool isTerminator(Instruction * t) { return is_br(t) || is_ret(t); }



ConstantInt * CalcInt(Instruction::OpID op, Module* m, ConstantInt *v1, ConstantInt *v2) {
  auto a = v1->value_, b = v2->value_;
  ConstantInt *  ans;
  switch (op) {
  case Instruction::Add:
    ans = new ConstantInt(T_INT_32, a + b);
  case Instruction::Sub:
    ans = new ConstantInt(T_INT_32, a - b);
  case Instruction::Mul:
    ans = new ConstantInt(T_INT_32, a * b);
  case Instruction::SDiv:
    ans = new ConstantInt(T_INT_32, a / b);
  case Instruction::SRem:
    ans = new ConstantInt(T_INT_32, a % b);
  case Instruction::Shl:
    ans = new ConstantInt(T_INT_32, a << b);
  case Instruction::LShr:
    ans = new ConstantInt(T_INT_32, (unsigned)a >> b);
  case Instruction::AShr:
    ans = new ConstantInt(T_INT_32, a >> b);
  case Instruction::And:
    ans = new ConstantInt(T_INT_32, a & b);
  case Instruction::Or:
    ans = new ConstantInt(T_INT_32, a | b);
  case Instruction::Xor:
    ans = new ConstantInt(T_INT_32, a ^ b);
  default:
    return nullptr;
  }
  return ans;
}

ConstantFloat * CalcFloat(Instruction::OpID op, Module* m, ConstantFloat *v1, ConstantFloat *v2) {
  auto a = v1->value_, b = v2->value_;
  ConstantFloat * ans;
  switch (op) {
  case Instruction::FAdd:
    ans = new ConstantFloat(T_FLOAT_, a + b);
  case Instruction::FSub:
    ans = new ConstantFloat(T_FLOAT_, a - b);
  case Instruction::FMul:
    ans = new ConstantFloat(T_FLOAT_, a * b);
  case Instruction::FDiv:
    ans = new ConstantFloat(T_FLOAT_, a / b);
  default:
    return nullptr;
  }
  return ans;
}

ConstantInt * CalcICMP(ICmpInst::ICmpOp op, Module* m, ConstantInt *v1, ConstantInt *v2) {
  auto a = v1->value_, b = v2->value_;
  switch (op) {
  case ICmpInst::ICMP_EQ:
    return new ConstantInt(T_INT_1, a == b);
  case ICmpInst::ICMP_NE:
    return new ConstantInt(T_INT_1, a != b);
  case ICmpInst::ICMP_SGT:
    return new ConstantInt(T_INT_1, a > b);
  case ICmpInst::ICMP_SGE:
    return new ConstantInt(T_INT_1, a >= b);
  case ICmpInst::ICMP_SLE:
    return new ConstantInt(T_INT_1, a <= b);
  case ICmpInst::ICMP_SLT:
    return new ConstantInt(T_INT_1, a < b);
  case ICmpInst::ICMP_UGE:
    return new ConstantInt(T_INT_1, (unsigned)a >= (unsigned)b);
  case ICmpInst::ICMP_ULE:
    return new ConstantInt(T_INT_1, (unsigned)a <= (unsigned)b);
  case ICmpInst::ICMP_ULT:
    return new ConstantInt(T_INT_1, (unsigned)a < (unsigned)b);
  case ICmpInst::ICMP_UGT:
    return new ConstantInt(T_INT_1, (unsigned)a > (unsigned)b);
  default:
    return nullptr;
  }
}

ConstantInt * CalcFCMP(FCmpInst::FCmpOp op, Module* m, ConstantFloat *v1, ConstantFloat *v2) {
  auto a = v1->value_, b = v2->value_;
  switch (op) {
  case FCmpInst::FCMP_UEQ:
    return new ConstantInt(T_INT_1, a == b);
  case FCmpInst::FCMP_UNE:
    return new ConstantInt(T_INT_1, a != b);
  case FCmpInst::FCMP_UGT:
    return new ConstantInt(T_INT_1, a > b);
  case FCmpInst::FCMP_UGE:
    return new ConstantInt(T_INT_1, a >= b);
  case FCmpInst::FCMP_ULE:
    return new ConstantInt(T_INT_1, a <= b);
  case FCmpInst::FCMP_ULT:
    return new ConstantInt(T_INT_1, a < b);
  case FCmpInst::FCMP_FALSE:
    return new ConstantInt(T_INT_1, 0);
  case FCmpInst::FCMP_TRUE:
    return new ConstantInt(T_INT_1, 1);
  case FCmpInst::FCMP_OEQ:
    return new ConstantInt(T_INT_1, a == b);
  case FCmpInst::FCMP_ONE:
    return new ConstantInt(T_INT_1, a != b);
  case FCmpInst::FCMP_OGE:
    return new ConstantInt(T_INT_1, a >= b);
  case FCmpInst::FCMP_OGT:
    return new ConstantInt(T_INT_1, a > b);
  case FCmpInst::FCMP_OLE:
    return new ConstantInt(T_INT_1, a <= b);
  case FCmpInst::FCMP_OLT:
    return new ConstantInt(T_INT_1, a < b);
  default:
    return nullptr;
  }
}



