#include "opt.h"
#include <functional>
#include <vector>


#define T_INT_32 (m->int32_ty_)
#define T_FLOAT_ (m->float32_ty_)
#define T_INT_1 (m->int1_ty_)


template <typename T>
void process_if_PhiInst(T* val, int arg_no) {
      
    auto instr = static_cast<PhiInst*>(val);
    instr->remove_operands(arg_no - 1, arg_no);
}




std::vector<BasicBlock *> dfs (std::set<BasicBlock *>& tempv, BasicBlock * place) {
  std::vector<BasicBlock *> ans;
  tempv.insert(place);
  for (auto t : place->succ_bbs_) {
    if (tempv.find(t) == tempv.end())
      dfs(tempv, t);
  }
  ans.emplace_back(place);
  return ans;
};

void deepgraph(BasicBlock *bb, std::set<BasicBlock *> &tempv) {
  if (bb == nullptr)
    return;
  tempv.insert(bb);
  for (auto suc = bb->succ_bbs_.rbegin() ; suc != bb->succ_bbs_.rend(); ++suc) {
    if (tempv.find(*suc) == tempv.end())
      deepgraph(*suc, tempv);
  }
}

bool phs_(BasicBlock *bb, BasicBlock *suc) {
  std::list<Instruction *> phi_us;
  auto list_inst = suc -> instr_list_;
  for (auto instr = list_inst.begin() ; instr != list_inst.end(); ++instr) {
    if((*instr)->op_id_ != Instruction::PHI) {
      continue;
    }
    if ((*instr)->op_id_ == Instruction::PHI) {
      for (unsigned int j = 1; j < (*instr)->num_ops_; j = j + 2)
        if ((*instr)->get_operand(j) == bb) {
          (*instr)->remove_operands(j - 1, j);
          break;
        }
      unsigned int size_instr = (*instr)->parent_->pre_bbs_.size();
      if (size_instr == 1) {
        valptr v1 = (*instr)->get_operand(0);
        (*instr)->replace_all_use_with(v1);
        phi_us.emplace_back(*instr);
      }
    }
  }
  for (auto i = phi_us.rbegin(); i != phi_us.rend() ;++i)
    suc->delete_instr(*i);
  return false;
}


void Deletebbs(Fptr  func) {
  std::set<BasicBlock *> tempv; std::vector<BasicBlock*> bb = func->basic_blocks_;
  int n = func->basic_blocks_.size();
  for (int i=0; i < n; ++i)
    if (bb[i]->name_ == "label_entry") {
      deepgraph(bb[i], tempv);
      break;
    }
  for (int i=0; i<n;++i)
    if (tempv.find(bb[i]) == tempv.end()) {
      bb[i]->parent_->remove_bb(bb[i]);
      for (auto suc : bb[i]->succ_bbs_)
        phs_(bb[i], suc);
    }
}


void DTree::implement() {
  for (auto f : m->function_list_)
    if (f->basic_blocks_.size() > 0) {
      Blockdom(f);
      BlockFront(f);
      for (auto bb : f->basic_blocks_) {
        bb->rdoms_.clear();
        bb->rdom_frontier_.clear();
      }
      getDomR(f);
      FrontR(f);
      getRDom(f);
    }
}

BasicBlock *DTree::compare_judge(BasicBlock *b1, BasicBlock *b2) {
  auto tval1 = b1, tval2 = b2;
  while (tval1 != tval2) {
    while (TInd[tval1] < TInd[tval2])
      tval1 = doBBs[TInd[tval1]];
    while (TInd[tval2] < TInd[tval1])
      tval2 = doBBs[TInd[tval2]];
  }
  return tval1;
}

std::vector<BasicBlock *> DTree::postTraverse(BasicBlock *bb) {
  std::set<BasicBlock *> tempv;
  return dfs(tempv, bb);
}

void DTree::getPostRraverse_2(Fptr  f) {
  doBBs.clear();
  rPostTraverse.clear();
  TInd.clear();
  auto entryBlock = *f->basic_blocks_.begin();
  auto seq = postTraverse(entryBlock);
  std::reverse(rPostTraverse.begin(), rPostTraverse.end());
  for (int i = 0; i < seq.size(); i++) {
    TInd[seq[i]] = i;
  }
  rPostTraverse = seq;
}

void DTree::Blockdom(Fptr  f) {
  getPostRraverse_2(f);
  auto t = *f->basic_blocks_.begin();
  auto t_id = TInd[t];
  doBBs.resize(t_id + 1, nullptr);
  doBBs.back() = t;
  bool change = true;
  while (change) {
    change = false;
    for (auto bb : rPostTraverse)
      if (bb != t) {
        auto preds = bb->pre_bbs_;
        BasicBlock *curDom = nullptr;
        for (auto pred_bb : preds)
          if (doBBs[TInd[pred_bb]] != nullptr) {
            curDom = pred_bb;
            break;
          }
        for (auto pred_bb : preds)
          if (doBBs[TInd[pred_bb]] != nullptr)
            curDom = compare_judge(pred_bb, curDom);
        if (doBBs[TInd[bb]] != curDom) {
          doBBs[TInd[bb]] = curDom;
          change = true;
        }
      }
  }
  for (auto bb : rPostTraverse)
    bb->idom_ = doBBs[TInd[bb]];
}

void DTree::BlockFront(Fptr  arg_f) {
  for (auto b : arg_f->basic_blocks_) {
    auto b_pred = b->pre_bbs_;
    if (b_pred.size() >= 2) {
      for (auto pred : b_pred) {
        auto r = pred;
        while (r != doBBs[TInd[b]]) {
          r->dom_frontier_.insert(b);
          r = doBBs[TInd[r]];
        }
      }
    }
  }
}



bool DTree::TraversePost(BasicBlock *bb, std::set<BasicBlock *> &tempvited) {
  tempvited.insert(bb);
  for (auto parent : bb->pre_bbs_)
    if (tempvited.find(parent) == tempvited.end())
      TraversePost(parent, tempvited);
  rTInd[bb] = rTraverse.size();
  rTraverse.emplace_back(bb);
  return true;
}

bool DTree::getPostRraverse(Fptr  f) {
  rDomainBlock.clear();
  rTraverse.clear();
  rTInd.clear();
  for (auto bb : f->basic_blocks_) {
    auto terminate_instr = bb->get_terminator();
    if (terminate_instr->op_id_ == Instruction::Ret) {
      BB_exit = bb;
      break;
    }
  }
  assert(BB_exit != nullptr);
  std::set<BasicBlock *> tempvited = {};
  TraversePost(BB_exit, tempvited);
  std::reverse(rTraverse.begin(), rTraverse.end());
  return false;
}

void DTree::getDomR(Fptr  f) {
  getPostRraverse(f);
  auto t = BB_exit;
  auto t_id = rTInd[t];
  for (int i = 0; i < t_id; i++)
    rDomainBlock.emplace_back(nullptr);
  rDomainBlock.emplace_back(t);
  bool change = true;
  while (change) {
    change = false;
    for (auto bb : rTraverse) {
      if (bb != t) {
        BasicBlock *new_irdom = nullptr;
        for (auto rpred_bb : bb->succ_bbs_)
          if (rDomainBlock[rTInd[rpred_bb]] != nullptr) {
            new_irdom = rpred_bb;
            break;
          }
        for (auto rpred_bb : bb->succ_bbs_)
          if (rDomainBlock[rTInd[rpred_bb]] != nullptr)
            new_irdom = compare_judge(rpred_bb, new_irdom);
        if (rDomainBlock[rTInd[bb]] != new_irdom) {
          rDomainBlock[rTInd[bb]] = new_irdom;
          change = true;
        }
      }
    }
  }
}

void DTree::getRDom(Fptr  f) {
  for (auto bb : f->basic_blocks_) {
    if (bb == BB_exit)
      continue;
    auto current = bb;
    while (current != BB_exit) {
      bb->rdoms_.insert(current);
      current = rDomainBlock[rTInd[current]];
    }
  }
}
void DTree::FrontR(Fptr  f) {
  for (auto bb_iter = f->basic_blocks_.rbegin();
       bb_iter != f->basic_blocks_.rend(); bb_iter++) {
    auto bb = *bb_iter;
    if (bb->succ_bbs_.size() >= 2) {
      for (auto rpred : bb->succ_bbs_) {
        auto r = rpred;
        while (r != rDomainBlock[rTInd[bb]]) {
          r->rdom_frontier_.insert(bb);
          r = rDomainBlock[rTInd[r]];
        }
      }
    }
  }
}



   void SJump::implement() {
     for (auto arg_f : m->function_list_)
       if (arg_f->basic_blocks_.empty()) {
         solveunREACH(arg_f);
         mergeBBs(arg_f);
         solvePHI(arg_f);
         solveJUMP(arg_f);
       }
   }

   void SJump::deleteUselessBlock(Fptr  arg_f, std::vector<BasicBlock *> &uselessBlock) {
     for (auto bb : uselessBlock)
       arg_f->remove_bb(bb);
   }

   bool SJump::checkUselessJump(BasicBlock *bb) {
     auto JumpPos = bb->get_terminator()->get_operand(0);
     for (auto preBB : bb->pre_bbs_) {
       auto br = preBB->get_terminator();
       if (br->operands_.size() == 1)
         continue;
       auto truepos = br->get_operand(1);
       auto falsepos = br->get_operand(2);
       if (truepos == JumpPos || falsepos == JumpPos)
         return false;
     }
     return true;
   }

   void SJump::solveunREACH(Fptr  arg_f) {
     std::vector<BasicBlock *> uselessBlock;
     for (int i = 2; i < arg_f->basic_blocks_.size(); i++) {
       auto curbb = arg_f->basic_blocks_[i];
       if (curbb->pre_bbs_.empty()) {
         uselessBlock.emplace_back(curbb);
         for (auto use : curbb->use_list_) {
           if(dynamic_cast<PhiInst *>(use.val_) != nullptr)
             process_if_PhiInst(use.val_, use.arg_no_) ;
         }
       }
     }
   }

   void SJump::mergeBBs(Fptr  arg_f) {
     std::vector<BasicBlock *> uselessBlock;
     for (int i = 2; i < arg_f->basic_blocks_.size(); i++) {
       auto bb = arg_f->basic_blocks_[i];
       if (bb->pre_bbs_.size() == 1) {
         auto preBlock = *bb->pre_bbs_.begin();
         auto preBr = preBlock->get_terminator();
         if (preBlock->succ_bbs_.size() != 1)
           continue;
         preBlock->delete_instr(preBr);
         for (auto instr : bb->instr_list_) {
           preBlock->add_instruction(instr);
           bb->remove_instr(instr);
         }
         preBlock->remove_succ_basic_block(bb);
         for (auto suc : bb->succ_bbs_) {
           preBlock->add_succ_basic_block(suc);
           suc->remove_pre_basic_block(bb);
           suc->add_pre_basic_block(preBlock);
         }
         bb->replace_all_use_with(preBlock);
         uselessBlock.emplace_back(bb);
       }
     }
     deleteUselessBlock(arg_f, uselessBlock);
   }

   void SJump::solvePHI(Fptr  arg_f) {
     for (auto bb : arg_f->basic_blocks_)
       if (bb->pre_bbs_.size() == 1)
         for (auto instr : bb->instr_list_)
           if (is_phi(instr)) {
             instr->replace_all_use_with(instr->get_operand(0));
             bb->delete_instr(instr);
           }
   }

   void SJump::solveJUMP(Fptr  arg_f) {
     std::vector<BasicBlock *> uselessBlock;
     for (int i = 2; i < arg_f->basic_blocks_.size(); i++) {
       BasicBlock *curbb = arg_f->basic_blocks_[i];
       if (curbb->instr_list_.size() != 1)
         continue;
       auto branchInstr = curbb->get_terminator();
       if (branchInstr->operands_.size() != 1 || is_ret(branchInstr))
         continue;
       if (!checkUselessJump(curbb))
         continue;
       uselessBlock.emplace_back(curbb);
       auto JumpTarget = dynamic_cast<BasicBlock *>(branchInstr->get_operand(0));
       for (auto instr : JumpTarget->instr_list_)
         if (is_phi(instr)) {
           for (int i = 1; i < instr->operands_.size(); i += 2) {
             if (instr->get_operand(i) == curbb) {
               auto val = instr->get_operand(i - 1);
               instr->remove_operands(i - 1, i);
               for (auto preBB : curbb->pre_bbs_) {
                 instr->add_operand(val);
                 instr->add_operand(preBB);
               }
               break;
             }
           }
         }
       curbb->replace_all_use_with(JumpTarget);
       for (auto preBB : curbb->pre_bbs_) {
         preBB->add_succ_basic_block(JumpTarget);
         JumpTarget->add_pre_basic_block(preBB);
       }
     }
     deleteUselessBlock(arg_f, uselessBlock);
   }

   void deleteUse(valptr opnd, Instruction *inst) {
     for (auto it = opnd->use_list_.begin(); it != opnd->use_list_.end(); ++it)
       if (it->val_ == inst) {
         opnd->use_list_.erase(it);
         return;
       }
   }



   void CInstr::implement() {
     assert(m != nullptr);
     for (auto arg_f : m->function_list_)
       if (!arg_f->basic_blocks_.empty())
         for (BasicBlock *bb : arg_f->basic_blocks_)
           checkBBs(bb);
  
     for (Fptr  arg_f : m->function_list_) {
       if (arg_f->basic_blocks_.size()) {
         bool change = true;
         while (change) {
           change = false;
           change |= forwardconst(arg_f);
           change |= solveBranch(arg_f);
           Deletebbs(arg_f);
         }
       }
     }
   }

   void CInstr::checkBBs(BasicBlock *bb) {
     bool change = true;
     while (change) {
       change = false;
       for (auto instr : bb->instr_list_) {
         if (instr->op_id_ != Instruction::Add &&
             instr->op_id_ != Instruction::Sub)
           continue;
         if (instr->use_list_.size() != 1 || instr->use_list_.back().arg_no_ != 0)
           continue;
         Instruction *nextInstr = dynamic_cast<Instruction *>(instr->use_list_.back().val_);
         if (nextInstr == nullptr || instr->op_id_ != nextInstr->op_id_ ||
             instr->parent_ != nextInstr->parent_)
           continue;
         std::map<valptr , unsigned int> Optime;
         Instruction *StartInstr = instr, *EndInstr = nullptr;
         Instruction *invalidStart = nullptr, *invalidEnd = nullptr; 
         bool isRepeat = false;
         valptr dupOp = nullptr;
         valptr Candi0 = instr->get_operand(0);
         valptr  Candi1 = instr->get_operand(1);
         Optime[Candi0]++;
         Optime[Candi1]++;
         Optime[nextInstr->get_operand(1)]++;
         instr = nextInstr;
         while (instr->use_list_.size() == 1 &&
                instr->use_list_.back().arg_no_ == 0) {
           nextInstr = dynamic_cast<Instruction *>(instr->use_list_.back().val_);
           if (nextInstr == nullptr || instr->op_id_ != nextInstr->op_id_ ||
               instr->parent_ != nextInstr->parent_)
             break;
           if (!isRepeat) {
             if (Optime.find(nextInstr->get_operand(1)) != Optime.end())
               isRepeat = true;
             EndInstr = instr;
             invalidStart = nextInstr;
           } else if (Optime.find(nextInstr->get_operand(1)) == Optime.end())
             break;
           Optime[nextInstr->get_operand(1)]++;
           instr = nextInstr;
         }
         invalidEnd = instr;
         unsigned int dupTime = 0;
         if (Optime[Candi1] == 1 && Optime[Candi0] > 1) {
           dupOp = Candi0;
           dupTime = Optime[Candi0];
         } else if (Optime[Candi0] == 1 && Optime[Candi1] > 1) {
           dupOp = Candi1;
           dupTime = Optime[Candi1];
         } else
           continue;
         for (auto p : Optime) {
           if (p.second == 1) {
             if (p.first != dupOp) {
               dupTime = 0;
               break;
             }
           } else {
             if (dupTime != p.second) {
               dupTime = 0;
               break;
             }
           }
         }
         if (!dupTime)
           continue;
         ConstantInt *dupTimeConst =
             new ConstantInt(instr->parent_->parent_->parent_->int32_ty_, dupTime);
         Instruction *toMulInst = new BinaryInst(
             bb->parent_->parent_->int32_ty_, Instruction::Mul,
             static_cast<valptr >(EndInstr), dupTimeConst, bb, true);
         toMulInst->name_ = invalidStart->name_;
         bb->add_instruction_before_inst(toMulInst, invalidStart);
         invalidEnd->replace_all_use_with(toMulInst);
         for (Instruction *ins = invalidStart; ins != nextInstr;) {
           bb->delete_instr(ins);
           ins = dynamic_cast<Instruction *>(ins->use_list_.back().val_);
         }
         change = true;
         break;
       }
     }
   }



   bool CInstr::forwardconst(Fptr  func) {
     uselessInstr.clear();
     for (auto bb : func->basic_blocks_) {
       val_constant.clear();
       val_float.clear();
       for (auto instr : bb->instr_list_) {
         ConstantInt* tempa = nullptr , *tempb= nullptr;
         ConstantFloat* testConstFloata = nullptr, * testConstFloatb = nullptr;
         auto id = instr->op_id_;
         if(id == Instruction::Add||id == Instruction::Sub || id == Instruction::Mul || id== Instruction::SDiv||id == Instruction::UDiv|| id == Instruction::SRem||id == Instruction::URem 
             || id ==Instruction::And||id == Instruction::Or||id == Instruction::Xor||id == Instruction::Shl||id == Instruction::AShr||id == Instruction::LShr) {

               tempa = dynamic_cast<ConstantInt *>(instr->get_operand(0));
               tempb = dynamic_cast<ConstantInt *>(instr->get_operand(1));
               if (tempa && tempb) {
                 auto intRes = CalcInt(instr->op_id_, this->m, tempa, tempb);
                 if (intRes) {
                   instr->replace_all_use_with(intRes);
                   uselessInstr[instr] = bb;
                 }
               }
             }
         else if (id == Instruction::ICmp) {
           tempa = dynamic_cast<ConstantInt *>(instr->get_operand(0));
           tempb = dynamic_cast<ConstantInt *>(instr->get_operand(1));
           if (tempa && tempb) {
             auto res = CalcICMP(dynamic_cast<ICmpInst *>(instr)->icmp_op_,this->m, tempa, tempb);
             if (res) {
               instr->replace_all_use_with(res);
               uselessInstr[instr] = bb;
             }
           }
         }
         else if (id == Instruction::FCmp) {
           testConstFloata = dynamic_cast<ConstantFloat *>(instr->get_operand(0));
           testConstFloatb = dynamic_cast<ConstantFloat *>(instr->get_operand(1));
           if (testConstFloata && testConstFloatb) {
             auto res = CalcFCMP(dynamic_cast<FCmpInst *>(instr)->fcmp_op_,this->m, 
                                       testConstFloata, testConstFloatb);
             if (res) {
               instr->replace_all_use_with(res);
               uselessInstr[instr] = bb;
             }
           }
         }
         else if (id == Instruction::FAdd||id == Instruction::FSub||id == Instruction::FMul||id == Instruction::FDiv) {
           testConstFloata = dynamic_cast<ConstantFloat *>(instr->get_operand(0));
           testConstFloatb = dynamic_cast<ConstantFloat *>(instr->get_operand(1));
           if (testConstFloata && testConstFloatb) {
             auto floaRes =
                 CalcFloat(instr->op_id_, this->m, testConstFloata, testConstFloatb);
             if (floaRes) {
               instr->replace_all_use_with(floaRes);
               uselessInstr[instr] = bb;
             }
           }
         }
         else if (id == Instruction::FNeg) {
           testConstFloata = dynamic_cast<ConstantFloat *>(instr->get_operand(0));
           if (testConstFloata) {
             instr->replace_all_use_with(
                 new ConstantFloat(T_FLOAT_, -testConstFloata->value_));
             uselessInstr[instr] = bb;
           }
         }
         else if(id == Instruction::FPtoSI) {
           testConstFloata = dynamic_cast<ConstantFloat *>(instr->get_operand(0));
           if (testConstFloata) {
             instr->replace_all_use_with(
                 new ConstantInt(T_INT_32, testConstFloata->value_));
             uselessInstr[instr] = bb;
           }
         }
         else if (id == Instruction::SItoFP) {

           tempa = dynamic_cast<ConstantInt *>(instr->get_operand(0));
           if (tempa) {
             instr->replace_all_use_with(
                 new ConstantFloat(T_FLOAT_, tempa->value_));
             uselessInstr[instr] = bb;
           }
         }
         else if(id == Instruction::ZExt) {

           tempa = dynamic_cast<ConstantInt *>(instr->get_operand(0));
           if (tempa) {
             instr->replace_all_use_with(
                 new ConstantInt(T_INT_32, tempa->value_));
             uselessInstr[instr] = bb;
           }
         }
         else if(Instruction::Call) {

           val_constant.clear();
           val_float.clear();
           auto globalVar = dynamic_cast<GlobalVariable *>(instr->get_operand(0));
           if (globalVar) {
             auto iterInt = val_constant.find(globalVar);
             auto iterFloat = val_float.find(globalVar);
             if (iterInt != val_constant.end()) {
               instr->replace_all_use_with(iterInt->second);
               uselessInstr[instr] = bb;
             } else if (iterFloat != val_float.end()) {
               instr->replace_all_use_with(iterFloat->second);
               uselessInstr[instr] = bb;
             }
           } else if (dynamic_cast<AllocaInst *>(instr->get_operand(0))) {
             auto pos = dynamic_cast<AllocaInst *>(instr->get_operand(0));
             if (pos->alloca_ty_->tid_ == Type::IntegerTyID) {
               auto iterInt = val_constant.find(pos);
               if (iterInt != val_constant.end()) {
                 instr->replace_all_use_with(iterInt->second);
                 uselessInstr[instr] = bb;
               }
             } else if (pos->alloca_ty_->tid_ == Type::FloatTyID) {
               auto iterFloat = val_float.find(pos);
               if (iterFloat != val_float.end()) {
                 instr->replace_all_use_with(iterFloat->second);
                 uselessInstr[instr] = bb;
               }
             }
           }
         }
         else if(id == Instruction::Load) {

           auto globalVar = dynamic_cast<GlobalVariable *>(instr->get_operand(0));
           if (globalVar) {
             auto iterInt = val_constant.find(globalVar);
             auto iterFloat = val_float.find(globalVar);
             if (iterInt != val_constant.end()) {
               instr->replace_all_use_with(iterInt->second);
               uselessInstr[instr] = bb;
             } else if (iterFloat != val_float.end()) {
               instr->replace_all_use_with(iterFloat->second);
               uselessInstr[instr] = bb;
             }
           } else if (dynamic_cast<AllocaInst *>(instr->get_operand(0))) {
             auto pos = dynamic_cast<AllocaInst *>(instr->get_operand(0));
             if (pos->alloca_ty_->tid_ == Type::IntegerTyID) {
               auto iterInt = val_constant.find(pos);
               if (iterInt != val_constant.end()) {
                 instr->replace_all_use_with(iterInt->second);
                 uselessInstr[instr] = bb;
               }
             } else if (pos->alloca_ty_->tid_ == Type::FloatTyID) {
               auto iterFloat = val_float.find(pos);
               if (iterFloat != val_float.end()) {
                 instr->replace_all_use_with(iterFloat->second);
                 uselessInstr[instr] = bb;
               }
             }
           }
         } 
         else if(id == Instruction::Store) 
         {
           std::cout << "EVER STORE\n";
           auto storePos = instr->get_operand(1);
           auto storeValInt = dynamic_cast<ConstantInt *>(instr->get_operand(0));
           auto storeValFloat = dynamic_cast<ConstantFloat *>(instr->get_operand(0));
           if (storeValInt!=nullptr) 
           {
             auto iter1 = val_constant.find(storePos);
             if (iter1 != val_constant.end()) {
               if (iter1->second->value_ == storeValInt->value_)
                 uselessInstr[instr] = bb;
               else
                 iter1->second = storeValInt;
             } 
             else
               val_constant[storePos] = storeValInt;
           } 
           else if (storeValFloat!=nullptr) 
           {
             auto iter = val_float.find(storePos);
             if (iter != val_float.end()) 
             {
               if (iter->second->value_ == storeValInt->value_)
                 uselessInstr[instr] = bb;
               else
                 iter->second = storeValFloat;
             } 
             else
               val_float[storePos] = storeValFloat;
           } 
           else 
           {
             auto iterInt = val_constant.find(storePos);
             auto iterFloat = val_float.find(storePos);
             if (iterInt != val_constant.end())
               val_constant.erase(iterInt);
             if (iterFloat != val_float.end())
               val_float.erase(iterFloat);
           }
         } 
       }
     }
     if (!uselessInstr.empty()) {
       for (auto [instr, bb] : uselessInstr)
         bb->delete_instr(instr);
       return true;
     }
     return false;
   }

   bool CInstr::solveBranch(Fptr  func) {
     bool change = false;
     for (auto bb : func->basic_blocks_) {
       auto br = bb->get_terminator();
       if (!br)
         continue;
       if (br->op_id_ == Instruction::Br &&
           dynamic_cast<BranchInst *>(br)->num_ops_ == 3) {
         auto cond = dynamic_cast<ConstantInt *>(br->get_operand(0));
         auto truebb = br->get_operand(1);
         auto falsebb = br->get_operand(2);
         if (!cond)
           continue;
         change = true;
         if (cond->value_ == 0) {
           bb->delete_instr(br);
           for (auto succ_bb : bb->succ_bbs_) {
             succ_bb->remove_pre_basic_block(bb);
             if (succ_bb != falsebb) {
               phs_(bb, succ_bb);
             }
           }
           bb->succ_bbs_.clear();
           new BranchInst(dynamic_cast<BasicBlock *>(falsebb), bb);
         } else {
           bb->delete_instr(br);
           for (auto succ_bb : bb->succ_bbs_) {
             succ_bb->remove_pre_basic_block(bb);
             if (succ_bb != truebb) {
               phs_(bb, succ_bb);
             }
           }
           bb->succ_bbs_.clear();
           new BranchInst(dynamic_cast<BasicBlock *>(truebb), bb);
         }
       }
     }
     return change;
   }



void DCodeSolve::initFuncPtrArg() {
  for (auto arg_f : m->function_list_) {
    if (arg_f->basic_blocks_.empty())
      continue;
    for (auto arg : arg_f->arguments_)
      if (arg->type->tid_ == Type::PointerTyID) {
        if (!funcPtrArgs.count(arg_f))
          funcPtrArgs[arg_f] = {};
        funcPtrArgs[arg_f].insert(arg);
      }
  }
}
void DCodeSolve::Init(Fptr  arg_f) {
  storePos.clear();
  for (auto bb : arg_f->basic_blocks_) {
    for (auto ins : bb->instr_list_) {
      auto id = ins->op_id_;
      switch (id)
      {
        case Instruction::Store:
        {
          if (!storePos.count(ins->get_operand(1))) 
          {
            storePos.insert({ins->get_operand(1), {}});
          }
          storePos[ins->get_operand(1)].emplace_back(ins);
          if (dynamic_cast<GlobalVariable *>(ins->get_operand(1)))
            OptFunc.insert(arg_f->name_);
          if (dynamic_cast<GetElementPtrInst *>(ins->get_operand(1)))
            OptFunc.insert(arg_f->name_);
          if (funcPtrArgs[arg_f].count(ins->get_operand(1)))
            OptFunc.insert(arg_f->name_);
        } break;
        case Instruction::Call: 
        {
          auto f = ins->get_operand(ins->operands_.size() - 1);
          if (OptFunc.count(f->name_))
            OptFunc.insert(arg_f->name_);
        }break;
        default:
          break;
      }
    }
  }
}
bool DCodeSolve::checkOpt(Fptr  arg_f, Instruction *ins) {
  if (ins->op_id_ == Instruction::Ret) {
    BB_exit = ins->parent_;
    return true;
  } else if (ins->op_id_ == Instruction::Call) {
    auto f = ins->get_operand(ins->operands_.size() - 1);
    return OptFunc.count(f->name_);
  } else if (ins->op_id_ == Instruction::Store) {
    if (dynamic_cast<GlobalVariable *>(ins->get_operand(1)))
      return true;
    if (dynamic_cast<GetElementPtrInst *>(ins->get_operand(1)))
      return true;
    if (funcPtrArgs[arg_f].count(ins->get_operand(1)))
      return true;
    return false;
  }
  return false;
}

void DCodeSolve::findInstr(Fptr  arg_f) {
  std::list<valptr > workList;
  for (auto bb : arg_f->basic_blocks_) {
    for (auto ins : bb->instr_list_) {
      if (checkOpt(arg_f, ins)) {
        uselessInstr.insert(ins);
        workList.emplace_back(ins);
      }
    }
  }
  while (!workList.empty()) {
    auto ins = dynamic_cast<Instruction *>(workList.back());
    workList.pop_back();
    if (ins == nullptr) {
      continue;
    }
    for (auto operand : ins->operands_) {
      auto temp = dynamic_cast<Instruction *>(operand);
      if (!temp)
        continue;
      if (uselessInstr.insert(temp).second) {
        workList.emplace_back(temp);
      }
    }
    if (ins->op_id_ == Instruction::PHI) {
      for (int i = 1; i < ins->operands_.size(); i += 2) {
        auto bb = dynamic_cast<BasicBlock *>(ins->get_operand(i));
        auto br = bb->get_terminator();
        if (uselessInstr.insert(br).second) {
          workList.emplace_back(br);
        }
      }
    }
    if (storePos.count(ins)) {
      for (auto curInstr : storePos[ins]) {
        if (uselessInstr.insert(dynamic_cast<Instruction *>(curInstr)).second) {
          workList.emplace_back(curInstr);
        }
      }
      storePos.erase(ins);
    }
    if (uselessBlock.insert(ins->parent_).second) {
      for (auto RFrontier : ins->parent_->rdom_frontier_) {
        auto t = RFrontier->get_terminator();
        if (uselessInstr.insert(t).second) {
          workList.emplace_back(t);
        }
      }
    }
  }
}
void DCodeSolve::deleteInstr(Fptr  arg_f) {
  int dCnt = 0, cCnt = 0;
  for (auto bb : arg_f->basic_blocks_) {
    std::list<Instruction *> tempinstr;
    for (auto ins : bb->instr_list_) {
      if (!uselessInstr.count(ins)) {
        if (ins->op_id_ != Instruction::Br) {
          tempinstr.emplace_back(ins);
        } 
        else 
        {
          if (ins->operands_.size() == 3) {
            cCnt++;
            auto truepos = dynamic_cast<BasicBlock *>(ins->get_operand(1));
            auto falsepos = dynamic_cast<BasicBlock *>(ins->get_operand(2));
            truepos->remove_pre_basic_block(bb);
            falsepos->remove_pre_basic_block(bb);
            bb->remove_succ_basic_block(truepos);
            bb->remove_succ_basic_block(falsepos);
            BasicBlock *temp = BB_exit;
            std::vector<BasicBlock *> rdoms(bb->rdoms_.begin(),
                                            bb->rdoms_.end());
            std::sort(rdoms.begin(), rdoms.end(),
                      [=](BasicBlock *x, BasicBlock *y) -> bool {
                        return x->rdoms_.count(y);
                      });
            for (auto rdbb : rdoms) {
              if (rdbb != bb && uselessBlock.count(rdbb)) {
                temp = rdbb;
                break;
              }
            }
            ins->remove_operands(0, 2);
            ins->num_ops_ = 1;
            ins->operands_.resize(1);
            ins->use_pos_.resize(1);
            ins->set_operand(0, temp);
            bb->add_succ_basic_block(temp);
            temp->add_pre_basic_block(bb);
          }
        }
      }
    }
    dCnt += tempinstr.size();
    for (auto ins : tempinstr) {
      bb->delete_instr(ins);
    }
  }
}

void DCodeSolve::implement() {
  /*Reverse*/DTree rDTree(m);
  rDTree.implement();
  initFuncPtrArg();
  for (auto arg_f : m->function_list_)
    if (!arg_f->basic_blocks_.empty()) {
      Init(arg_f);
      findInstr(arg_f);
      deleteInstr(arg_f);
      Deletebbs(arg_f);
    }
}

   void LInvariant::implement() {
     searchLoop();
     while (!store_loop.empty()) {
       auto loop = store_loop.top();
       store_loop.pop();
       std::set<valptr > assignVals;
       std::set<Instruction *> tempvInstr; 
       std::vector<Instruction *> invarInstrs; 
       std::map<Instruction *, BasicBlock *> instrPos;
       for (auto bb : *loop)
         for (Instruction * inst : bb->instr_list_)
           if (is_binary(inst) || is_cmp(inst) || is_fcmp(inst) || is_call(inst) || is_phi(inst) || is_zext(inst) || is_fptosi(inst) || is_sitofp(inst) || is_gep(inst) || is_load(inst))
             assignVals.insert(inst);
       bool changed = true;
       while (changed) {
         changed = false;
         for (auto bb : *loop) {
           for (auto instr : bb->instr_list_) {
             if (tempvInstr.find(instr) != tempvInstr.end())
               continue;
             if (!is_gep(instr) && !is_alloca(instr) && !is_br(instr) && !is_ret(instr) && !is_phi(instr) && !is_store(instr) && !is_load(instr) && !(is_call(instr) &&
                   instr->get_operand(instr->num_ops_ - 1)->print() != "rand")) {
               bool move = true;
               for (unsigned int i = 0; i < instr->num_ops_; i++)
                 if (assignVals.find(instr->get_operand(i)) != assignVals.end())
                   move = false;
               if (move) {
                 instrPos[instr] = bb;
                 invarInstrs.emplace_back(instr);
                 assignVals.erase(instr);
                 tempvInstr.insert(instr);
                 changed = true;
               }
             }
           }
         }
       }
       auto enter = enterptr[loop];
       for (auto prev : enter->pre_bbs_)
         if (loop->find(prev) == loop->end())
           for (auto inst : invarInstrs)
             prev->add_instruction_before_terminator(inst);
     }
   }

   void LInvariant::searchLoop() {
     for (auto arg_f : m->function_list_) {
       if (arg_f->basic_blocks_.empty())
         continue;
       std::set<nodule *> nodules;
       std::set<nodule *> entry;
       std::set<std::set<nodule *> *> CSS;
       std::map<BasicBlock *, nodule *> noduleMap;
       for (auto bb : arg_f->basic_blocks_) {
         auto cur = new nodule(bb, -1, -1, 0);
         noduleMap[bb] = cur;
         nodules.insert(cur);
       }
       for (auto bb : arg_f->basic_blocks_) {
         auto BlockNode = noduleMap[bb];
         for (auto suc : bb->succ_bbs_)
           BlockNode->suc.insert(noduleMap[suc]);
         for (auto pre : bb->succ_bbs_)
           BlockNode->pre.insert(noduleMap[pre]);
       }

       while (LInvariant::searchC(nodules, CSS)) {
         for (auto C : CSS) {
           nodule *enter = nullptr;
           for (auto curBlock : *C)
             for (auto pre : curBlock->pre)
               if (C->find(pre) == C->end())
                 enter = curBlock;
               else if (entry.find(pre) != entry.end())
                 enter = pre;

           auto curLoop = new std::set<BasicBlock *>;
           for (auto curBlock : *C)
             curLoop->insert(curBlock->bb);
           enterptr[curLoop] = enter->bb;
           store_loop.push(curLoop);
           entry.insert(enter);
           nodules.erase(enter);
           for (auto pre : enter->pre)
             pre->suc.erase(enter);
           for (auto suc : enter->suc)
             suc->pre.erase(enter);
         }
         for (auto C : CSS)
           C->clear();
         CSS.clear();
         for (auto NodeBlock : nodules)
           NodeBlock = new nodule(nullptr, -1, -1, false);
       }
       for (auto nodule : nodules)
         delete nodule;
     }
   }

   bool LInvariant::searchC(std::set<nodule *> &basicBlocks, std::set<std::set<nodule *> *> &CSS) {
     ind = 0;
     while (!store_tj.empty())
       store_tj.pop();
     for (auto cur : basicBlocks)
       if (cur-> d_num == -1)
         tj(cur, CSS);
     return !CSS.empty();
   }

   void LInvariant::tj(nodule *cur, std::set<std::set<nodule *> *> &CSS) {
     cur-> d_num = cur->low = ++ind;
     cur->inStack = true;
     store_tj.push(cur);
     for (auto succ : cur->suc)
       if (succ-> d_num == -1) {
         tj(succ, CSS);
         if (succ->low < cur->low)
           cur->low = succ->low;
       } else if (succ->inStack && succ->low < cur->low)
         cur->low = succ->low;
     if (cur-> d_num == cur->low) {
       if (cur == store_tj.top()) {
         store_tj.pop();
         cur->inStack = false;
         return;
       }
       auto C = new std::set<nodule *>;
       nodule *tp = nullptr;
       do {
         tp = store_tj.top();
         C->insert(tp);
         store_tj.pop();
         tp->inStack = false;
       } while (tp != cur);
       CSS.insert(C);
     }
   }
