#include <iostream>
#include <map>
#include <memory>
#include "build.h"
#include "ircond.h"

int CreateStmt::calc_type_size(Type* ty) {
    int totalsize = 1;
    while (ty->tid_ == Type::ArrayTyID) {
        totalsize *= static_cast<ArrayType*>(ty)->num_elements_;
        ty = static_cast<ArrayType*>(ty)->contained_;
    }
   
    totalsize *= 4;
    return totalsize;
}

void CreateStmt::global_arr_init(LIRGlobalVariable* lirgv, Constant* gv) {
    Type* gv_type = gv->type;
    if (gv_type->tid_ == Type::IntegerTyID) {
        int gv_int_val = static_cast<ConstantInt*>(gv)->value_;
        lirgv->add_data_def(lirgv->parent_->get_LIRDataDef(LIRDataDef::LWordTy, gv_int_val));
    } else if (gv_type->tid_ == Type::FloatTyID) {
        float float_val = static_cast<ConstantFloat*>(gv)->value_;
        int float_val2 = *(int*)(&float_val);
        lirgv->add_data_def(lirgv->parent_->get_LIRDataDef(LIRDataDef::LWordTy, float_val2));
    } else if (gv_type->tid_ == Type::ArrayTyID) {
        if (dynamic_cast<ConstantArray*>(gv)) {
            unsigned int gv_arr_elemnum = static_cast<ArrayType*>(gv_type)->num_elements_;
            for (unsigned int i = 0; i < gv_arr_elemnum; i++) {
                global_arr_init(lirgv, (static_cast<ConstantArray*>(gv)->const_array)[i]);
            }
        } else if (dynamic_cast<ConstantZero*>(gv)) {
            int arr_size = calc_type_size(gv_type);
            lirgv->add_data_def(lirgv->parent_->get_LIRDataDef(LIRDataDef::LZeroTy, arr_size));
        } 
    } 
}
bool CreateStmt::is_constant_all_zero(Constant* gv) {
    Type* gv_type = gv->type;
    if (gv_type->tid_ == Type::IntegerTyID) {
        int gv_int_val = static_cast<ConstantInt*>(gv)->value_;
        if (gv_int_val == 0) {
            return true;
        } else {
            return false;
        }
    } else if (gv_type->tid_ == Type::FloatTyID) {
        float float_val = static_cast<ConstantFloat*>(gv)->value_;
        unsigned int float_val2 = *(unsigned int*)(&float_val);
        if (float_val2 == 0) {
            return true;
        } else {
            return false;
        }
    } else if (gv_type->tid_ == Type::ArrayTyID) {
        if (dynamic_cast<ConstantArray*>(gv)) {
            unsigned int gv_arr_elemnum = static_cast<ArrayType*>(gv_type)->num_elements_;
            for (unsigned int i = 0; i < gv_arr_elemnum; i++) {
                if (!is_constant_all_zero((static_cast<ConstantArray*>(gv)->const_array)[i])) {
                    return false;
                }
            }
            return true;
        } else if (dynamic_cast<ConstantZero*>(gv)) {
            return true;
        }
    } 
}
LIROperand* CreateStmt::get_LIROpnd(Value* v) {
    if (auto* a = dynamic_cast<ConstantInt*>(v)) {
        return m_lir->get_LIRConstant(a->value_);
    }
    if (auto* b = dynamic_cast<ConstantFloat*>(v)) {
        float float_val = b->value_;
        int float_val2 = *(int*)(&float_val);
        return m_lir->get_LIRConstant(float_val2);
    }
    if (auto* c = dynamic_cast<Argument*>(v)) {
          auto it = arg_map_.find(c);
        if (it != arg_map_.end()) {
            return it->second;
        } else {
            assert(0 && "Argument not found in arg_map!");
        }
    }
    if (dynamic_cast<AllocaInst*>(v)) {
       
        return nullptr;
    }
    if (auto* i = dynamic_cast<Instruction*>(v)) {
        auto it = inst_map_.find(i);
        if (it != inst_map_.end()) {
            return it->second;
        } else {
            assert(0 && "Instruction not found in inst_map!");
        }
    } else {
      
        return nullptr;
    }
}
void CreateStmt::clir() {
    for (auto* gv : m_->global_list_) {
        LIRGlobalVariable::LIRDataSectionTy lirgv_tid;
        if (is_constant_all_zero(gv->init_val_)) {
            lirgv_tid = LIRGlobalVariable::LBssTy;
        } else {
            lirgv_tid = LIRGlobalVariable::LDataTy;
        }
        LIRGlobalVariable* lirgv = new LIRGlobalVariable(gv->name_, m_lir, lirgv_tid);
        gv_map_[gv] = lirgv;
       
        Constant* gv_init = gv->init_val_;
        global_arr_init(lirgv, gv_init);
        if (print_s_verbosity_) {
        }
    }
    for (Function* f : m_->function_list_) {
        if (f->is_declaration()) {
            int result_type;
            switch (f->get_return_type()->tid_) {
                case Type::VoidTyID:
                    result_type = 0;
                    break;
                case Type::IntegerTyID:
                    result_type = 1;
                    break;
                case Type::FloatTyID:
                    result_type = 2;
                    break;
                default:
                   
                    break;
            }
            LIRFunction* lirf = new LIRFunction(f->name_, result_type, f->arguments_.size());
            func_map_[f] = lirf;
            int num_int_args = 0, num_float_args = 0;
            for (size_t idx = 0; idx < f->arguments_.size(); idx++) {
                auto* arg = f->arguments_[idx];
                if (arg->type->tid_ == Type::FloatTyID) {
                    assert(num_float_args < 16);
                    lirf->args_passmethod_.push_back(m_lir->get_LIRPhysFReg(-1 - num_float_args));
                    num_float_args++;
                } else {
                    assert(num_int_args < 4);
                    lirf->args_passmethod_.push_back(m_lir->get_LIRPhysReg(num_int_args + 1));
                    num_int_args++;
                }
            }
            continue;
        }
        int result_type;
        switch (f->get_return_type()->tid_) {
            case Type::VoidTyID:
                result_type = 0;
                break;
            case Type::IntegerTyID:
                result_type = 1;
                break;
            case Type::FloatTyID:
                result_type = 2;
                break;
            default:
              
                break;
        }
        LIRFunction* lirf = new LIRFunction(f->name_, m_lir, result_type, f->arguments_.size());
        func_map_[f] = lirf;
        LIRBasicBlock* lirbb;
        lirbb = new LIRBasicBlock("init", lirf);
        set_insert_point(lirbb);
        int num_int_args = 0, num_float_args = 0;
        int mempass_idx = 0;
        for (size_t idx = 0; idx < f->arguments_.size(); idx++) {
            auto* arg = f->arguments_[idx];
            if (arg->type->tid_ == Type::FloatTyID) {
                auto* lirvfreg = m_lir->get_LIRVirtFReg();
                arg_map_[arg] = lirvfreg;
                if (num_float_args >= 16) {
                    auto* ttt = new LIRVldrInst(LIRInstruction::LCondNone, lirvfreg, m_lir->get_LIRPhysReg(14), m_lir->get_LIRConstant(0));
                    lirbb->add_inst_back(ttt);
                    std::pair<std::map<LIRInstruction*,int>::iterator,bool> rrr =lirf->args_offset_.emplace(ttt,mempass_idx);
                    if(rrr.second){
                        ttt->args_offset_use_pos_.push_back(rrr.first);
                    }
                    lirf->args_passmethod_.push_back(nullptr);
                    mempass_idx += 4;
                } else {
                    create_vmovf32(LIRInstruction::LCondNone, lirvfreg, m_lir->get_LIRPhysFReg(-1 - num_float_args));
                    lirf->args_passmethod_.push_back(m_lir->get_LIRPhysFReg(-1 - num_float_args));
                }
                num_float_args++;
            } else {
               
                auto* lirvreg = m_lir->get_LIRVirtReg();
                arg_map_[arg] = lirvreg;
                if (num_int_args >= 4) {
                    LIRInstruction* ttt = new LIRLdrInst(LIRInstruction::LCondNone, lirvreg, m_lir->get_LIRPhysReg(14), m_lir->get_LIRConstant(0), LIRInstruction::LRegisterOffsetTy);
                    lirbb->add_inst_back(ttt);
                    std::pair<std::map<LIRInstruction*,int>::iterator,bool> rrr =lirf->args_offset_.emplace(ttt,mempass_idx);
                    if(rrr.second){
                        ttt->args_offset_use_pos_.push_back(rrr.first);
                    }
                    lirf->args_passmethod_.push_back(nullptr);
                    mempass_idx += 4;
                } else {
                    create_mov(LIRInstruction::LCondNone, lirvreg, m_lir->get_LIRPhysReg(num_int_args + 1));
                    lirf->args_passmethod_.push_back(m_lir->get_LIRPhysReg(num_int_args + 1));
                }
                num_int_args++;
            }
        }
        lirf->stackframe_args_size_ = (mempass_idx % 8 == 0) ? (mempass_idx) : ((mempass_idx / 8 + 1) * 8);
        int alloca_offset = 0;
        for (auto* bb : f->basic_blocks_) {
            lirbb = new LIRBasicBlock(bb->name_, lirf);
            bb_map_[bb] = lirbb;
            for (Instruction* inst : bb->instr_list_) {
                if (inst->op_id_ == Instruction::Alloca) {
                    auto* alloca_type = static_cast<PointerType*>(inst->type)->contained_;
                    auto* ai = static_cast<AllocaInst*>(inst);
                    int alloca_type_size = calc_type_size(alloca_type);
                    lirf->stackframe_allocas_.emplace_back(ai, alloca_type_size);
                    lirf->allocas_offset_[ai] = alloca_offset;
                    alloca_offset += alloca_type_size;
                } else if (inst->type->tid_ != Type::VoidTyID) {
                    if (inst->type->tid_ == Type::FloatTyID) {
                        auto* lirvfreg = m_lir->get_LIRVirtFReg();
                        inst_map_[inst] = lirvfreg;
                    } else {
                        auto* lirvreg = m_lir->get_LIRVirtReg();
                        inst_map_[inst] = lirvreg;
                    }
                }
            }
        }
        lirf->allocas_offset_end_ = alloca_offset;
        lirf->basic_blocks_[0]->add_succ_basic_block(lirf->basic_blocks_[1]);
        lirf->basic_blocks_[1]->add_pre_basic_block(lirf->basic_blocks_[0]);
        for (auto* bb : f->basic_blocks_) {
            for (auto* prebb : bb->pre_bbs_) {
                bb_map_[bb]->add_pre_basic_block(bb_map_.at(prebb));
            }
            for (auto* succbb : bb->succ_bbs_) {
                bb_map_[bb]->add_succ_basic_block(bb_map_.at(succbb));
            }
        }
    }
    for (auto* f : m_->function_list_) {
        if (f->is_declaration())
            continue;
        for (auto* bb : f->basic_blocks_) {
            set_insert_point(bb_map_.at(bb));
            for (std::list<Instruction*>::iterator iter3 = bb->instr_list_.begin(); iter3 != bb->instr_list_.end(); iter3++) {
                cinstr(iter3);
            }
        }
    }
    for (auto* f : m_->function_list_) {
        if (f->is_declaration())
            continue;
        for (auto* bb : f->basic_blocks_) {
            auto* lirbb = bb_map_.at(bb);
            for (std::list<Instruction*>::iterator iter3 = bb->instr_list_.begin(); iter3 != bb->instr_list_.end(); iter3++) {
                Instruction* inst = *iter3;
                if (inst->op_id_ == Instruction::PHI) {
                    auto* phi = dynamic_cast<PhiInst*>(inst);
                    if (inst->type->tid_ == Type::IntegerTyID) {
                        auto* phi_slot_reg = m_lir->get_LIRVirtReg();
                        auto* movinst0 = new LIRMovInst(LIRInstruction::LCondNone, get_LIROpnd(inst), phi_slot_reg);
                        lirbb->add_inst_front(movinst0);
                        std::set<BasicBlock*> prebbs(bb->pre_bbs_.begin(), bb->pre_bbs_.end());
                        for (int i = 0; i < phi->num_ops_ / 2; i++) {
                            auto* valFrom = phi->get_operand(2 * i);
                            auto* bbFrom = static_cast<BasicBlock*>(phi->get_operand(2 * i + 1));
                            auto* lirbbFrom = bb_map_.at(bbFrom);
                            prebbs.erase(bbFrom);
                            auto* movinst = new LIRMovInst(LIRInstruction::LCondNone, phi_slot_reg, get_LIROpnd(valFrom));
                            lirbbFrom->add_inst_before_br(movinst);
                        }
                    } else {
                       
                        auto* phi_slot_freg = m_lir->get_LIRVirtFReg();
                        auto* vmovf32inst0 = new LIRVmovf32Inst(LIRInstruction::LCondNone, get_LIROpnd(inst), phi_slot_freg);
                        lirbb->add_inst_front(vmovf32inst0);
                        std::set<BasicBlock*> prebbs(bb->pre_bbs_.begin(), bb->pre_bbs_.end());
                        for (int i = 0; i < phi->num_ops_ / 2; i++) {
                            auto* bbFrom = static_cast<BasicBlock*>(phi->get_operand(2 * i + 1));
                            auto* lirbbFrom = bb_map_.at(bbFrom);
                            prebbs.erase(bbFrom);
                            auto* valFrom = phi->get_operand(2 * i);
                            auto* lirvalFrom = get_LIROpnd(valFrom);
                            if (dynamic_cast<ConstantFloat*>(valFrom)) {
                                auto* tmpfreg = m_lir->get_LIRVirtFReg();
                                auto* tmpreg = m_lir->get_LIRVirtReg();
                                auto* movinst = new LIRMovInst(LIRInstruction::LCondNone, tmpreg, get_LIROpnd(valFrom));
                                lirbbFrom->add_inst_before_br(movinst);
                                auto* vmovinst = new LIRVmovInst(LIRInstruction::LCondNone, tmpfreg, tmpreg);
                                lirbbFrom->add_inst_before_br(vmovinst);
                                lirvalFrom = tmpfreg;
                            }
                            auto* vmovf32inst = new LIRVmovf32Inst(LIRInstruction::LCondNone, phi_slot_freg, lirvalFrom);
                            lirbbFrom->add_inst_before_br(vmovf32inst);
                        }
                    }
                }
            }
        }
    }
}

void CreateStmt::carm() {
    
    for (auto* lirf : m_lir->function_list_) {
        ra_ = new Register(lirf, is_o2_);
        ra_->run();
        delete ra_;
    }
   
    resolve_function_header();
    if (is_o2_) {
        remove_redundant_load_store();
    }
 
    resolve_imm();
    if (is_o2_) {
        resolve_libfunc();
        remove_useless_insts();
    }
}
void CreateStmt::combine_mul_addsub() {
    for (auto* lirf : m_lir->function_list_) {
        for (auto* lirbb : lirf->basic_blocks_) {
            std::map<LIRInstruction*, LIRInstruction*> replaceinst_map;
            for (auto iter = lirbb->instr_list_.begin(); iter != std::prev(lirbb->instr_list_.end()); iter++) {
                auto* lirinst = *iter;
                if (lirinst->op_id_ == LIRInstruction::LMulTy) {
                    auto* v0 = lirinst->result_;
                    auto* v1 = lirinst->get_operand(0);
                    auto* v2 = lirinst->get_operand(1);
                    auto* nextinst = *(std::next(iter));
                   auto* v3 = nextinst->result_;
                    if (nextinst->op_id_ == LIRInstruction::LAddTy) {
                        if(nextinst->get_operand(0) == v0){
                            auto* v4 = nextinst->get_operand(1);
                            if(v4->tid_==LIROperand::LOVirtRegTy||
                            v4->tid_==LIROperand::LOPhysRegTy){
                                replaceinst_map[lirinst] = new LIRIntInst(LIRInstruction::LMlaTy,LIRInstruction::LCondNone,v3,v1,v2,v4);
                                replaceinst_map[nextinst] = nullptr;
                            }
                        }else if(nextinst->get_operand(1) == v0){
                            auto* v4 = nextinst->get_operand(0);
                            if(v4->tid_==LIROperand::LOVirtRegTy||
                            v4->tid_==LIROperand::LOPhysRegTy){
                                replaceinst_map[lirinst] = new LIRIntInst(LIRInstruction::LMlaTy,LIRInstruction::LCondNone,v3,v1,v2,v4);
                                replaceinst_map[nextinst] = nullptr;
                            }
                        }
                    } else if (nextinst->op_id_ == LIRInstruction::LSubTy) {
                        if(nextinst->get_operand(1) == v0){
                            auto* v4 = nextinst->get_operand(0);
                            if(v4->tid_==LIROperand::LOVirtRegTy||
                            v4->tid_==LIROperand::LOPhysRegTy){
                                replaceinst_map[lirinst] = new LIRIntInst(LIRInstruction::LMlsTy,LIRInstruction::LCondNone,v3,v1,v2,v4);
                                replaceinst_map[nextinst] = nullptr;
                            }
                        }
                    } else if (nextinst->op_id_ == LIRInstruction::LRsbTy) {
                        if(nextinst->get_operand(0) == v0){
                            auto* v4 = nextinst->get_operand(1);
                            if(v4->tid_==LIROperand::LOVirtRegTy||
                            v4->tid_==LIROperand::LOPhysRegTy){
                                replaceinst_map[lirinst] = new LIRIntInst(LIRInstruction::LMlsTy,LIRInstruction::LCondNone,v3,v1,v2,v4);
                                replaceinst_map[nextinst] = nullptr;
                            }
                        }
                    }
                }else if (lirinst->op_id_ == LIRInstruction::LVmulTy) {
                    auto* v0 = lirinst->result_;
                    auto* v1 = lirinst->get_operand(0);
                    auto* v2 = lirinst->get_operand(1);
                   auto* nextinst = *(std::next(iter));
                    auto* v3 = nextinst->result_;
                    if (nextinst->op_id_ == LIRInstruction::LVaddTy) {
                        if(nextinst->get_operand(0) == v0 && v3==nextinst->get_operand(1)){
                            replaceinst_map[lirinst] = new LIRFloatInst(LIRInstruction::LVmlaTy,LIRInstruction::LCondNone,v3,v1,v2);
                            replaceinst_map[nextinst] = nullptr;
                        }else if(nextinst->get_operand(1) == v0 && v3==nextinst->get_operand(0)){
                            replaceinst_map[lirinst] = new LIRFloatInst(LIRInstruction::LVmlaTy,LIRInstruction::LCondNone,v3,v1,v2);
                            replaceinst_map[nextinst] = nullptr;
                        }
                    } else if (nextinst->op_id_ == LIRInstruction::LVsubTy) {
                        if(nextinst->get_operand(1) == v0 && v3==nextinst->get_operand(0)){
                            replaceinst_map[lirinst] = new LIRFloatInst(LIRInstruction::LVmlsTy,LIRInstruction::LCondNone,v3,v1,v2);
                            replaceinst_map[nextinst] = nullptr;
                        }
                    }
                }
            }
            for (const auto& pair : replaceinst_map) {
                if (pair.second == nullptr) {
                    lirbb->delete_instr(pair.first);
                } else {
                    lirbb->replace_instr(pair.first, pair.second);
                }
            }
        }
    }
}


void CreateStmt::resolve_libfunc() {
    for (auto* lirf : m_lir->function_list_) {
        for (auto* lirbb : lirf->basic_blocks_) {
            std::map<LIRInstruction*, std::pair<std::list<LIRInstruction*>, std::list<LIRInstruction*>>> addinst_map;
            for (auto* lirinst : lirbb->instr_list_) {
                std::list<LIRInstruction*> addbefore;
                std::list<LIRInstruction*> addafter;
                if (lirinst->op_id_ == LIRInstruction::LBlTy) {
                    auto* calllirf = static_cast<LIRFunction*>(lirinst->get_operand(0));
                    std::string callname = calllirf->name_;
                    
                    if (callname == "putint" || callname == "putch" || callname == "putarray" || callname == "_sysy_starttime" || callname == "_sysy_stoptime" || callname == "__aeabi_memclr4" || callname == "__aeabi_memset4" || callname == "__aeabi_memcpy4") {
                        std::vector<LIROperand*> reglist = {m_lir->get_LIRPhysReg(1), m_lir->get_LIRPhysReg(2), m_lir->get_LIRPhysReg(3), m_lir->get_LIRPhysReg(4)};
                        addbefore.push_front(new LIRPushInst(LIRInstruction::LCondNone, reglist));
                        addafter.push_back(new LIRPopInst(LIRInstruction::LCondNone, reglist));
                    } else if (callname == "getint" || callname == "getch" || callname == "getarray") {
                        std::vector<LIROperand*> reglist = {m_lir->get_LIRPhysReg(2), m_lir->get_LIRPhysReg(3), m_lir->get_LIRPhysReg(4), m_lir->get_LIRPhysReg(5)};
                        addbefore.push_front(new LIRPushInst(LIRInstruction::LCondNone, reglist));
                        addafter.push_back(new LIRPopInst(LIRInstruction::LCondNone, reglist));
                    } else if (callname == "getfloat") {
                        std::vector<LIROperand*> reglist = {m_lir->get_LIRPhysReg(1), m_lir->get_LIRPhysReg(2), m_lir->get_LIRPhysReg(3), m_lir->get_LIRPhysReg(4)};
                        std::vector<LIROperand*> freglist = {m_lir->get_LIRPhysFReg(-2), m_lir->get_LIRPhysFReg(-3), m_lir->get_LIRPhysFReg(-4), m_lir->get_LIRPhysFReg(-5), m_lir->get_LIRPhysFReg(-6), m_lir->get_LIRPhysFReg(-7), m_lir->get_LIRPhysFReg(-8), m_lir->get_LIRPhysFReg(-9), m_lir->get_LIRPhysFReg(-10), m_lir->get_LIRPhysFReg(-11), m_lir->get_LIRPhysFReg(-12), m_lir->get_LIRPhysFReg(-13), m_lir->get_LIRPhysFReg(-14), m_lir->get_LIRPhysFReg(-15), m_lir->get_LIRPhysFReg(-16), m_lir->get_LIRPhysFReg(-17)};
                        addbefore.push_front(new LIRVpushInst(LIRInstruction::LCondNone, freglist));
                        addbefore.push_front(new LIRPushInst(LIRInstruction::LCondNone, reglist));
                        addafter.push_back(new LIRVpopInst(LIRInstruction::LCondNone, freglist));
                        addafter.push_back(new LIRPopInst(LIRInstruction::LCondNone, reglist));
                    } else if (callname == "getfarray") {
                        std::vector<LIROperand*> reglist = {m_lir->get_LIRPhysReg(2), m_lir->get_LIRPhysReg(3), m_lir->get_LIRPhysReg(4), m_lir->get_LIRPhysReg(5)};
                        std::vector<LIROperand*> freglist = {m_lir->get_LIRPhysFReg(-1), m_lir->get_LIRPhysFReg(-2), m_lir->get_LIRPhysFReg(-3), m_lir->get_LIRPhysFReg(-4), m_lir->get_LIRPhysFReg(-5), m_lir->get_LIRPhysFReg(-6), m_lir->get_LIRPhysFReg(-7), m_lir->get_LIRPhysFReg(-8), m_lir->get_LIRPhysFReg(-9), m_lir->get_LIRPhysFReg(-10), m_lir->get_LIRPhysFReg(-11), m_lir->get_LIRPhysFReg(-12), m_lir->get_LIRPhysFReg(-13), m_lir->get_LIRPhysFReg(-14), m_lir->get_LIRPhysFReg(-15), m_lir->get_LIRPhysFReg(-16)};
                        addbefore.push_front(new LIRVpushInst(LIRInstruction::LCondNone, freglist));
                        addbefore.push_front(new LIRPushInst(LIRInstruction::LCondNone, reglist));
                        addafter.push_back(new LIRVpopInst(LIRInstruction::LCondNone, freglist));
                        addafter.push_back(new LIRPopInst(LIRInstruction::LCondNone, reglist));
                    } else if (callname == "putfloat" || callname == "putfarray") {
                        std::vector<LIROperand*> reglist = {m_lir->get_LIRPhysReg(1), m_lir->get_LIRPhysReg(2), m_lir->get_LIRPhysReg(3), m_lir->get_LIRPhysReg(4)};
                        std::vector<LIROperand*> freglist = {m_lir->get_LIRPhysFReg(-1), m_lir->get_LIRPhysFReg(-2), m_lir->get_LIRPhysFReg(-3), m_lir->get_LIRPhysFReg(-4), m_lir->get_LIRPhysFReg(-5), m_lir->get_LIRPhysFReg(-6), m_lir->get_LIRPhysFReg(-7), m_lir->get_LIRPhysFReg(-8), m_lir->get_LIRPhysFReg(-9), m_lir->get_LIRPhysFReg(-10), m_lir->get_LIRPhysFReg(-11), m_lir->get_LIRPhysFReg(-12), m_lir->get_LIRPhysFReg(-13), m_lir->get_LIRPhysFReg(-14), m_lir->get_LIRPhysFReg(-15), m_lir->get_LIRPhysFReg(-16)};
                        addbefore.push_front(new LIRVpushInst(LIRInstruction::LCondNone, freglist));
                        addbefore.push_front(new LIRPushInst(LIRInstruction::LCondNone, reglist));
                        addafter.push_back(new LIRVpopInst(LIRInstruction::LCondNone, freglist));
                        addafter.push_back(new LIRPopInst(LIRInstruction::LCondNone, reglist));
                    }
                }
                addinst_map[lirinst] = {addbefore, addafter};
            }
            for (const auto& pair : addinst_map) {
                for (auto inst : pair.second.first) {
                    lirbb->add_inst_before(inst, pair.first);
                }
                for (auto inst = pair.second.second.rbegin(); inst != pair.second.second.rend(); ++inst) {
                    lirbb->add_inst_after(*inst, pair.first);
                }
            }
        }
    }
}

void CreateStmt::return_operand(Value* operand) {  
    auto* operand_type = operand->type;  
    
    switch (operand_type->tid_) {  
        case Type::IntegerTyID:  
        case Type::PointerTyID:  
            create_mov(LIRInstruction::LCondNone, m_lir->get_LIRPhysReg(1), get_LIROpnd(operand));  
            break;  

        case Type::FloatTyID:  
            return_process_float(operand);  
            break;  

        default:  
            assert(0 && "Return instruction type invalid!");  
    }  
}  

void CreateStmt::return_process_float(Value* operand) {  
    if (dynamic_cast<ConstantFloat*>(operand)) {  
        auto* vr = m_lir->get_LIRVirtReg();  
        create_mov(LIRInstruction::LCondNone, vr, get_LIROpnd(operand));  
        create_vmov(LIRInstruction::LCondNone, m_lir->get_LIRPhysFReg(-1), vr);  
    } else {  
        auto* liropnd = get_LIROpnd(operand);  
        assert(liropnd->tid_ == LIROperand::LOVirtFRegTy);  
        create_vmovf32(LIRInstruction::LCondNone, m_lir->get_LIRPhysFReg(-1), liropnd);  
    }  
}  

void CreateStmt::cinstr(std::list<Instruction*>::iterator iter3) {
    auto* inst = *iter3;
    switch (inst->op_id_) {
          case Instruction::Ret: {  
            if (inst->num_ops_ > 0) {  
                auto* opnd0 = inst->get_operand(0);  
                return_operand(opnd0);  
            }  
            break;  
        }  
        
        case Instruction::Call: {
            Function* callf = static_cast<Function*>(inst->get_operand(inst->num_ops_ - 1));
            LIRFunction* calllirf = func_map_.at(callf);
            if (inst->num_ops_ >= 2) {
                std::vector<LIROperand*> arg_vec;
                for (int i = inst->num_ops_ - 2; i >= 0; i--) {
                    Value* opi = inst->get_operand(i);
                    LIROperand* destreg;
                    if (calllirf->args_passmethod_[i]) {
                        destreg = calllirf->args_passmethod_[i];
                    } else {
                        destreg = m_lir->get_LIRVirtReg();
                    }
                    if (auto gv = dynamic_cast<GlobalVariable*>(opi)) {
                        create_mov32(LIRInstruction::LCondNone, destreg, gv_map_.at(gv));
                    } else if (auto oi = dynamic_cast<Instruction*>(opi)) {
                        if (oi->op_id_ == Instruction::Alloca) {
                            int var_offset = BB_->parent_->allocas_offset_.at(static_cast<AllocaInst*>(oi));
                            if (var_offset == 0) {
                                create_mov(LIRInstruction::LCondNone, destreg, m_lir->get_LIRPhysReg(14));
                            } else {
                                create_lirinst(LIRInstruction::LAddTy, LIRInstruction::LCondNone, destreg, m_lir->get_LIRPhysReg(14), m_lir->get_LIRConstant(var_offset));
                            }
                        } else {
                            if (destreg->tid_ == LIROperand::LOPhysRegTy) {
                                create_mov(LIRInstruction::LCondNone, destreg, get_LIROpnd(opi));
                            } else if (destreg->tid_ == LIROperand::LOPhysFRegTy) {
                                create_vmovf32(LIRInstruction::LCondNone, destreg, get_LIROpnd(opi));
                            } else {
                               
                                LIROperand* liropi = get_LIROpnd(opi);
                                if (liropi->tid_ == LIROperand::LOVirtRegTy) {
                                    create_mov(LIRInstruction::LCondNone, destreg, liropi);
                                } else {
                                 
                                    create_vmov(LIRInstruction::LCondNone, destreg, liropi);
                                }
                            }
                        }
                    } else if (dynamic_cast<Argument*>(opi)) {
                        if (destreg->tid_ == LIROperand::LOPhysRegTy) {
                            create_mov(LIRInstruction::LCondNone, destreg, get_LIROpnd(opi));
                        } else if (destreg->tid_ == LIROperand::LOPhysFRegTy) {
                            create_vmovf32(LIRInstruction::LCondNone, destreg, get_LIROpnd(opi));
                        } else {
                           
                            LIROperand* liropi = get_LIROpnd(opi);
                            if (liropi->tid_ == LIROperand::LOVirtRegTy) {
                                create_mov(LIRInstruction::LCondNone, destreg, liropi);
                            } else {
                             
                                create_vmov(LIRInstruction::LCondNone, destreg, liropi);
                            }
                        }
                    } else {
                       
                        if (destreg->tid_ == LIROperand::LOPhysRegTy || destreg->tid_ == LIROperand::LOVirtRegTy) {
                            create_mov(LIRInstruction::LCondNone, destreg, get_LIROpnd(opi));
                        } else {
                          
                            LIRVirtReg* tmpreg = m_lir->get_LIRVirtReg();
                            create_mov(LIRInstruction::LCondNone, tmpreg, get_LIROpnd(opi));
                            create_vmov(LIRInstruction::LCondNone, destreg, tmpreg);
                        }
                    }
                    if (!calllirf->args_passmethod_[i]) {
                        arg_vec.push_back(destreg);
                    }
                }
                if (!arg_vec.empty()) {
                    int curoffset = (arg_vec.size() % 2 == 0) ? (-4) : (-8);
                    for (int i = 0; i < arg_vec.size(); i++) {
                        create_str(LIRInstruction::LCondNone, arg_vec[i], m_lir->get_LIRPhysReg(14), m_lir->get_LIRConstant(curoffset), LIRInstruction::LRegisterOffsetTy);
                        curoffset -= 4;
                    }
                    create_lirinst(LIRInstruction::LSubTy, LIRInstruction::LCondNone, m_lir->get_LIRPhysReg(14), m_lir->get_LIRPhysReg(14), m_lir->get_LIRConstant(calllirf->stackframe_args_size_));
                }
            }
            bool res_unused=((inst->use_list_).empty())?(true):(false);
            create_bl(LIRInstruction::LCondNone, calllirf,res_unused);
            if (calllirf->stackframe_args_size_ > 0) {
                create_lirinst(LIRInstruction::LAddTy, LIRInstruction::LCondNone, m_lir->get_LIRPhysReg(14), m_lir->get_LIRPhysReg(14), m_lir->get_LIRConstant(calllirf->stackframe_args_size_));
            }
            if (calllirf->has_result_ && !(inst->use_list_).empty()) {
                LIROperand* lirresult = get_LIROpnd(inst);
                if (lirresult->tid_ == LIROperand::LOVirtRegTy) {
                    create_mov(LIRInstruction::LCondNone, lirresult, m_lir->get_LIRPhysReg(1));
                } else {
                   
                    create_vmovf32(LIRInstruction::LCondNone, lirresult, m_lir->get_LIRPhysFReg(-1));
                }
            }
            break;
        }
        case Instruction::GetElementPtr: {
            Value* op0 = inst->get_operand(0);
            LIROperand* destreg = get_LIROpnd(inst);
          
            if (auto gv = dynamic_cast<GlobalVariable*>(op0)) {
                create_mov32(LIRInstruction::LCondNone, destreg, gv_map_.at(gv));
            } else if (auto oi = dynamic_cast<Instruction*>(op0)) {
                if (oi->op_id_ == Instruction::Alloca) {
                    int var_offset = BB_->parent_->allocas_offset_.at(static_cast<AllocaInst*>(oi));
                    if (var_offset == 0) {
                        create_mov(LIRInstruction::LCondNone, destreg, m_lir->get_LIRPhysReg(14));
                    } else {
                        create_lirinst(LIRInstruction::LAddTy, LIRInstruction::LCondNone, destreg, m_lir->get_LIRPhysReg(14), m_lir->get_LIRConstant(var_offset));
                    }
                } else {
                    create_mov(LIRInstruction::LCondNone, destreg, get_LIROpnd(op0));
                }
            } else {
              
                create_mov(LIRInstruction::LCondNone, destreg, get_LIROpnd(op0));
            }
            int cur_type_size = 0;
            unsigned int num_operands = inst->num_ops_;
            int index_val;
            int total_offset = 0;
            Type* cur_type = static_cast<PointerType*>(inst->get_operand(0)->type)->contained_;
            for (unsigned int i = 1; i <= num_operands - 1; i++) {
                if (i > 1) {
                    cur_type = static_cast<ArrayType*>(cur_type)->contained_;
                }
                Value* opi = inst->get_operand(i);
                cur_type_size = calc_type_size(cur_type);
                if (auto ci = dynamic_cast<ConstantInt*>(opi)) {
                    index_val = ci->value_;
                    total_offset += index_val * cur_type_size;
                } else {
                  
                    LIRVirtReg* tmpreg = m_lir->get_LIRVirtReg();
                    create_mov(LIRInstruction::LCondNone, tmpreg, m_lir->get_LIRConstant(cur_type_size));
                    create_lirinst(LIRInstruction::LMlaTy, LIRInstruction::LCondNone, destreg, get_LIROpnd(opi), tmpreg, destreg);
                }
            }
            if (total_offset > 0) {
                create_lirinst(LIRInstruction::LAddTy, LIRInstruction::LCondNone, destreg, destreg, m_lir->get_LIRConstant(total_offset));
            }
            break;
        }
        case Instruction::Load: {
            if (inst->type->tid_ == Type::IntegerTyID || inst->type->tid_ == Type::PointerTyID) {
               
                Value* opnd = inst->get_operand(0);
                if (auto gv = dynamic_cast<GlobalVariable*>(opnd)) {
                    LIRVirtReg* tmpreg = m_lir->get_LIRVirtReg();
                    create_mov32(LIRInstruction::LCondNone, tmpreg, gv_map_.at(gv));
                    create_ldr(LIRInstruction::LCondNone, get_LIROpnd(inst), tmpreg, m_lir->get_LIRConstant(0), LIRInstruction::LRegisterOffsetTy);
                } else if (auto oi = dynamic_cast<Instruction*>(opnd)) {
                    if (oi->op_id_ == Instruction::Alloca) {
                        create_ldr(LIRInstruction::LCondNone, get_LIROpnd(inst), m_lir->get_LIRPhysReg(14), m_lir->get_LIRConstant(BB_->parent_->allocas_offset_.at(static_cast<AllocaInst*>(oi))), LIRInstruction::LRegisterOffsetTy);
                    } else {
                        create_ldr(LIRInstruction::LCondNone, get_LIROpnd(inst), get_LIROpnd(oi), m_lir->get_LIRConstant(0), LIRInstruction::LRegisterOffsetTy);
                    }
                } else if (auto oa = dynamic_cast<Argument*>(opnd)) {
                    create_ldr(LIRInstruction::LCondNone, get_LIROpnd(inst), get_LIROpnd(oa), m_lir->get_LIRConstant(0), LIRInstruction::LRegisterOffsetTy);
                } else {
                    assert(0 && "load_opnd instruction opnd type invalid!");
                }
                
            } else if (inst->type->tid_ == Type::FloatTyID) {
                Value* opnd = inst->get_operand(0);
                if (auto gv = dynamic_cast<GlobalVariable*>(opnd)) {
                    LIRVirtReg* tmpreg = m_lir->get_LIRVirtReg();
                    create_mov32(LIRInstruction::LCondNone, tmpreg, gv_map_.at(gv));
                    create_vldr(LIRInstruction::LCondNone, get_LIROpnd(inst), tmpreg, m_lir->get_LIRConstant(0));
                } else if (auto oi = dynamic_cast<Instruction*>(opnd)) {
                    if (oi->op_id_ == Instruction::Alloca) {
                        create_vldr(LIRInstruction::LCondNone, get_LIROpnd(inst), m_lir->get_LIRPhysReg(14), m_lir->get_LIRConstant(BB_->parent_->allocas_offset_.at(static_cast<AllocaInst*>(oi))));
                    } else {
                        create_vldr(LIRInstruction::LCondNone, get_LIROpnd(inst), get_LIROpnd(oi), m_lir->get_LIRConstant(0));
                    }
                } else if (auto oa = dynamic_cast<Argument*>(opnd)) {
                    create_vldr(LIRInstruction::LCondNone, get_LIROpnd(inst), get_LIROpnd(oa), m_lir->get_LIRConstant(0));
                }
            } 
            break;
        }
        case Instruction::Store: {
            Value* op0 = inst->get_operand(0);
            LIROperand* lirop0 = get_LIROpnd(op0);
            if (dynamic_cast<Constant*>(op0)) {
               
                LIRVirtReg* tmpreg = m_lir->get_LIRVirtReg();
                create_mov(LIRInstruction::LCondNone, tmpreg, get_LIROpnd(op0));
                lirop0 = tmpreg;
            }
            Value* opnd = inst->get_operand(1);
            if (auto gv = dynamic_cast<GlobalVariable*>(opnd)) {
                LIRVirtReg* tmpreg = m_lir->get_LIRVirtReg();
                create_mov32(LIRInstruction::LCondNone, tmpreg, gv_map_.at(gv));
                if (lirop0->tid_ == LIROperand::LOVirtRegTy) {
                    create_str(LIRInstruction::LCondNone, lirop0, tmpreg, m_lir->get_LIRConstant(0), LIRInstruction::LRegisterOffsetTy);
                } else {
                   
                    create_vstr(LIRInstruction::LCondNone, lirop0, tmpreg, m_lir->get_LIRConstant(0));
                }
            } else if (auto oi = dynamic_cast<Instruction*>(opnd)) {
                if (oi->op_id_ == Instruction::Alloca) {
                    if (lirop0->tid_ == LIROperand::LOVirtRegTy) {
                        create_str(LIRInstruction::LCondNone, lirop0, m_lir->get_LIRPhysReg(14), m_lir->get_LIRConstant(BB_->parent_->allocas_offset_.at(static_cast<AllocaInst*>(oi))), LIRInstruction::LRegisterOffsetTy);
                    } else {
                       
                        create_vstr(LIRInstruction::LCondNone, lirop0, m_lir->get_LIRPhysReg(14), m_lir->get_LIRConstant(BB_->parent_->allocas_offset_.at(static_cast<AllocaInst*>(oi))));
                    }
                } else {
                    if (lirop0->tid_ == LIROperand::LOVirtRegTy) {
                        create_str(LIRInstruction::LCondNone, lirop0, get_LIROpnd(oi), m_lir->get_LIRConstant(0), LIRInstruction::LRegisterOffsetTy);
                    } else {
                     
                        create_vstr(LIRInstruction::LCondNone, lirop0, get_LIROpnd(oi), m_lir->get_LIRConstant(0));
                    }
                }
            } else if (auto oa = dynamic_cast<Argument*>(opnd)) {
                if (lirop0->tid_ == LIROperand::LOVirtRegTy) {
                    create_str(LIRInstruction::LCondNone, lirop0, get_LIROpnd(oa), m_lir->get_LIRConstant(0), LIRInstruction::LRegisterOffsetTy);
                } else {
                  
                    create_vstr(LIRInstruction::LCondNone, lirop0, get_LIROpnd(oa), m_lir->get_LIRConstant(0));
                }
            } 
            break;
        }
        case Instruction::Alloca: {
            break;
        }
        case Instruction::Br: {
            Instruction* previnst = *(std::prev(iter3, 1));
            if (inst->num_ops_ == 1) {
                create_b(LIRInstruction::LCondNone, bb_map_.at(static_cast<BasicBlock*>(inst->get_operand(0))));
            } else {
                assert(inst->num_ops_ == 3);
                Value* op0 = inst->get_operand(0);
                if (auto ci = dynamic_cast<ConstantInt*>(op0)) {
                    if (ci->value_) {
                        create_b(LIRInstruction::LCondNone, bb_map_.at(static_cast<BasicBlock*>(inst->get_operand(1))));
                    } else {
                        create_b(LIRInstruction::LCondNone, bb_map_.at(static_cast<BasicBlock*>(inst->get_operand(2))));
                    }
                } else if (auto op0inst = dynamic_cast<Instruction*>(op0)) {
                    if ((op0inst->op_id_ == Instruction::ICmp || op0inst->op_id_ == Instruction::FCmp) && op0inst == previnst && op0inst->use_list_.size() == 1) {
                    } else {
                        create_cmp(LIRInstruction::LCondNone, get_LIROpnd(op0), m_lir->get_LIRConstant(0));
                        create_b(LIRInstruction::LCondNE, bb_map_.at(static_cast<BasicBlock*>(inst->get_operand(1))));
                        create_b(LIRInstruction::LCondNone, bb_map_.at(static_cast<BasicBlock*>(inst->get_operand(2))));
                    }
                } 
            }
            break;
        }
        case Instruction::ICmp: {
            Instruction* nextinst = *(std::next(iter3, 1));
            auto icmpi = dynamic_cast<ICmpInst*>(inst);
            Value* icmpiop0 = icmpi->get_operand(0);
            Value* icmpiop1 = icmpi->get_operand(1);
            bool is_op_inverse;
            if (dynamic_cast<ConstantInt*>(icmpiop0)) {
                if (dynamic_cast<ConstantInt*>(icmpiop1)) {
                    LIRVirtReg* tmpreg = m_lir->get_LIRVirtReg();
                    create_mov(LIRInstruction::LCondNone, tmpreg, get_LIROpnd(icmpiop0));
                    create_cmp(LIRInstruction::LCondNone, tmpreg, get_LIROpnd(icmpiop1));
                    is_op_inverse = false;
                } else {
                    create_cmp(LIRInstruction::LCondNone, get_LIROpnd(icmpiop1), get_LIROpnd(icmpiop0));
                    is_op_inverse = true;
                }
            } else {
                create_cmp(LIRInstruction::LCondNone, get_LIROpnd(icmpiop0), get_LIROpnd(icmpiop1));
                is_op_inverse = false;
            }
            LIRInstruction::LIRCondTy brcondty = is_op_inverse ? Ircond(Icmpcond(icmpi->icmp_op_)) : Icmpcond(icmpi->icmp_op_);
            if (inst->use_list_.size() == 1 && nextinst->op_id_ == Instruction::Br && nextinst->operands_[0] == inst) {
                create_b(brcondty, bb_map_.at(static_cast<BasicBlock*>(nextinst->get_operand(1))));
                create_b(LIRInstruction::LCondNone, bb_map_.at(static_cast<BasicBlock*>(nextinst->get_operand(2))));
            } else {
                create_mov(LIRInstruction::LCondNone, get_LIROpnd(inst), m_lir->get_LIRConstant(0));
                create_mov(brcondty, get_LIROpnd(inst), m_lir->get_LIRConstant(1));
            }
            break;
        }
        case Instruction::FCmp: {
            Instruction* nextinst = *(std::next(iter3, 1));
            auto fcmpi = dynamic_cast<FCmpInst*>(inst);
            Value* fcmpiop0 = fcmpi->get_operand(0);
            Value* fcmpiop1 = fcmpi->get_operand(1);
            LIROperand* lirfcmpiop0 = get_LIROpnd(fcmpiop0);
            LIROperand* lirfcmpiop1 = get_LIROpnd(fcmpiop1);
            bool is_op_inverse;
            if (lirfcmpiop0->tid_ == LIROperand::LOConstantTy && static_cast<LIRConstant*>(lirfcmpiop0)->value_ == 0 && lirfcmpiop1->tid_ == LIROperand::LOConstantTy && static_cast<LIRConstant*>(lirfcmpiop1)->value_ == 0) {
                is_op_inverse = false;
                LIRVirtFReg* tmpfreg = m_lir->get_LIRVirtFReg();
                LIRVirtReg* tmpreg = m_lir->get_LIRVirtReg();
                create_mov(LIRInstruction::LCondNone, tmpreg, lirfcmpiop0);
                create_vmov(LIRInstruction::LCondNone, tmpfreg, tmpreg);
                create_vcmp(LIRInstruction::LCondNone, tmpfreg, lirfcmpiop1);
            } else if (lirfcmpiop0->tid_ == LIROperand::LOConstantTy && static_cast<LIRConstant*>(lirfcmpiop0)->value_ == 0 && lirfcmpiop1->tid_ == LIROperand::LOVirtFRegTy) {
                is_op_inverse = true;
                create_vcmp(LIRInstruction::LCondNone, lirfcmpiop1, lirfcmpiop0);
            } else if (lirfcmpiop0->tid_ == LIROperand::LOVirtFRegTy && lirfcmpiop1->tid_ == LIROperand::LOConstantTy && static_cast<LIRConstant*>(lirfcmpiop1)->value_ == 0) {
                is_op_inverse = false;
                create_vcmp(LIRInstruction::LCondNone, lirfcmpiop0, lirfcmpiop1);
            } else {
                is_op_inverse = false;
                if (dynamic_cast<ConstantFloat*>(fcmpiop0)) {
                    LIRVirtFReg* tmpfreg = m_lir->get_LIRVirtFReg();
                    LIRVirtReg* tmpreg = m_lir->get_LIRVirtReg();
                    create_mov(LIRInstruction::LCondNone, tmpreg, lirfcmpiop0);
                    create_vmov(LIRInstruction::LCondNone, tmpfreg, tmpreg);
                    lirfcmpiop0 = tmpfreg;
                }
                if (dynamic_cast<ConstantFloat*>(fcmpiop1)) {
                    LIRVirtFReg* tmpfreg = m_lir->get_LIRVirtFReg();
                    LIRVirtReg* tmpreg = m_lir->get_LIRVirtReg();
                    create_mov(LIRInstruction::LCondNone, tmpreg, lirfcmpiop1);
                    create_vmov(LIRInstruction::LCondNone, tmpfreg, tmpreg);
                    lirfcmpiop1 = tmpfreg;
                }
                create_vcmp(LIRInstruction::LCondNone, lirfcmpiop0, lirfcmpiop1);
            }
            create_vmrs();
            LIRInstruction::LIRCondTy brcondty = is_op_inverse ? Ircond(Fcmpcond(fcmpi->fcmp_op_)) : Fcmpcond(fcmpi->fcmp_op_);
            if (inst->use_list_.size() == 1 && nextinst->op_id_ == Instruction::Br && nextinst->operands_[0] == inst) {
                create_b(brcondty, bb_map_.at(static_cast<BasicBlock*>(nextinst->get_operand(1))));
                create_b(LIRInstruction::LCondNone, bb_map_.at(static_cast<BasicBlock*>(nextinst->get_operand(2))));
            } else {
                create_mov(LIRInstruction::LCondNone, get_LIROpnd(inst), m_lir->get_LIRConstant(0));
                create_mov(brcondty, get_LIROpnd(inst), m_lir->get_LIRConstant(1));
            }
            break;
        }
        case Instruction::Add: {
            Value* op0 = inst->get_operand(0);
            Value* op1 = inst->get_operand(1);
            bool is_op0_ci, is_op1_ci;
            ConstantInt *ciop0, *ciop1;
            is_op0_ci = (ciop0 = dynamic_cast<ConstantInt*>(op0)) ? true : false;
            is_op1_ci = (ciop1 = dynamic_cast<ConstantInt*>(op1)) ? true : false;
            if (is_op0_ci && is_op1_ci) {
                int result = ciop0->value_ + ciop1->value_;
                create_mov(LIRInstruction::LCondNone, get_LIROpnd(inst), m_lir->get_LIRConstant(result));
            } else if (is_op0_ci && !is_op1_ci) {
                if (ciop0->value_ == 0) {
                    create_mov(LIRInstruction::LCondNone, get_LIROpnd(inst), get_LIROpnd(op1));
                } else {
                    create_lirinst(LIRInstruction::LAddTy, LIRInstruction::LCondNone, get_LIROpnd(inst), get_LIROpnd(op1), get_LIROpnd(op0));
                }
            } else if (!is_op0_ci && is_op1_ci) {
                if (ciop1->value_ == 0) {
                    create_mov(LIRInstruction::LCondNone, get_LIROpnd(inst), get_LIROpnd(op0));
                } else {
                    create_lirinst(LIRInstruction::LAddTy, LIRInstruction::LCondNone, get_LIROpnd(inst), get_LIROpnd(op0), get_LIROpnd(op1));
                }
            } else {
                create_lirinst(LIRInstruction::LAddTy, LIRInstruction::LCondNone, get_LIROpnd(inst), get_LIROpnd(op0), get_LIROpnd(op1));
            }
            break;
        }
        case Instruction::Sub: {
            Value* op0 = inst->get_operand(0);
            Value* op1 = inst->get_operand(1);
            bool is_op0_ci, is_op1_ci;
            ConstantInt *ciop0, *ciop1;
            is_op0_ci = (ciop0 = dynamic_cast<ConstantInt*>(op0)) ? true : false;
            is_op1_ci = (ciop1 = dynamic_cast<ConstantInt*>(op1)) ? true : false;
            if (is_op0_ci && is_op1_ci) {
                int result = ciop0->value_ - ciop1->value_;
                create_mov(LIRInstruction::LCondNone, get_LIROpnd(inst), m_lir->get_LIRConstant(result));
            } else if (is_op0_ci && !is_op1_ci) {
                create_lirinst(LIRInstruction::LRsbTy, LIRInstruction::LCondNone, get_LIROpnd(inst), get_LIROpnd(op1), get_LIROpnd(op0));
            } else if (!is_op0_ci && is_op1_ci) {
                if (ciop1->value_ == 0) {
                    create_mov(LIRInstruction::LCondNone, get_LIROpnd(inst), get_LIROpnd(op0));
                } else {
                    create_lirinst(LIRInstruction::LSubTy, LIRInstruction::LCondNone, get_LIROpnd(inst), get_LIROpnd(op0), get_LIROpnd(op1));
                }
            } else {
                create_lirinst(LIRInstruction::LSubTy, LIRInstruction::LCondNone, get_LIROpnd(inst), get_LIROpnd(op0), get_LIROpnd(op1));
            }
            break;
        }
        case Instruction::Mul: {
            Value* op0 = inst->get_operand(0);
            LIROperand* lirop0 = get_LIROpnd(op0);
            Value* op1 = inst->get_operand(1);
            LIROperand* lirop1 = get_LIROpnd(op1);
            if (dynamic_cast<ConstantInt*>(op0)) {
                LIRVirtReg* tmpreg = m_lir->get_LIRVirtReg();
                create_mov(LIRInstruction::LCondNone, tmpreg, get_LIROpnd(op0));
                lirop0 = tmpreg;
            }
            if (dynamic_cast<ConstantInt*>(op1)) {
                LIRVirtReg* tmpreg = m_lir->get_LIRVirtReg();
                create_mov(LIRInstruction::LCondNone, tmpreg, get_LIROpnd(op1));
                lirop1 = tmpreg;
            }
            create_lirinst(LIRInstruction::LMulTy, LIRInstruction::LCondNone, get_LIROpnd(inst), lirop0, lirop1);
            break;
        }
        case Instruction::SDiv: {
            Value* op0 = inst->get_operand(0);
            LIROperand* lirop0 = get_LIROpnd(op0);
            Value* op1 = inst->get_operand(1);
            LIROperand* lirop1 = get_LIROpnd(op1);
            if (dynamic_cast<ConstantInt*>(op0)) {
                LIRVirtReg* tmpreg = m_lir->get_LIRVirtReg();
                create_mov(LIRInstruction::LCondNone, tmpreg, get_LIROpnd(op0));
                lirop0 = tmpreg;
            }
            if (dynamic_cast<ConstantInt*>(op1)) {
                LIRVirtReg* tmpreg = m_lir->get_LIRVirtReg();
                create_mov(LIRInstruction::LCondNone, tmpreg, get_LIROpnd(op1));
                lirop1 = tmpreg;
            }
            create_lirinst(LIRInstruction::LSdivTy,LIRInstruction::LCondNone, get_LIROpnd(inst), lirop0, lirop1);
            break;
        }
        case Instruction::SRem: {
            Value* op0 = inst->get_operand(0);
            LIROperand* lirop0 = get_LIROpnd(op0);
            Value* op1 = inst->get_operand(1);
            LIROperand* lirop1 = get_LIROpnd(op1);
            if (dynamic_cast<ConstantInt*>(op0)) {
                LIRVirtReg* tmpreg = m_lir->get_LIRVirtReg();
                create_mov(LIRInstruction::LCondNone, tmpreg, get_LIROpnd(op0));
                lirop0 = tmpreg;
            }
            if (dynamic_cast<ConstantInt*>(op1)) {
                LIRVirtReg* tmpreg = m_lir->get_LIRVirtReg();
                create_mov(LIRInstruction::LCondNone, tmpreg, get_LIROpnd(op1));
                lirop1 = tmpreg;
            }
            create_lirinst(LIRInstruction::LSdivTy, LIRInstruction::LCondNone, get_LIROpnd(inst), lirop0, lirop1);
            create_lirinst(LIRInstruction::LMlsTy,  LIRInstruction::LCondNone, get_LIROpnd(inst), get_LIROpnd(inst), lirop1, lirop0);
            break;
        }
        case Instruction::FAdd: {
            Value* op0 = inst->get_operand(0);
            LIROperand* lirop0 = get_LIROpnd(op0);
            Value* op1 = inst->get_operand(1);
            LIROperand* lirop1 = get_LIROpnd(op1);
            if (dynamic_cast<ConstantFloat*>(op0)) {
                LIRVirtFReg* tmpfreg = m_lir->get_LIRVirtFReg();
                LIRVirtReg* tmpreg = m_lir->get_LIRVirtReg();
                create_mov(LIRInstruction::LCondNone, tmpreg, get_LIROpnd(op0));
                create_vmov(LIRInstruction::LCondNone, tmpfreg, tmpreg);
                lirop0 = tmpfreg;
            }
            if (dynamic_cast<ConstantFloat*>(op1)) {
                LIRVirtFReg* tmpfreg = m_lir->get_LIRVirtFReg();
                LIRVirtReg* tmpreg = m_lir->get_LIRVirtReg();
                create_mov(LIRInstruction::LCondNone, tmpreg, get_LIROpnd(op1));
                create_vmov(LIRInstruction::LCondNone, tmpfreg, tmpreg);
                lirop1 = tmpfreg;
            }
            create_lirfinst(LIRInstruction::LVaddTy, LIRInstruction::LCondNone, get_LIROpnd(inst), lirop0, lirop1);
            break;
        }
        case Instruction::FSub: {
            Value* op0 = inst->get_operand(0);
            LIROperand* lirop0 = get_LIROpnd(op0);
            Value* op1 = inst->get_operand(1);
            LIROperand* lirop1 = get_LIROpnd(op1);
            if (dynamic_cast<ConstantFloat*>(op0)) {
                LIRVirtFReg* tmpfreg = m_lir->get_LIRVirtFReg();
                LIRVirtReg* tmpreg = m_lir->get_LIRVirtReg();
                create_mov(LIRInstruction::LCondNone, tmpreg, get_LIROpnd(op0));
                create_vmov(LIRInstruction::LCondNone, tmpfreg, tmpreg);
                lirop0 = tmpfreg;
            }
            if (dynamic_cast<ConstantFloat*>(op1)) {
                LIRVirtFReg* tmpfreg = m_lir->get_LIRVirtFReg();
                LIRVirtReg* tmpreg = m_lir->get_LIRVirtReg();
                create_mov(LIRInstruction::LCondNone, tmpreg, get_LIROpnd(op1));
                create_vmov(LIRInstruction::LCondNone, tmpfreg, tmpreg);
                lirop1 = tmpfreg;
            }
            create_lirfinst(LIRInstruction::LVsubTy, LIRInstruction::LCondNone, get_LIROpnd(inst), lirop0, lirop1);
            break;
        }
        case Instruction::FMul: {
            Value* op0 = inst->get_operand(0);
            LIROperand* lirop0 = get_LIROpnd(op0);
            Value* op1 = inst->get_operand(1);
            LIROperand* lirop1 = get_LIROpnd(op1);
            if (dynamic_cast<ConstantFloat*>(op0)) {
                LIRVirtFReg* tmpfreg = m_lir->get_LIRVirtFReg();
                LIRVirtReg* tmpreg = m_lir->get_LIRVirtReg();
                create_mov(LIRInstruction::LCondNone, tmpreg, get_LIROpnd(op0));
                create_vmov(LIRInstruction::LCondNone, tmpfreg, tmpreg);
                lirop0 = tmpfreg;
            }
            if (dynamic_cast<ConstantFloat*>(op1)) {
                LIRVirtFReg* tmpfreg = m_lir->get_LIRVirtFReg();
                LIRVirtReg* tmpreg = m_lir->get_LIRVirtReg();
                create_mov(LIRInstruction::LCondNone, tmpreg, get_LIROpnd(op1));
                create_vmov(LIRInstruction::LCondNone, tmpfreg, tmpreg);
                lirop1 = tmpfreg;
            }
            create_lirfinst(LIRInstruction::LVmulTy, LIRInstruction::LCondNone, get_LIROpnd(inst), lirop0, lirop1);
            break;
        }
        case Instruction::FDiv: {
            Value* op0 = inst->get_operand(0);
            LIROperand* lirop0 = get_LIROpnd(op0);
            Value* op1 = inst->get_operand(1);
            LIROperand* lirop1 = get_LIROpnd(op1);
            if (dynamic_cast<ConstantFloat*>(op0)) {
                LIRVirtFReg* tmpfreg = m_lir->get_LIRVirtFReg();
                LIRVirtReg* tmpreg = m_lir->get_LIRVirtReg();
                create_mov(LIRInstruction::LCondNone, tmpreg, get_LIROpnd(op0));
                create_vmov(LIRInstruction::LCondNone, tmpfreg, tmpreg);
                lirop0 = tmpfreg;
            }
            if (dynamic_cast<ConstantFloat*>(op1)) {
                LIRVirtFReg* tmpfreg = m_lir->get_LIRVirtFReg();
                LIRVirtReg* tmpreg = m_lir->get_LIRVirtReg();
                create_mov(LIRInstruction::LCondNone, tmpreg, get_LIROpnd(op1));
                create_vmov(LIRInstruction::LCondNone, tmpfreg, tmpreg);
                lirop1 = tmpfreg;
            }
            create_lirfinst(LIRInstruction::LVdivTy, LIRInstruction::LCondNone, get_LIROpnd(inst), lirop0, lirop1);
            break;
        }
        case Instruction::FNeg: {
            Value* op0 = inst->get_operand(0);
            LIROperand* lirop0 = get_LIROpnd(op0);
            if (dynamic_cast<ConstantFloat*>(op0)) {
                LIRVirtFReg* tmpfreg = m_lir->get_LIRVirtFReg();
                LIRVirtReg* tmpreg = m_lir->get_LIRVirtReg();
                create_mov(LIRInstruction::LCondNone, tmpreg, get_LIROpnd(op0));
                create_vmov(LIRInstruction::LCondNone, tmpfreg, tmpreg);
                lirop0 = tmpfreg;
            }
            create_lirfinst(LIRInstruction::LVnegTy, LIRInstruction::LCondNone, get_LIROpnd(inst), lirop0);
            break;
        }
        case Instruction::SItoFP: {
            Value* op0 = inst->get_operand(0);
            LIROperand* lirop0 = get_LIROpnd(op0);
            if (dynamic_cast<ConstantInt*>(op0)) {
                LIRVirtReg* tmpreg = m_lir->get_LIRVirtReg();
                create_mov(LIRInstruction::LCondNone, tmpreg, get_LIROpnd(op0));
                lirop0 = tmpreg;
            }
            create_vmov(LIRInstruction::LCondNone, get_LIROpnd(inst), lirop0);
            create_lirfinst(LIRInstruction::LVcvtFSTy, LIRInstruction::LCondNone, get_LIROpnd(inst), get_LIROpnd(inst));
            break;
        }
        case Instruction::FPtoSI: {
            Value* op0 = inst->get_operand(0);
            LIROperand* lirop0 = get_LIROpnd(op0);
            if (dynamic_cast<ConstantFloat*>(op0)) {
                LIRVirtFReg* tmpfreg = m_lir->get_LIRVirtFReg();
                LIRVirtReg* tmpreg = m_lir->get_LIRVirtReg();
                create_mov(LIRInstruction::LCondNone, tmpreg, lirop0);
                create_vmov(LIRInstruction::LCondNone, tmpfreg, tmpreg);
                lirop0 = tmpfreg;
            }
            LIRVirtFReg* tmpfreg2 = m_lir->get_LIRVirtFReg();
            create_lirfinst(LIRInstruction::LVcvtSFTy, LIRInstruction::LCondNone, tmpfreg2, lirop0);
            create_vmov(LIRInstruction::LCondNone, get_LIROpnd(inst), tmpfreg2);
            break;
        }
        case Instruction::ZExt: {
            Value* op0 = inst->get_operand(0);
            create_mov(LIRInstruction::LCondNone, get_LIROpnd(inst), get_LIROpnd(op0));
            break;
        }
        case Instruction::BitCast: {
            Value* op0 = inst->get_operand(0);
            LIROperand* destreg = get_LIROpnd(inst);
          
            if (auto gv = dynamic_cast<GlobalVariable*>(op0)) {
                create_mov32(LIRInstruction::LCondNone, destreg, gv_map_.at(gv));
            } else if (auto oi = dynamic_cast<Instruction*>(op0)) {
                if (oi->op_id_ == Instruction::Alloca) {
                    int var_offset = BB_->parent_->allocas_offset_.at(static_cast<AllocaInst*>(oi));
                    if (var_offset == 0) {
                        create_mov(LIRInstruction::LCondNone, destreg, m_lir->get_LIRPhysReg(14));
                    } else {
                        create_lirinst(LIRInstruction::LAddTy, LIRInstruction::LCondNone, destreg, m_lir->get_LIRPhysReg(14), m_lir->get_LIRConstant(var_offset));
                    }
                } else {
                    create_mov(LIRInstruction::LCondNone, destreg, get_LIROpnd(op0));
                }
            } else {
               
                create_mov(LIRInstruction::LCondNone, destreg, get_LIROpnd(op0));
            }
            break;
        }
        case Instruction::UDiv:
        case Instruction::URem:
        case Instruction::Shl:
        case Instruction::LShr:
        case Instruction::AShr:
        case Instruction::And:
        case Instruction::Or:
        case Instruction::Xor: {
           
            break;
        }
    }
}
