#ifndef CreateStmt_H
#define CreateStmt_H

#include <iostream>
#include <map>
#include <memory>
#include "register.h"
#include "../IR/llvm.h"
#include "asm.h"


class CreateStmt {
public:
    LIRBasicBlock* BB_;
    Module* m_;
    LIRModule* m_lir;
    Register* ra_;
    bool is_o2_;
    bool print_s_verbosity_;
    std::map<GlobalVariable*, LIRGlobalVariable*> gv_map_;
    std::map<Function*, LIRFunction*> func_map_;
    std::map<Argument*, LIROperand*> arg_map_;
    std::map<BasicBlock*, LIRBasicBlock*> bb_map_;
    std::map<Instruction*, LIROperand*> inst_map_;
    CreateStmt(Module* m,bool psv,bool is_o2) : m_(m),print_s_verbosity_(psv),is_o2_(is_o2){
        m_lir=new LIRModule();
        ra_=nullptr;
        BB_=nullptr;
    };
    ~CreateStmt() = default;
    void clir();
    void carm();
    void global_arr_init(LIRGlobalVariable* lirgv, Constant* gv);
    int calc_type_size(Type* ty);
    bool is_constant_all_zero(Constant* gv);
    LIRBasicBlock* get_insert_block() { return this->BB_; }
    void set_insert_point(LIRBasicBlock* bb) { this->BB_ = bb; }
    void cinstr(std::list<Instruction*>::iterator iter3);

    void return_operand(Value* operand);
    void return_process_float(Value* operand);
   

    void remove_useless_insts();
    void remove_redundant_load_store();
    void combine_mul_addsub();
    void resolve_function_header();
    void resolve_imm();
    void resolve_libfunc();
    LIROperand* get_LIROpnd(Value *v);
    LIRFunction *getLFunction(Function *ori) {
       
        auto it = func_map_.find(ori);
    if (it != func_map_.end()) {
        return it->second;
    }
    return nullptr;
    }
    LIRBasicBlock *getLBB(BasicBlock *bb) {
       auto it = bb_map_.find(bb);
    if (it != bb_map_.end()) {
        return it->second;
    }
    return nullptr; 
    }
    
    LIRIntInst* create_lirinst(LIRInstruction::LIRInstTy id, LIRInstruction::LIRCondTy cond_id, LIROperand* result, LIROperand* v1, LIROperand* v2) ;
    LIRIntInst* create_lirinst(LIRInstruction::LIRInstTy id, LIRInstruction::LIRCondTy cond_id, LIROperand* result, 
                                    LIROperand* v1, LIROperand* v2, LIROperand* v3) ;
    LIRFloatInst* create_lirfinst(LIRInstruction::LIRInstTy id, LIRInstruction::LIRCondTy cond_id, LIROperand* result, LIROperand* v1, LIROperand* v2) ;
    LIRFloatInst* create_lirfinst(LIRInstruction::LIRInstTy id, LIRInstruction::LIRCondTy cond_id, LIROperand* result, LIROperand* v1) ;
    LIRIntInst* create_cmp(LIRInstruction::LIRCondTy cond_id, LIROperand* v1, LIROperand* v2) ;
    LIRFloatInst* create_vcmp(LIRInstruction::LIRCondTy cond_id, LIROperand* v1, LIROperand* v2) ;
    LIRVmrsInst* create_vmrs() ;
    LIRMovInst* create_mov(LIRInstruction::LIRCondTy cond_id, LIROperand* result, LIROperand* v1) ;
    LIRVmovInst* create_vmov(LIRInstruction::LIRCondTy cond_id, LIROperand* result, LIROperand* v1) ;
    LIRVmovf32Inst* create_vmovf32(LIRInstruction::LIRCondTy cond_id, LIROperand* result, LIROperand* v1) ;
    LIRMov32Inst* create_mov32(LIRInstruction::LIRCondTy cond_id, LIROperand* result, LIRConstant* v1) ;
    LIRMov32Inst* create_mov32(LIRInstruction::LIRCondTy cond_id, LIROperand* result, LIRFunction* v1) ;
    LIRMov32Inst* create_mov32(LIRInstruction::LIRCondTy cond_id, LIROperand* result, LIRGlobalVariable* v1) ;
    LIRLdrInst* create_ldr(LIRInstruction::LIRCondTy cond_id, LIROperand* result, LIROperand* v1, LIROperand* v2, LIRInstruction::LIRAddrTy addr_tid) ;
    LIRVldrInst* create_vldr(LIRInstruction::LIRCondTy cond_id, LIROperand* result, LIROperand* v1, LIROperand* v2) ;
    LIRStrInst* create_str(LIRInstruction::LIRCondTy cond_id, LIROperand* v1, LIROperand* v2, LIROperand* v3, LIRInstruction::LIRAddrTy addr_tid) ;
    LIRVstrInst* create_vstr(LIRInstruction::LIRCondTy cond_id, LIROperand* v1, LIROperand* v2, LIROperand* v3) ;
    LIRBInst* create_b(LIRInstruction::LIRCondTy cond_id, LIRBasicBlock* target_bb) ;
    LIRBlInst* create_bl(LIRInstruction::LIRCondTy cond_id, LIRFunction* target_func,bool result_unused) ;
    LIRPushInst *create_push(LIRInstruction::LIRCondTy cond_id, const std::vector<LIROperand *> &reglist) ;
    LIRPopInst *create_pop(LIRInstruction::LIRCondTy cond_id, const std::vector<LIROperand *> &reglist) ;
 };
#endif
