#ifndef IR_COND_H
#define IR_COND_H

#include "../IR/llvm.h"
#include "asm.h"


std::map<Value*, unsigned int> var_offset;
std::map<Value*, unsigned int> var_type;
std::map<BasicBlock*, std::string> bb_label_name;

LIRInstruction::LIRCondTy Ircond (LIRInstruction::LIRCondTy id) {
    switch (id) {
        case LIRInstruction::LIRCondTy::LCondEQ:
          return LIRInstruction::LIRCondTy::LCondEQ;
        case LIRInstruction::LIRCondTy::LCondNE:
          return LIRInstruction::LIRCondTy::LCondNE;
        case LIRInstruction::LIRCondTy::LCondGT:
            return LIRInstruction::LIRCondTy::LCondLT;
            break;
        case LIRInstruction::LIRCondTy::LCondGE:
            return LIRInstruction::LIRCondTy::LCondLE;
            break;
        case LIRInstruction::LIRCondTy::LCondLT:
            return LIRInstruction::LIRCondTy::LCondGT;
            break;
        case LIRInstruction::LIRCondTy::LCondLE:
            return LIRInstruction::LIRCondTy::LCondGE;
            break;
        case LIRInstruction::LIRCondTy::LCondHI:
            return LIRInstruction::LIRCondTy::LCondCC;
            break;
        case LIRInstruction::LIRCondTy::LCondCS:
            return LIRInstruction::LIRCondTy::LCondLS;
            break;
        case LIRInstruction::LIRCondTy::LCondCC:
            return  LIRInstruction::LIRCondTy::LCondHI;
            break;
        case LIRInstruction::LIRCondTy::LCondLS:
            return  LIRInstruction::LIRCondTy::LCondCS;
            break;
        default:
            break;
    }
    return id;
}


LIRInstruction::LIRCondTy Icmpcond (ICmpInst::ICmpOp id){
    switch (id) {
        case ICmpInst::ICmpOp::ICMP_EQ:
            return LIRInstruction::LIRCondTy::LCondEQ;
            break;
        case ICmpInst::ICmpOp::ICMP_NE:
            return LIRInstruction::LIRCondTy::LCondNE;
            break;
        case ICmpInst::ICmpOp::ICMP_UGT:
            return LIRInstruction::LIRCondTy::LCondHI;
            break;
        case ICmpInst::ICmpOp::ICMP_UGE:
            return LIRInstruction::LIRCondTy::LCondCS;
            break;
        case ICmpInst::ICmpOp::ICMP_ULT:
            return LIRInstruction::LIRCondTy::LCondCC;
            break;
        case ICmpInst::ICmpOp::ICMP_ULE:
            return LIRInstruction::LIRCondTy::LCondLS;
            break;
        case ICmpInst::ICmpOp::ICMP_SGT:
            return LIRInstruction::LIRCondTy::LCondGT;
            break;
        case ICmpInst::ICmpOp::ICMP_SGE:
            return LIRInstruction::LIRCondTy::LCondGE;
            break;
        case ICmpInst::ICmpOp::ICMP_SLT:
            return LIRInstruction::LIRCondTy::LCondLT;
            break;
        case ICmpInst::ICmpOp::ICMP_SLE:
            return LIRInstruction::LIRCondTy::LCondLE;
            break;
        default:
            break;
    }
    return LIRInstruction::LIRCondTy::LCondLE;
}


LIRInstruction::LIRCondTy Fcmpcond (FCmpInst::FCmpOp id) {
    switch (id) {
        case FCmpInst::FCmpOp::FCMP_OEQ:
            return LIRInstruction::LIRCondTy::LCondEQ;
            break;
        case FCmpInst::FCmpOp::FCMP_OGE:
            return LIRInstruction::LIRCondTy::LCondGE;
            break;
        case FCmpInst::FCmpOp::FCMP_OGT:
            return LIRInstruction::LIRCondTy::LCondGT;
            break;
        case FCmpInst::FCmpOp::FCMP_OLT:
            return LIRInstruction::LIRCondTy::LCondCC;
            break;
        case FCmpInst::FCmpOp::FCMP_OLE:
            return LIRInstruction::LIRCondTy::LCondLS;
            break;
        case FCmpInst::FCmpOp::FCMP_ONE:
            return LIRInstruction::LIRCondTy::LCondNE;
            break;
        case FCmpInst::FCmpOp::FCMP_ORD:
            return LIRInstruction::LIRCondTy::LCondVC;
            break;
        case FCmpInst::FCmpOp::FCMP_UNO:
            return LIRInstruction::LIRCondTy::LCondVS;
            break;
        case FCmpInst::FCmpOp::FCMP_UEQ:
            return LIRInstruction::LIRCondTy::LCondEQ;
            break;
        case FCmpInst::FCmpOp::FCMP_UGT:
            return LIRInstruction::LIRCondTy::LCondHI;
            break;
        case FCmpInst::FCmpOp::FCMP_UGE:
            return LIRInstruction::LIRCondTy::LCondCS;
            break;
        case FCmpInst::FCmpOp::FCMP_ULT:
            return LIRInstruction::LIRCondTy::LCondLT;
            break;
        case FCmpInst::FCmpOp::FCMP_ULE:
            return LIRInstruction::LIRCondTy::LCondLE;
            break;
        case FCmpInst::FCmpOp::FCMP_UNE:
            return LIRInstruction::LIRCondTy::LCondNE;
            break;
        case FCmpInst::FCmpOp::FCMP_TRUE:
            return LIRInstruction::LIRCondTy::LCondNone;
            break;
        default:
            break;

    }
    return LIRInstruction::LIRCondTy::LCondLE;
}


#endif