#include "asm.h"
std::string LIRPhysRegname_(int id) {
    switch (id) {
        case 1:
            return "r0";
        case 2:
            return "r1";
        case 3:
            return "r2";
        case 4:
            return "r3";
        case 5:
            return "r4";
        case 6:
            return "r5";
        case 7:
            return "r6";
        case 8:
            return "r7";
        case 9:
            return "r8";
        case 10:
            return "r9";
        case 11:
            return "r10";
        case 12:
            return "r11";
        case 13:
            return "ip";
        case 14:
            return "sp";
        case 15:
            return "lr";
        case 16:
            return "pc";
        default:
            break;
    }
    return "";
}
std::string LIRPhysFRegname_ (int id) {
    switch (id) {
        case -1:
            return "s0";
        case -2:
            return "s1";
        case -3:
            return "s2";
        case -4:
            return "s3";
        case -5:
            return "s4";
        case -6:
            return "s5";
        case -7:
            return "s6";
        case -8:
            return "s7";
        case -9:
            return "s8";
        case -10:
            return "s9";
        case -11:
            return "s10";
        case -12:
            return "s11";
        case -13:
            return "s12";
        case -14:
            return "s13";
        case -15:
            return "s14";
        case -16:
            return "s15";
        case -17:
            return "s16";
        case -18:
            return "s17";
        case -19:
            return "s18";
        case -20:
            return "s19";
        case -21:
            return "s20";
        case -22:
            return "s21";
        case -23:
            return "s22";
        case -24:
            return "s23";
        case -25:
            return "s24";
        case -26:
            return "s25";
        case -27:
            return "s26";
        case -28:
            return "s27";
        case -29:
            return "s28";
        case -30:
            return "s29";
        case -31:
            return "s30";
        case -32:
            return "s31";
        default:
            break;
    }
    return "";
}              
const std::map<int, std::string> LIRPhysReg::name_ = {{1, "r0"}, {2, "r1"}, {3, "r2"}, {4, "r3"}, {5, "r4"},
                                                      {6, "r5"}, {7, "r6"}, {8, "r7"}, {9, "r8"}, {10, "r9"}, {11, "r10"},
                                                      {12, "r11"}, {13, "ip"}, {14, "sp"}, {15, "lr"}, {16, "pc"}};
                                        
const std::map<int, std::string> LIRPhysFReg::name_ = {
    {-1, "s0"}, {-2, "s1"}, {-3, "s2"}, {-4, "s3"}, {-5, "s4"}, {-6, "s5"}, {-7, "s6"}, {-8, "s7"}, {-9, "s8"}, {-10, "s9"},
    {-11, "s10"}, {-12, "s11"}, {-13, "s12"}, {-14, "s13"}, {-15, "s14"}, {-16, "s15"}, {-17, "s16"}, {-18, "s17"}, {-19, "s18"}, {-20, "s19"}, {-21, "s20"},
    {-22, "s21"}, {-23, "s22"}, {-24, "s23"}, {-25, "s24"}, {-26, "s25"}, {-27, "s26"}, {-28, "s27"}, {-29, "s28"}, {-30, "s29"}, {-31, "s30"}, {-32, "s31"}};


const std::map<LIRShift::LIRShiftTy, std::string> LIRShift::name_ = {
    {LIRShift::LSAsrTy, "asr"},
    {LIRShift::LSLslTy, "lsl"},
    {LIRShift::LSLsrTy, "lsr"},
    {LIRShift::LSRorTy, "ror"},
};


const std::map<LIRInstruction::LIRInstTy, std::string> LIRInstruction::inst_name_ = {{LIRInstruction::LAddTy, "add"},
                                                                                     {LIRInstruction::LSubTy, "sub"},
                                                                                     {LIRInstruction::LRsbTy, "rsb"},
                                                                                     {LIRInstruction::LMulTy, "mul"},
                                                                                     {LIRInstruction::LMlaTy, "mla"},
                                                                                     {LIRInstruction::LMlsTy, "mls"},
                                                                                     {LIRInstruction::LSdivTy, "sdiv"},
                                                                                     {LIRInstruction::LUdivTy, "udiv"},
                                                                                     {LIRInstruction::LAndTy, "and"},
                                                                                     {LIRInstruction::LOrrTy, "orr"},
                                                                                     {LIRInstruction::LEorTy, "eor"},
                                                                                     {LIRInstruction::LLslTy, "lsl"},
                                                                                     {LIRInstruction::LLsrTy, "lsr"},
                                                                                     {LIRInstruction::LAsrTy, "asr"},
                                                                                     {LIRInstruction::LRorTy, "ror"},
                                                                                     {LIRInstruction::LBTy, "b"},
                                                                                     {LIRInstruction::LBlTy, "bl"},
                                                                                     {LIRInstruction::LBxTy, "bx"},
                                                                                     {LIRInstruction::LMovTy, "mov"},
                                                                                     {LIRInstruction::LVmovTy, "vmov"},
                                                                                     {LIRInstruction::LVmovf32Ty, "vmov.f32"},
                                                                                     {LIRInstruction::LMov32Ty, "mov32"},
                                                                                     {LIRInstruction::LLdrTy, "ldr"},
                                                                                     {LIRInstruction::LVldrTy, "vldr"},
                                                                                     {LIRInstruction::LStrTy, "str"},
                                                                                     {LIRInstruction::LVstrTy, "vstr"},
                                                                                     {LIRInstruction::LPushTy, "push"},
                                                                                     {LIRInstruction::LPopTy, "pop"},
                                                                                     {LIRInstruction::LVpushTy, "vpush"},
                                                                                     {LIRInstruction::LVpopTy, "vpop"},
                                                                                     {LIRInstruction::LCmpTy, "cmp"},
                                                                                     {LIRInstruction::LVcmpTy, "vcmp.f32"},
                                                                                     {LIRInstruction::LVaddTy, "vadd.f32"},
                                                                                     {LIRInstruction::LVsubTy, "vsub.f32"},
                                                                                     {LIRInstruction::LVdivTy, "vdiv.f32"},
                                                                                     {LIRInstruction::LVmulTy, "vmul.f32"},
                                                                                     {LIRInstruction::LVmlaTy, "vmla.f32"},
                                                                                     {LIRInstruction::LVmlsTy, "vmls.f32"},
                                                                                     {LIRInstruction::LVnegTy, "vneg.f32"},
                                                                                     {LIRInstruction::LVcvtFSTy, "vcvt.f32.s32"},
                                                                                     {LIRInstruction::LVcvtSFTy, "vcvt.s32.f32"},
                                                                                     {LIRInstruction::LVmrsTy, "vmrs"}};

const std::map<LIRInstruction::LIRCondTy, std::string> LIRInstruction::cond_name_ = {
    {LIRInstruction::LCondGT, "gt"}, {LIRInstruction::LCondGE, "ge"}, {LIRInstruction::LCondLT, "lt"}, {LIRInstruction::LCondLE, "le"},
    {LIRInstruction::LCondHI, "hi"}, {LIRInstruction::LCondCS, "cs"}, {LIRInstruction::LCondCC, "cc"}, {LIRInstruction::LCondLS, "ls"},
    {LIRInstruction::LCondEQ, "eq"}, {LIRInstruction::LCondNE, "ne"}, {LIRInstruction::LCondVC, "vc"}, {LIRInstruction::LCondVS, "vs"}, {LIRInstruction::LCondNone, ""}};


template <typename... Args>
std::string strfmt(const std::string& format, Args... args) {
    int size_s = std::snprintf(nullptr, 0, format.c_str(), args...) + 1;
    if (size_s <= 0) {
        throw std::runtime_error("Error during formatting.");
    }
    auto size = static_cast<size_t>(size_s);
    std::unique_ptr<char[]> buf(new char[size]);
    std::snprintf(buf.get(), size, format.c_str(), args...);
    return std::string(buf.get(), buf.get() + size - 1);
}
std::string LIRMovInst::print() {
    return "\t" + inst_name_.at(op_id_) + cond_name_.at(cond_id_) + " "+result_->print() + ","+operands_[0]->print()+"\n";
}
std::string LIRConstant::print() {
    return "#" + std::to_string(value_);
}
std::string LIRShift::print() {
    return name_.at(tid_)+" #" + std::to_string(amount_);

}
std::string LIRVirtReg::print() {
    return "v" + std::to_string(rid_);
}
std::string LIRPhysReg::print() {
    return name_.at(rid_);
}
std::string LIRVirtFReg::print() {
    return "x" + std::to_string(frid_);
}
std::string LIRPhysFReg::print() {
    return name_.at(frid_);

}
std::string LIRBInst::print() {
    return "\t" + inst_name_.at(op_id_) + cond_name_.at(cond_id_) +" "+static_cast<LIRLabel*>(operands_[0])->name_+"\n";
}
std::string LIRBlInst::print() {
    return "\t" + inst_name_.at(op_id_) + cond_name_.at(cond_id_) + " "+static_cast<LIRLabel*>(operands_[0])->name_+"\n";
}
std::string LIRBxInst::print() {
    return "\t" + inst_name_.at(op_id_) + cond_name_.at(cond_id_) + " "+operands_[0]->print()+"\n";
}
std::string LIRVmrsInst::print() {
    return "\t" + inst_name_.at(op_id_) + cond_name_.at(cond_id_) + " APSR_nzcv, FPSCR\n";
}
std::string LIRModule::print() {
    std::string ret =
        "\t.arch armv7-a\n"
        "\t.eabi_attribute 28, 1\n"
        "\t.eabi_attribute 20, 1\n"
        "\t.eabi_attribute 21, 1\n"
        "\t.eabi_attribute 23, 3\n"
        "\t.eabi_attribute 24, 1\n"
        "\t.eabi_attribute 25, 1\n"
        "\t.eabi_attribute 26, 2\n"
        "\t.eabi_attribute 30, 6\n"
        "\t.eabi_attribute 34, 1\n"
        "\t.eabi_attribute 18, 4\n"
        "\t.arm\n"
        "\t.cpu cortex-a15\n"
        "\t.fpu neon-vfpv4\n"
        "\t.macro mov32, reg, val\n"
        "\t\tmovw \\reg, #:lower16:\\val\n"
        "\t\tmovt \\reg, #:upper16:\\val\n"
        "\t.endm\n";
    for (LIRGlobalVariable* gv : global_list_) {
        ret += gv->print();
    }
    ret += "\t.text\n";
    for (LIRFunction* f : function_list_) {
        ret += f->print();
    }
    return ret;
}
std::string LIRGlobalVariable::print() {
    std::string ret = "\t.global " + name_ + "\n";
    if (tid_ == LDataTy) {
        ret += "\t.data\n";
    } else {
        ret += "\t.bss\n";
    }
    ret += "\t.align 2\n";
    ret += "\t.type " + name_ + ", %object\n";
    ret += name_ + ":\n";
    for (LIRDataDef* dd : init_val_) {
        ret += dd->print();
    }
    return ret;
}
std::string LIRDataDef::print() {
    std::string ret = "\t";
    if (tid_ == LWordTy) {
        ret += ".word ";
    } else {
        ret += ".zero ";
    }
    ret += strfmt("%u\n", expr_);
    return ret;
}
std::string LIRFunction::print() {
    std::string ret = "\t.global " + name_ + "\n" + name_ + ":\t\t\t@num_args_ = "+std::to_string(num_args_)+"; has_result_ = "+(has_result_?"true":"false")+"\n";
    for (LIRBasicBlock* bb : basic_blocks_) {
        ret += bb->print();
    }
    return ret;
}
std::string LIRBasicBlock::print() {
    std::string ret = name_ + ":\t\t\t@";
    ret+="pre_bbs_ = ";
    std::string delim = "";
    for(LIRBasicBlock* prebb:pre_bbs_){
        ret+=delim+prebb->name_;
        delim=", ";
    }
    ret+="; succ_bbs_ = ";
    delim="";
    for(LIRBasicBlock* succbb:succ_bbs_){
        ret+=delim+succbb->name_;
        delim=", ";
    }
    ret+=";\n";
    for (LIRInstruction* inst : instr_list_) {
        ret += inst->print();
    }
    return ret;
}
std::string LIRIntInst::print() {
    std::string ret = "\t" + inst_name_.at(op_id_) + cond_name_.at(cond_id_) + " ";
    std::string delim = "";
    if (result_ != nullptr) {
        ret += delim + result_->print();
        delim=",";
    }
    for (LIROperand* op : operands_) {
        ret += delim + op->print();
        delim=",";
    }
    ret+="\n";
    return ret;
}
std::string LIRFloatInst::print() {
    std::string ret = "\t" + inst_name_.at(op_id_) + cond_name_.at(cond_id_) + " ";
    std::string delim = "";
    if (result_ != nullptr) {
        ret += delim + result_->print();
        delim=",";
    }
    for (LIROperand* op : operands_) {
        ret += delim + op->print();
        delim=",";
    }
    ret+="\n";
    return ret;
}

std::string LIRVmovInst::print() {
    std::string ret="";
    LIROperand::LIROprTy result_tid=result_->tid_;
    LIROperand::LIROprTy op0_tid=operands_[0]->tid_;
   
     ret+= "\t" + inst_name_.at(op_id_) + cond_name_.at(cond_id_) + " "+result_->print() + ","+operands_[0]->print()+"\n";
    return ret;
}
std::string LIRVmovf32Inst::print() {
    
    std::string ret = "\t" + inst_name_.at(op_id_) + cond_name_.at(cond_id_) + " "+result_->print() + ","+operands_[0]->print()+"\n";
    return ret;
}
std::string LIRMov32Inst::print() {
    
    std::string ret = "\t" + inst_name_.at(op_id_) + cond_name_.at(cond_id_) + " "+result_->print() + ",";
    if(dynamic_cast<LIRConstant*>(operands_[0])){
        ret+=(operands_[0]->print()).substr(1,std::string::npos)+"\n";
    }else{
       
        ret+=static_cast<LIRLabel*>(operands_[0])->name_+"\n";
    }
    return ret;
}
std::string LIRLdrInst::print() {
    std::string ret = "\t" + inst_name_.at(op_id_) + cond_name_.at(cond_id_) + " ";
   
    ret += result_->print() + ",";
        if(addr_tid_==LRegisterOffsetTy){
            ret += "[" + operands_[0]->print() + "," + operands_[1]->print() + "]\n";
        }else if(addr_tid_==LPreIndexedTy){
            ret += "[" + operands_[0]->print() + "," + operands_[1]->print() + "]!\n";
        }else{
            ret += "[" + operands_[0]->print() + "]," + operands_[1]->print() + "\n";
        }
    return ret;
}
std::string LIRVldrInst::print() {
    std::string ret = "\t" + inst_name_.at(op_id_) + cond_name_.at(cond_id_) + " ";
   
    ret += result_->print() + ",";
        ret += "[" + operands_[0]->print() + "," + operands_[1]->print() + "]\n";
    return ret;
}
std::string LIRStrInst::print() {
    std::string ret = "\t" + inst_name_.at(op_id_) + cond_name_.at(cond_id_) + " ";
   
        if(addr_tid_==LRegisterOffsetTy){
            ret += operands_[0]->print() + ",[" + operands_[1]->print() + "," + operands_[2]->print() + "]\n";
        }else if(addr_tid_==LPreIndexedTy){
            ret += operands_[0]->print() + ",[" + operands_[1]->print() + "," + operands_[2]->print() + "]!\n";
        }else{
           
            ret += operands_[0]->print() + ",[" + operands_[1]->print() + "]," + operands_[2]->print() + "\n";
        }
    return ret;
}
std::string LIRVstrInst::print() {
    std::string ret = "\t" + inst_name_.at(op_id_) + cond_name_.at(cond_id_) + " ";
   
        ret += operands_[0]->print() + ",[" + operands_[1]->print() + "," + operands_[2]->print() + "]\n";
    return ret;
}

std::string LIRPushInst::print() {
    std::string ret = "\t" + inst_name_.at(op_id_) + cond_name_.at(cond_id_) + " {";
    std::string delim="";
    for (LIROperand* op : operands_) {
        ret += delim + op->print();
        delim=",";
    }
    ret+="}\n";
    return ret;
}
std::string LIRPopInst::print() {
    std::string ret = "\t" + inst_name_.at(op_id_) + cond_name_.at(cond_id_) + " {";
    std::string delim="";
    for (LIROperand* op : operands_) {
        ret += delim + op->print();
        delim=",";
    }
    ret+="}\n";
    return ret;
}
std::string LIRVpushInst::print() {
    std::string ret = "\t" + inst_name_.at(op_id_) + cond_name_.at(cond_id_) + " {";
    std::string delim="";
    for (LIROperand* op : operands_) {
        ret += delim + op->print();
        delim=",";
    }
    ret+="}\n";
    return ret;
}
std::string LIRVpopInst::print() {
    std::string ret = "\t" + inst_name_.at(op_id_) + cond_name_.at(cond_id_) + " {";
    std::string delim="";
    for (LIROperand* op : operands_) {
        ret += delim + op->print();
        delim=",";
    }
    ret+="}\n";
    return ret;
}

