   #include "asm.h"
   
   void LIRInstruction::set_operand(int i, LIROperand* v) {
        operands_[i] = v;
        opr_use_pos_[i] = v->add_use(this, i);
    }
    void LIRInstruction::remove_operand(int i){
        operands_[i]->remove_use(opr_use_pos_[i]);
        operands_[i]=nullptr;
    }
    void LIRInstruction::remove_operand_real(int i) {
        operands_[i]->remove_use(opr_use_pos_[i]);
        operands_.erase(operands_.begin() + i);
        opr_use_pos_.erase(opr_use_pos_.begin() + i);
        num_ops_ = operands_.size();
    }
    void LIRInstruction::set_result(LIROperand* v) {
        result_ = v;
        res_use_pos_ = v->add_use(this, -1);
    }
    void LIRInstruction::remove_result(){
        result_->remove_use(res_use_pos_);
        result_=nullptr;
    }
    void LIRInstruction::remove_result_and_operands() {
        if(result_!=nullptr){
            remove_result();
        }
        for (size_t i = 0; i < operands_.size(); i++ ) {
            remove_operand(i);
        }
    }
    std::vector<LIROperand*> LIRInstruction::get_operands_and_result(){
        std::vector<LIROperand*> ans(operands_);
        if(result_!=nullptr){
            ans.push_back(result_);
        }
        return ans;
    }

void LIRBasicBlock::delete_instr(LIRInstruction* instr){
    instr_list_.remove(instr);
    for(auto it:instr->args_offset_use_pos_){
        parent_->args_offset_.erase(it);
    }
    instr->remove_result_and_operands();
}
void LIRBasicBlock::replace_instr(LIRInstruction* oldinst,LIRInstruction* newinst){
    auto iter = std::find(instr_list_.begin(),instr_list_.end(),oldinst);
    oldinst->remove_result_and_operands();
    *iter = newinst;
}
bool LIRBasicBlock::add_inst_before_br(LIRInstruction* instr){
    for (auto it = instr_list_.begin(); it != instr_list_.end(); ++it) {
        if ((*it)->op_id_ == LIRInstruction::LBTy) {
            instr->parent_=this;
            instr_list_.insert(it,instr);
            return true;
        }
    }
    return false;
}
bool LIRBasicBlock::add_inst_before(LIRInstruction *instr, LIRInstruction *dst){
    for (auto it = instr_list_.begin(); it != instr_list_.end(); ++it) {
        if ((*it) == dst) {
            instr->parent_=this;
            instr_list_.insert(it,instr);
            return true;
        }
    }
    return false;
}
bool LIRBasicBlock::add_inst_after(LIRInstruction *instr, LIRInstruction *dst){
    if(dst->op_id_==LIRInstruction::LBTy){
        return this->add_inst_before(instr, dst);
    }
    for (auto it = instr_list_.begin(); it != instr_list_.end(); ++it) {
        if ((*it) == dst) {
            it++;
            instr->parent_=this;
            instr_list_.insert(it,instr);
            return true;
        }
    }
    return false;
}
void LIRBasicBlock::add_inst_back(LIRInstruction* instr) {
    instr->parent_=this;
    instr_list_.push_back(instr);
}
void LIRBasicBlock::add_inst_front(LIRInstruction* instr){
    instr->parent_=this;
    instr_list_.push_front(instr);
}

 LIRConstant* LIRModule::get_LIRConstant(int val) {
        if (!LIRConstant_map_.count(val)) {
            LIRConstant_map_[val] = new LIRConstant(val);
        }
        return LIRConstant_map_[val];
    }
    LIRVirtReg* LIRModule::get_LIRVirtReg(int rid) { return LIRVirtReg_map_.at(rid); }
    LIRVirtReg* LIRModule::get_LIRVirtReg() {
        int rid = total_rid_++;
        LIRVirtReg_map_[rid] = new LIRVirtReg(rid);
        return LIRVirtReg_map_[rid];
    }
    LIRPhysReg* LIRModule::get_LIRPhysReg(int rid) {
        if (!LIRPhysReg_map_.count(rid)) {
            LIRPhysReg_map_[rid] = new LIRPhysReg(rid);
        }
        return LIRPhysReg_map_[rid];
    }
    LIRVirtFReg* LIRModule::get_LIRVirtFReg(int frid) { return LIRVirtFReg_map_.at(frid); }
    LIRVirtFReg* LIRModule::get_LIRVirtFReg() {
        int frid = total_frid_++;
        LIRVirtFReg_map_[frid] = new LIRVirtFReg(frid);
        return LIRVirtFReg_map_[frid];
    }
    LIRPhysFReg* LIRModule::get_LIRPhysFReg(int frid) {
        if (!LIRPhysFReg_map_.count(frid)) {
            LIRPhysFReg_map_[frid] = new LIRPhysFReg(frid);
        }
        return LIRPhysFReg_map_[frid];
    }
    LIRDataDef* LIRModule::get_LIRDataDef(LIRDataDef::LIRDataDefTy tid, int expr) {
        if (!LIRDataDef_map_.count({tid, expr})) {
            LIRDataDef_map_[{tid, expr}] = new LIRDataDef(tid, expr);
        }
        return LIRDataDef_map_[{tid, expr}];
    }
    LIRShift* LIRModule::get_LIRShift(LIRShift::LIRShiftTy tid, int amount) {
        if (!LIRShift_map_.count({tid, amount})) {
            LIRShift_map_[{tid, amount}] = new LIRShift(tid, amount);
        }
        return LIRShift_map_[{tid, amount}];
    }