#include "build.h"

void CreateStmt::resolve_function_header() {
    for (LIRFunction* lirf : m_lir->function_list_) {
        LIRBasicBlock *initBB, *exitBB;
        std::map<int, int> useStatus;
        bool isLeaf = true;
        for (size_t ittt = 0; ittt < lirf->basic_blocks_.size(); ittt++) {
            LIRBasicBlock* lirbb = lirf->basic_blocks_[ittt];
            if (ittt == 0) {
                initBB = lirbb;
            } else if (ittt == 2) {
                exitBB = lirbb;
            }
            for (LIRInstruction* lirinst : lirbb->instr_list_) {
                for (LIROperand* lirop : lirinst->get_operands_and_result()) {
                    if (auto pr = dynamic_cast<LIRPhysReg*>(lirop)) {
                        if (pr->rid_ <= 12) {
                            useStatus[pr->rid_] = 1;
                        }
                    } else if (auto pfr = dynamic_cast<LIRPhysFReg*>(lirop)) {
                        if (pfr->frid_ >= -31) {
                            useStatus[pfr->frid_] = 1;
                        }
                    }
                }
                if (lirinst->op_id_ == LIRInstruction::LBlTy) {
                    isLeaf = false;
                }
            }
        }
        lirf->stackframe_offset_end_ = (lirf->spills_offset_end_ % 8 != 0) ? ((lirf->spills_offset_end_ / 8 + 1) * 8) : (lirf->spills_offset_end_);
        if (lirf->stackframe_offset_end_ > 0) {
            initBB->add_inst_front(new LIRIntInst(LIRInstruction::LSubTy, LIRInstruction::LCondNone, m_lir->get_LIRPhysReg(14), m_lir->get_LIRPhysReg(14), m_lir->get_LIRConstant(lirf->stackframe_offset_end_)));
            exitBB->add_inst_back(new LIRIntInst(LIRInstruction::LAddTy, LIRInstruction::LCondNone, m_lir->get_LIRPhysReg(14), m_lir->get_LIRPhysReg(14), m_lir->get_LIRConstant(lirf->stackframe_offset_end_)));
        }
        if (lirf->name_ == "main" && !isLeaf) {
            initBB->add_inst_front(new LIRStrInst(LIRInstruction::LCondNone, m_lir->get_LIRPhysReg(2), m_lir->get_LIRPhysReg(14), m_lir->get_LIRConstant(-8), LIRInstruction::LPreIndexedTy));
            initBB->add_inst_front(new LIRIntInst(LIRInstruction::LAddTy, LIRInstruction::LCondNone, m_lir->get_LIRPhysReg(14), m_lir->get_LIRPhysReg(1), m_lir->get_LIRConstant(67108864)));
            initBB->add_inst_front(new LIRMovInst(LIRInstruction::LCondNone, m_lir->get_LIRPhysReg(2), m_lir->get_LIRPhysReg(14)));
            initBB->add_inst_front(new LIRBlInst(LIRInstruction::LCondNone, new LIRFunction("malloc", 1, 1),false));
            initBB->add_inst_front(new LIRMovInst(LIRInstruction::LCondNone, m_lir->get_LIRPhysReg(1), m_lir->get_LIRConstant(67108864)));
            exitBB->add_inst_back(new LIRLdrInst(LIRInstruction::LCondNone, m_lir->get_LIRPhysReg(14), m_lir->get_LIRPhysReg(14), m_lir->get_LIRConstant(0), LIRInstruction::LRegisterOffsetTy));
        }
        std::vector<LIRPhysReg*> regs_to_protect;
        std::vector<LIRPhysFReg*> fregs_to_protect;
        for (int i = 1; i <= 12; i++) {
            if (useStatus.count(i) && useStatus[i] == 1) {
                if (i == 1 && lirf->has_result_ == 1) {
                    continue;
                }
                regs_to_protect.push_back(m_lir->get_LIRPhysReg(i));
            }
        }
        if (!isLeaf) {
            regs_to_protect.insert(regs_to_protect.begin(), m_lir->get_LIRPhysReg(15));
        }
        for (int i = -1; i >= -31; i--) {
            if (useStatus.count(i) && useStatus[i] == 1) {
                if (i == -1 && lirf->has_result_ == 2) {
                    continue;
                }
                fregs_to_protect.push_back(m_lir->get_LIRPhysFReg(i));
            }
        }
        std::list<LIRInstruction*> addfrontinst;
        std::list<LIRInstruction*> addbackinst;
        if (!regs_to_protect.empty()) {
            for (int i = 0; i < regs_to_protect.size(); i += 2) {
                std::vector<LIROperand*> new_vec_push;
                std::vector<LIROperand*> new_vec_pop;
                new_vec_push.push_back(regs_to_protect[i]);
                new_vec_pop.push_back((regs_to_protect[i]->rid_ == 15) ? (m_lir->get_LIRPhysReg(16)) : (regs_to_protect[i]));
                if (i + 1 < regs_to_protect.size()) {
                    if (regs_to_protect[i + 1]->rid_ < regs_to_protect[i]->rid_) {
                        new_vec_push.insert(new_vec_push.begin(), regs_to_protect[i + 1]);
                    } else {
                        new_vec_push.push_back(regs_to_protect[i + 1]);
                    }
                    if (((regs_to_protect[i + 1]->rid_ == 15) ? (16) : (regs_to_protect[i + 1]->rid_)) < regs_to_protect[i]->rid_) {
                        new_vec_pop.insert(new_vec_pop.begin(), (regs_to_protect[i + 1]->rid_ == 15) ? (m_lir->get_LIRPhysReg(16)) : (regs_to_protect[i + 1]));
                    } else {
                        new_vec_pop.push_back((regs_to_protect[i + 1]->rid_ == 15) ? (m_lir->get_LIRPhysReg(16)) : (regs_to_protect[i + 1]));
                    }
                }
                addfrontinst.push_back(new LIRPushInst(LIRInstruction::LCondNone, new_vec_push));
                addbackinst.push_front(new LIRPopInst(LIRInstruction::LCondNone, new_vec_pop));
            }
        }
        if (!fregs_to_protect.empty()) {
            for (int i = 0; i < fregs_to_protect.size(); i++) {
                std::vector<LIROperand*> new_vec;
                new_vec.push_back(fregs_to_protect[i]);
                addfrontinst.push_back(new LIRVpushInst(LIRInstruction::LCondNone, new_vec));
                addbackinst.push_front(new LIRVpopInst(LIRInstruction::LCondNone, new_vec));
            }
        }
        int saveregs_size = (regs_to_protect.size() + fregs_to_protect.size()) * 4;
        if (saveregs_size % 8 != 0) {
            saveregs_size += 4;
            addfrontinst.push_back(new LIRIntInst(LIRInstruction::LSubTy, LIRInstruction::LCondNone, m_lir->get_LIRPhysReg(14), m_lir->get_LIRPhysReg(14), m_lir->get_LIRConstant(4)));
            addbackinst.push_front(new LIRIntInst(LIRInstruction::LAddTy, LIRInstruction::LCondNone, m_lir->get_LIRPhysReg(14), m_lir->get_LIRPhysReg(14), m_lir->get_LIRConstant(4)));
        }
        lirf->saveregs_offset_end_ = lirf->stackframe_offset_end_ + saveregs_size;
        for (auto iter = addfrontinst.rbegin(); iter != addfrontinst.rend(); iter++) {
            initBB->add_inst_front(*iter);
        }
        for (auto inst : addbackinst) {
            exitBB->add_inst_back(inst);
        }
        if (isLeaf) {
            exitBB->add_inst_back(new LIRBxInst(LIRInstruction::LCondNone, m_lir->get_LIRPhysReg(15)));
        }
        for (auto pp : lirf->args_offset_) {
            LIRInstruction* lirinst = pp.first;
            int argoff = pp.second;
            lirinst->remove_operand(1);
            lirinst->set_operand(1, m_lir->get_LIRConstant(argoff + lirf->saveregs_offset_end_));
            lirf->args_offset_[lirinst] = argoff + lirf->saveregs_offset_end_;
        }
    }
}
void CreateStmt::resolve_imm() {
    for (LIRFunction* lirf : m_lir->function_list_) {
        for (LIRBasicBlock* lirbb : lirf->basic_blocks_) {
            std::map<LIRInstruction*, std::list<LIRInstruction*>> addinst_map;
            for (LIRInstruction* lirinst : lirbb->instr_list_) {
                std::list<LIRInstruction*> addbefore;
                switch (lirinst->op_id_) {
                    case LIRInstruction::LAddTy:
                    case LIRInstruction::LSubTy:
                    case LIRInstruction::LCmpTy: {
                        if (auto c = dynamic_cast<LIRConstant*>(lirinst->get_operand(1))) {
                            int val = std::abs(c->value_);
                            if (val <= 255 || val <= 1020 && val % 4 == 0 || val <= 4080 && val % 16 == 0 || val <= 16320 && val % 64 == 0 || val <= 65280 && val % 256 == 0 || val <= 261120 && val % 1024 == 0 || val <= 1044480 && val % 4096 == 0 || val <= 4177920 && val % 16384 == 0 || val <= 16711680 && val % 65536 == 0 || val <= 66846720 && val % 262114 == 0 || val <= 267386880 && val % 1048576 == 0 || val <= 1069547520 && val % 4194304 == 0) {
                                break; 
                            }
                            lirinst->remove_operand(1);
                            lirinst->set_operand(1, m_lir->get_LIRPhysReg(13));
                            if (c->value_ >= -257 && c->value_ <= 65536) {
                                addbefore.push_back(new LIRMovInst(LIRInstruction::LCondNone, m_lir->get_LIRPhysReg(13), m_lir->get_LIRConstant(c->value_)));
                            } else {
                                addbefore.push_back(new LIRMov32Inst(LIRInstruction::LCondNone, m_lir->get_LIRPhysReg(13), m_lir->get_LIRConstant(c->value_)));
                            }
                        }
                        break;
                    }
                    case LIRInstruction::LRsbTy: {
                        if (auto c = dynamic_cast<LIRConstant*>(lirinst->get_operand(1))) {
                            int val = c->value_;
                            if ((val >= 0) && (val <= 255 || val <= 1020 && val % 4 == 0 || val <= 4080 && val % 16 == 0 || val <= 16320 && val % 64 == 0 || val <= 65280 && val % 256 == 0 || val <= 261120 && val % 1024 == 0 || val <= 1044480 && val % 4096 == 0 || val <= 4177920 && val % 16384 == 0 || val <= 16711680 && val % 65536 == 0 || val <= 66846720 && val % 262114 == 0 || val <= 267386880 && val % 1048576 == 0 || val <= 1069547520 && val % 4194304 == 0)) {
                                break; 
                            }
                            lirinst->remove_operand(1);
                            lirinst->set_operand(1, m_lir->get_LIRPhysReg(13));
                            if (c->value_ >= -257 && c->value_ <= 65536) {
                                addbefore.push_back(new LIRMovInst(LIRInstruction::LCondNone, m_lir->get_LIRPhysReg(13), m_lir->get_LIRConstant(c->value_)));
                            } else {
                                addbefore.push_back(new LIRMov32Inst(LIRInstruction::LCondNone, m_lir->get_LIRPhysReg(13), m_lir->get_LIRConstant(c->value_)));
                            }
                        }
                        break;
                    }
                    case LIRInstruction::LMovTy: {
                        if (auto c = dynamic_cast<LIRConstant*>(lirinst->get_operand(0))) {
                            if (c->value_ >= -257 && c->value_ <= 65536) {
                                break; 
                            }
                            int val = std::abs(c->value_);
                            if (val <= 255 || val <= 1020 && val % 4 == 0 || val <= 4080 && val % 16 == 0 || val <= 16320 && val % 64 == 0 || val <= 65280 && val % 256 == 0 || val <= 261120 && val % 1024 == 0 || val <= 1044480 && val % 4096 == 0 || val <= 4177920 && val % 16384 == 0 || val <= 16711680 && val % 65536 == 0 || val <= 66846720 && val % 262114 == 0 || val <= 267386880 && val % 1048576 == 0 || val <= 1069547520 && val % 4194304 == 0) {
                                break; 
                            }
                            lirinst->remove_operand(0);
                            lirinst->set_operand(0, m_lir->get_LIRPhysReg(13));
                            addbefore.push_back(new LIRMov32Inst(LIRInstruction::LCondNone, m_lir->get_LIRPhysReg(13), m_lir->get_LIRConstant(c->value_)));
                        }
                        break;
                    }
                    case LIRInstruction::LLdrTy: {
                        if (lirinst->num_ops_ == 2) {
                            if (auto c = dynamic_cast<LIRConstant*>(lirinst->get_operand(1))) {
                                if (c->value_ >= -4095 && c->value_ <= 4095) {
                                    break; 
                                }
                                lirinst->remove_operand(1);
                                lirinst->set_operand(1, m_lir->get_LIRPhysReg(13));
                                if (c->value_ >= -257 && c->value_ <= 65536) {
                                    addbefore.push_back(new LIRMovInst(LIRInstruction::LCondNone, m_lir->get_LIRPhysReg(13), m_lir->get_LIRConstant(c->value_)));
                                } else {
                                    addbefore.push_back(new LIRMov32Inst(LIRInstruction::LCondNone, m_lir->get_LIRPhysReg(13), m_lir->get_LIRConstant(c->value_)));
                                }
                            }
                        }
                        break;
                    }
                    case LIRInstruction::LVldrTy: {
                        if (lirinst->num_ops_ == 2) {
                            if (auto c = dynamic_cast<LIRConstant*>(lirinst->get_operand(1))) {
                                if (c->value_ >= -1020 && c->value_ <= 1020) {
                                    break; 
                                }
                                lirinst->remove_operand(1);
                                lirinst->set_operand(1,m_lir->get_LIRConstant(0));
                                lirinst->remove_operand(0);
                                lirinst->set_operand(0, m_lir->get_LIRPhysReg(13));
                                int val = std::abs(c->value_);
                                if (val <= 255 || val <= 1020 && val % 4 == 0 || val <= 4080 && val % 16 == 0 || val <= 16320 && val % 64 == 0 || val <= 65280 && val % 256 == 0 || val <= 261120 && val % 1024 == 0 || val <= 1044480 && val % 4096 == 0 || val <= 4177920 && val % 16384 == 0 || val <= 16711680 && val % 65536 == 0 || val <= 66846720 && val % 262114 == 0 || val <= 267386880 && val % 1048576 == 0 || val <= 1069547520 && val % 4194304 == 0) {
                                    addbefore.push_back(new LIRIntInst(LIRInstruction::LAddTy, LIRInstruction::LCondNone, m_lir->get_LIRPhysReg(13), m_lir->get_LIRPhysReg(14), m_lir->get_LIRConstant(c->value_)));
                                } else {
                                    addbefore.push_back(new LIRMov32Inst(LIRInstruction::LCondNone, m_lir->get_LIRPhysReg(13), m_lir->get_LIRConstant(c->value_)));
                                    addbefore.push_back(new LIRIntInst(LIRInstruction::LAddTy, LIRInstruction::LCondNone, m_lir->get_LIRPhysReg(13), m_lir->get_LIRPhysReg(14), m_lir->get_LIRPhysReg(13)));
                                }
                            }
                        }
                        break;
                    }
                    case LIRInstruction::LStrTy: {
                        if (lirinst->num_ops_ == 3) {
                            if (auto c = dynamic_cast<LIRConstant*>(lirinst->get_operand(2))) {
                                if (c->value_ >= -4095 && c->value_ <= 4095) {
                                    break;
                                }
                                lirinst->remove_operand(2);
                                lirinst->set_operand(2, m_lir->get_LIRPhysReg(13));
                                if (c->value_ >= -257 && c->value_ <= 65536) {
                                    addbefore.push_back(new LIRMovInst(LIRInstruction::LCondNone, m_lir->get_LIRPhysReg(13), m_lir->get_LIRConstant(c->value_)));
                                } else {
                                    addbefore.push_back(new LIRMov32Inst(LIRInstruction::LCondNone, m_lir->get_LIRPhysReg(13), m_lir->get_LIRConstant(c->value_)));
                                }
                            }
                        }
                        break;
                    }
                    case LIRInstruction::LVstrTy: {
                        if (lirinst->num_ops_ == 3) {
                            if (auto c = dynamic_cast<LIRConstant*>(lirinst->get_operand(2))) {
                                if (c->value_ >= -1020 && c->value_ <= 1020) {
                                    break;
                                }
                                lirinst->remove_operand(2);
                                lirinst->set_operand(2,m_lir->get_LIRConstant(0));
                                lirinst->remove_operand(1);
                                lirinst->set_operand(1, m_lir->get_LIRPhysReg(13));
                                int val = std::abs(c->value_);
                                if (val <= 255 || val <= 1020 && val % 4 == 0 || val <= 4080 && val % 16 == 0 || val <= 16320 && val % 64 == 0 || val <= 65280 && val % 256 == 0 || val <= 261120 && val % 1024 == 0 || val <= 1044480 && val % 4096 == 0 || val <= 4177920 && val % 16384 == 0 || val <= 16711680 && val % 65536 == 0 || val <= 66846720 && val % 262114 == 0 || val <= 267386880 && val % 1048576 == 0 || val <= 1069547520 && val % 4194304 == 0) {
                                    addbefore.push_back(new LIRIntInst(LIRInstruction::LAddTy, LIRInstruction::LCondNone, m_lir->get_LIRPhysReg(13), m_lir->get_LIRPhysReg(14), m_lir->get_LIRConstant(c->value_)));
                                } else {
                                    addbefore.push_back(new LIRMov32Inst(LIRInstruction::LCondNone, m_lir->get_LIRPhysReg(13), m_lir->get_LIRConstant(c->value_)));
                                    addbefore.push_back(new LIRIntInst(LIRInstruction::LAddTy, LIRInstruction::LCondNone, m_lir->get_LIRPhysReg(13), m_lir->get_LIRPhysReg(14), m_lir->get_LIRPhysReg(13)));
                                }
                            }
                        }
                        break;
                    }
                    default: {
                        break;
                    }
                }
                addinst_map[lirinst] = addbefore;
            }
            for (const auto& pair : addinst_map) {
                for (auto inst : pair.second) {
                    lirbb->add_inst_before(inst, pair.first);
                }
            }
        }
    }
}
