%define parse.error verbose
%locations
%{
    #include "ast.h"
    #include <memory>
    #include <cstring>
    #include <stdarg.h>
    using namespace std;
    unique_ptr<COMPUnitAST> root; /* the top level root node of our final AST */

    //函数声明（可以改函数名字，函数内容在文件最下方）
    extern int yylineno;
	//yylloc.first_line=yylloc.last_line=yylineno
	
    extern int yylex();
	//声明由flex生成的词法分析器函数（应该不能改）
	
    extern void yyerror(const char *s);
	//报错函数
	
    extern void initFileName(char *name);
	//设置文件名字
	
    char filename[100];
	
%}

//定义联合体存储语义
%union {
    COMPUnitAST* compUnit;
    DECLDefAST* declDef;
    DECLAST* decl;
    DefListAST* defList;
    DEFAST* def;
    ArraysAST* arrays;
    InitValListAST* initValList;
    INITValAST* initVal;
    FUNCDefAST* funcDef;
    FuncFParamListAST* FuncFParamList;
    FUNCFParamAST* funcFParam;
    BLOCKAST* block;
    BlockItemListAST* blockItemList;
    BLOCKItemAST* blockItem;
    STMTAST* stmt;
    RETURNStmtAST* returnStmt;
    SELECTStmtAST* selectStmt;
    ITERATIONStmtAST* iterationStmt;
    LValAST* lVal;
    PRIMARYExpAST* primaryExp;
    NUMBERAST* number;
    UNARYExpAST* unaryExp;
    CALLAST* call;
    FuncCParamListAST* funcCParamList;
    MULExpAST* mulExp;
    ADDExpAST* addExp;
    RELExpAST* relExp;
    EQExpAST* eqExp;
    LANDExpAST* lAndExp;
    LORExpAST* lOrExp;

    TYPE ty;
    UOP op;
	string* token;
	int int_val;
	float float_val;
};


%type <compUnit> CompUnit;
%type <declDef> DeclDef;
%type <decl> Decl;
%type <defList> DefList;
%type <def> Def;
%type <arrays> Arrays;
%type <initValList> InitValList;
%type <initVal> InitVal;
%type <funcDef> FuncDef;
%type <FuncFParamList> FuncFParamList
%type <funcFParam> FuncFParam;
%type <block> Block;
%type <blockItemList> BlockItemList;
%type <blockItem> BlockItem;
%type <stmt> Stmt;
%type <returnStmt> ReturnStmt;
%type <selectStmt> SelectStmt;
%type <iterationStmt> IterationStmt;
%type <lVal> LVal;
%type <primaryExp> PrimaryExp;
%type <number> Number;
%type <unaryExp> UnaryExp;
%type <call> Call;
%type <funcCParamList> FuncCParamList;
%type <mulExp> MulExp;
%type <addExp> Exp AddExp;
%type <relExp> RelExp;
%type <eqExp> EqExp;
%type <lAndExp> LAndExp;
%type <lOrExp> Cond LOrExp;

%type <ty> BType VoidType
%type <op> UnaryOp

//% token 定义终结符的语义值类型
%token <int_val> INT           // 指定INT字面量的语义值是type_int，有词法分析得到的数值
%token <float_val> FLOAT       // 指定FLOAT字面量的语义值是type_float，有词法分析得到的数值
%token <token> ID            // 指定ID
%token greaterequal lessequal Greater Less equaL NOTequal    // 关系运算
%token INTNUM FLOATNUM VOIDTYPE  // 数据类型
%token CONSTANT RETURNstmt IF ELSE wHiLe BreAk ConTiNue
%token LParenthesis RParenthesis LBracket RBracket Lcurlybrace Rcurlybrace coMMa semiCOLON
//用bison对该文件编译时，带参数-d，生成的exp.tab.h中给这些单词进行编码，可在lex.l中包含parser.tab.h使用这些单词种类码
%token Not  Assign Minus PLUS Mul Div Mod And Or

%left Assign
%left Or And
%left equaL NOTequal
%left greaterequal lessequal Greater Less
%left PLUS Minus
%left Mod Mul Div
%right Not 

%nonassoc LOWER_THEN_ELSE
%nonassoc ELSE

%start Program

%%
Program:CompUnit {
          root = unique_ptr<COMPUnitAST>($1);
		}
		;

// 编译单元
CompUnit:CompUnit DeclDef {
            $$ = $1;
            $$->declDefList.push_back(unique_ptr<DECLDefAST>($2));
		}
		|DeclDef {
            $$ = new COMPUnitAST();
		    $$->declDefList.push_back(unique_ptr<DECLDefAST>($1));
		}
		;

//声明或者函数定义
DeclDef: Decl {
			$$ = new DECLDefAST();
			$$->Decl = unique_ptr<DECLAST>($1);
		}
		|FuncDef {
            $$ =  new DECLDefAST();
			$$->funcDef = unique_ptr<FUNCDefAST>($1);
		}
        ;

// 变量或常量声明
Decl:	CONSTANT BType DefList semiCOLON {
            $$ = new DECLAST();
            $$->isConst = true;
			$$->bType = $2;
			$$->defList.swap($3->list);
		}
        |BType DefList semiCOLON {
            $$ = new DECLAST();
            $$->isConst = false;
			$$->bType = $1;
			$$->defList.swap($2->list);
        }
		;

// 基本类型
BType:	INTNUM {
			$$ = TYPE_INT;
		}
		|FLOATNUM {
            $$ = TYPE_FLOAT;
        }
		;

// 定义列表
DefList:Def {
			$$ = new DefListAST();
            $$->list.push_back(unique_ptr<DEFAST>($1));
		}				
		|DefList coMMa Def {
			$$ = $1;
            $$->list.push_back(unique_ptr<DEFAST>($3));
		}
		;

// 定义
Def:	ID Arrays Assign InitVal {
			$$ = new DEFAST();
			$$->id = unique_ptr<string>($1);
			$$->arrays.swap($2->list);
			$$->initVal = unique_ptr<INITValAST>($4);
		}
        |ID Assign InitVal {
			$$ = new DEFAST();
			$$->id = unique_ptr<string>($1);
			$$->initVal = unique_ptr<INITValAST>($3);
        }
        |ID Arrays {
            $$ = new DEFAST();
            $$->id = unique_ptr<string>($1);
            $$->arrays.swap($2->list);
        }
        |ID {
            $$ = new DEFAST();
            $$->id = unique_ptr<string>($1);
        }
		;

// 数组
Arrays: LBracket Exp RBracket {
			$$ = new ArraysAST();
			$$->list.push_back(unique_ptr<ADDExpAST>($2));
		}
        |Arrays LBracket Exp RBracket {
			$$ = $1;
			$$->list.push_back(unique_ptr<ADDExpAST>($3));
		}
		;


// 变量或常量初值
InitVal: Exp {
			$$ = new INITValAST();
			$$->exp = unique_ptr<ADDExpAST>($1);
		}		
		|Lcurlybrace Rcurlybrace	{
			$$ = new INITValAST();
		}	
		|Lcurlybrace InitValList Rcurlybrace {
			$$ = new INITValAST();
			$$->initValList.swap($2->list);
		}	
		;

// 变量列表
InitValList:InitValList coMMa InitVal {
			$$ = $1;
			$$->list.push_back(unique_ptr<INITValAST>($3));
		}
		|InitVal {
			$$  = new InitValListAST();
			$$->list.push_back(unique_ptr<INITValAST>($1));
		}
		;

// 函数定义
FuncDef: BType ID LParenthesis FuncFParamList RParenthesis Block {
			$$ = new FUNCDefAST();
			$$->funcType = $1;
			$$->id = unique_ptr<string>($2);
			$$->funcFParamList.swap($4->list);
			$$->block = unique_ptr<BLOCKAST>($6);
		}
       	|BType ID LParenthesis RParenthesis Block {
			$$ = new FUNCDefAST();
			$$->funcType = $1;
			$$->id = unique_ptr<string>($2);
			$$->block = unique_ptr<BLOCKAST>($5);
		}
        |VoidType ID LParenthesis FuncFParamList RParenthesis Block {
			$$ = new FUNCDefAST();
			$$->funcType = $1;
			$$->id = unique_ptr<string>($2);
			$$->funcFParamList.swap($4->list);
			$$->block = unique_ptr<BLOCKAST>($6);
		}
       	|VoidType ID LParenthesis RParenthesis Block {
			$$ = new FUNCDefAST();
			$$->funcType = $1;
			$$->id = unique_ptr<string>($2);
			$$->block = unique_ptr<BLOCKAST>($5);
		}
		;

VoidType: VOIDTYPE {
			$$ = TYPE_VOID;
		}
		;

// 函数形参列表
FuncFParamList: FuncFParam {
				$$ = new FuncFParamListAST();
				$$->list.push_back(unique_ptr<FUNCFParamAST>($1));
			}	
			|FuncFParamList coMMa FuncFParam {
				$$ = $1;
				$$->list.push_back(unique_ptr<FUNCFParamAST>($3));
			}	
			;

// 函数形参
FuncFParam:	 BType ID {
				$$ = new FUNCFParamAST();
				$$->bType = $1;
				$$->id = unique_ptr<string>($2);
				$$->isArray = false;
			}
			|BType ID LBracket RBracket	{
				$$ = new FUNCFParamAST();
				$$->bType = $1;
				$$->id = unique_ptr<string>($2);
				$$->isArray = true;
			}
			|BType ID LBracket RBracket Arrays {
				$$ = new FUNCFParamAST();
				$$->bType = $1;
				$$->id = unique_ptr<string>($2);
				$$->isArray = true;
				$$->arrays.swap($5->list);
			}	
			;

// 语句块
Block:	 Lcurlybrace Rcurlybrace {
			$$ = new BLOCKAST();
		}
		|Lcurlybrace BlockItemList Rcurlybrace {
			$$ = new BLOCKAST();
			$$->blockItemList.swap($2->list);
		}	
		;

// 语句块项列表
BlockItemList:BlockItem	{
				$$ = new BlockItemListAST();
				$$->list.push_back(unique_ptr<BLOCKItemAST>($1));
			}
			|BlockItemList BlockItem {
				$$ = $1;
				$$->list.push_back(unique_ptr<BLOCKItemAST>($2));
			}		
			;

// 语句块项
BlockItem:	 Decl {
				$$ = new BLOCKItemAST();
				$$->decl = unique_ptr<DECLAST>($1);
			}
			|Stmt {
				$$ = new BLOCKItemAST();
				$$->stmt = unique_ptr<STMTAST>($1);
			}	
			;


Stmt:
 semiCOLON {
    $$ = new STMTAST();
    $$->sType = SEMI;
  }|
  LVal Assign Exp semiCOLON {
    $$ = new STMTAST();
    $$->sType = ASS;
    $$->lVal = unique_ptr<LValAST>($1);
    $$->exp = unique_ptr<ADDExpAST>($3);
  }|
  Exp semiCOLON {
    $$ = new STMTAST();
    $$->sType = EXP;
    $$->exp = unique_ptr<ADDExpAST>($1);
  }|
  ConTiNue semiCOLON {
    $$ = new STMTAST();
    $$->sType = CONT;
  }|
 BreAk semiCOLON {
    $$ = new STMTAST();
    $$->sType = BRE;
  }|
  Block {
    $$ = new STMTAST();
    $$->sType = BLK;
    $$->block = unique_ptr<BLOCKAST>($1);
  }|
  ReturnStmt {
    $$ = new STMTAST();
    $$->sType = RET;
    $$->returnStmt = unique_ptr<RETURNStmtAST>($1);
  }|
  SelectStmt {
    $$ = new STMTAST();
    $$->sType = SEL;
    $$->selectStmt = unique_ptr<SELECTStmtAST>($1);
  }|
  IterationStmt {
    $$ = new STMTAST();
    $$->sType = ITER;
    $$->iterationStmt = unique_ptr<ITERATIONStmtAST>($1);
  };



SelectStmt:  IF LParenthesis Cond RParenthesis Stmt %prec LOWER_THEN_ELSE{
	   $$ = new SELECTStmtAST();
	   $$-> cond = unique_ptr<LORExpAST>($3);
	   $$-> ifStmt = unique_ptr<STMTAST>($5);
	   $$->elseStmt = nullptr;
}|IF LParenthesis Cond RParenthesis Stmt ELSE Stmt{
	   $$ = new SELECTStmtAST();
	   $$-> cond = unique_ptr<LORExpAST>($3);
	   $$-> ifStmt = unique_ptr<STMTAST>($5);
	   $$-> elseStmt = unique_ptr<STMTAST>($7);
};
IterationStmt:  wHiLe LParenthesis Cond RParenthesis Stmt{
       $$ = new ITERATIONStmtAST();
	   $$-> cond = unique_ptr<LORExpAST>($3);
	   $$-> stmt = unique_ptr<STMTAST>($5);
};
ReturnStmt:  RETURNstmt semiCOLON{
	$$ = new RETURNStmtAST();
	$$->exp =nullptr;
}|RETURNstmt Exp semiCOLON{
    $$ = new RETURNStmtAST();
	$$-> exp = unique_ptr<ADDExpAST>($2);
};







// 表达式
Exp:	AddExp {
			$$ = $1;
		}
		;

// 条件表达式
Cond:	LOrExp {
			$$ = $1;
		}
		;

// 左值表达式
LVal:	ID {
			$$ = new LValAST();
			$$->id = unique_ptr<string>($1);
		}
		|ID Arrays {
			$$ = new LValAST();
			$$->id = unique_ptr<string>($1);
			$$->arrays.swap($2->list);
		}
		;

// 基本表达式
PrimaryExp:	 LParenthesis Exp RParenthesis {
				$$ = new PRIMARYExpAST();
				$$->exp = unique_ptr<ADDExpAST>($2);
				
			}
			|LVal {
				$$ = new PRIMARYExpAST();
				$$->lval = unique_ptr<LValAST>($1);
				
			}	
			|Number	{
				$$ = new PRIMARYExpAST();
				$$->number = unique_ptr<NUMBERAST>($1);
				
			}		
			;

// 数值
Number:	 INT {
			$$ = new NUMBERAST();
			$$->isInt = true;
			$$->intval = $1;
			
		}		
   		|FLOAT {
			$$ = new NUMBERAST();
			$$->isInt = false;
			$$->floatval = $1;
			
		}		
   		;

// 一元表达式
UnaryExp:	 PrimaryExp	{
				$$ = new UNARYExpAST();
				$$->primaryExp = unique_ptr<PRIMARYExpAST>($1);
				
			}					
			|Call {
				$$ = new UNARYExpAST();
				$$->call = unique_ptr<CALLAST>($1);
				
			}
			|UnaryOp UnaryExp {
				$$ = new UNARYExpAST();
				$$->op = $1;
				$$->unaryExp = unique_ptr<UNARYExpAST>($2);
			}		
			;

//函数调用
Call:ID LParenthesis RParenthesis {
		$$ = new CALLAST();
		$$->id = unique_ptr<string>($1);
		
	}
	 |ID LParenthesis FuncCParamList RParenthesis {
		$$ = new CALLAST();
		$$->id = unique_ptr<string>($1);
		$$->funcCParamList.swap($3->list);
		
	 }
	 ;	


UnaryOp: PLUS {
            $$ = UOP_ADD;
		}	
    	|Minus {
            $$ = UOP_MINUS;
		}
    	|Not {
            $$ = UOP_NOT;
		}	
		;


FuncCParamList: Exp {
				$$ = new FuncCParamListAST();
				$$->list.push_back(unique_ptr<ADDExpAST>($1));
			}
			|FuncCParamList coMMa Exp {
				$$ = (FuncCParamListAST*) $1;
				$$->list.push_back(unique_ptr<ADDExpAST>($3));
			}
			;
				
//乘除模表达式
MulExp:	 UnaryExp {
			$$ = new MULExpAST();
			$$->unaryExp = unique_ptr<UNARYExpAST>($1);
			
		}		
		|MulExp Mul UnaryExp {
			$$ = new MULExpAST();
			$$->mulExp = unique_ptr<MULExpAST>($1);
			$$->op = MOP_MUL;
			$$->unaryExp = unique_ptr<UNARYExpAST>($3);
			
		}	
		|MulExp Div UnaryExp {
			$$ = new MULExpAST();
			$$->mulExp = unique_ptr<MULExpAST>($1);
			$$->op = MOP_DIV;
			$$->unaryExp = unique_ptr<UNARYExpAST>($3);
			
		}	
		|MulExp Mod UnaryExp {
			$$ = new MULExpAST();
			$$->mulExp = unique_ptr<MULExpAST>($1);
			$$->op = MOP_MOD;
			$$->unaryExp = unique_ptr<UNARYExpAST>($3);
			
		}	
		;

// 加减表达式
AddExp:	 MulExp	{
			$$ = new ADDExpAST();
			$$->mulExp = unique_ptr<MULExpAST>($1);
			
		}			
		|AddExp PLUS MulExp {
			$$ = new ADDExpAST();
			$$->addExp = unique_ptr<ADDExpAST>($1);
			$$->op = AOP_ADD;
			$$->mulExp = unique_ptr<MULExpAST>($3);
			
		}
		|AddExp Minus MulExp {
			$$ = new ADDExpAST();
			$$->addExp = unique_ptr<ADDExpAST>($1);
			$$->op = AOP_MINUS;
			$$->mulExp = unique_ptr<MULExpAST>($3);
			
		}
		;

// 关系表达式
RelExp:	 AddExp	{
			$$ = new RELExpAST();
			$$->addExp = unique_ptr<ADDExpAST>($1);
			
		}				
		|RelExp greaterequal AddExp{
			$$ = new RELExpAST();
			$$->relExp = unique_ptr<RELExpAST>($1);
			$$->op = ROP_GTE;
			$$->addExp = unique_ptr<ADDExpAST>($3);
			
		}  //分析关系运算符号自身值保存在$2中
		|RelExp lessequal AddExp{
			$$ = new RELExpAST();
			$$->relExp = unique_ptr<RELExpAST>($1);
			$$->op = ROP_LTE;
			$$->addExp = unique_ptr<ADDExpAST>($3);
			
		}  //分析关系运算符号自身值保存在$2中
		|RelExp Greater AddExp {
			$$ = new RELExpAST();
			$$->relExp = unique_ptr<RELExpAST>($1);
			$$->op = ROP_GT;
			$$->addExp = unique_ptr<ADDExpAST>($3);
			
		}  //分析关系运算符号自身值保存在$2中
		|RelExp Less AddExp {
			$$ = new RELExpAST();
			$$->relExp = unique_ptr<RELExpAST>($1);
			$$->op = ROP_LT;
			$$->addExp = unique_ptr<ADDExpAST>($3);
			
		}  
		;

// 相等性表达式
EqExp:	 RelExp	{
			$$ = new EQExpAST();
			$$->relExp = unique_ptr<RELExpAST>($1);
			
		}				
		|EqExp equaL RelExp{
			$$ = new EQExpAST();
			$$->eqExp = unique_ptr<EQExpAST>($1);
			$$->op = EOP_EQ;
			$$->relExp = unique_ptr<RELExpAST>($3);
			
		} 	
		|EqExp NOTequal RelExp{
			$$ = new EQExpAST();
			$$->eqExp = unique_ptr<EQExpAST>($1);
			$$->op = EOP_NEQ;
			$$->relExp = unique_ptr<RELExpAST>($3);
			
		} 	
		;

// 逻辑与表达式
LAndExp: EqExp {
			$$ = new LANDExpAST();
			$$->eqExp = unique_ptr<EQExpAST>($1);
			
		}		
		|LAndExp And EqExp {
			$$ = new LANDExpAST();
			$$->lAndExp = unique_ptr<LANDExpAST>($1);
			$$->eqExp = unique_ptr<EQExpAST>($3);
		} 	
		;


LOrExp:	 LAndExp {
			$$ = new LORExpAST();
			$$->lAndExp = unique_ptr<LANDExpAST>($1);
			
		}				
		|LOrExp Or LAndExp {
			$$ = new LORExpAST();
			$$->lOrExp = unique_ptr<LORExpAST>($1);
			$$->lAndExp = unique_ptr<LANDExpAST>($3);
			
		} 	
		;
%%

void initFileName(char *name){
    strcpy(filename, name);
}
//将filename设置为文件名字

void yyerror(const char* fmt) {
    printf("%s:%d ", filename, yylloc.first_line);
    printf("%s\n", fmt);
}
//yyerror的具体内容 打印文件名 first_line在tokens.l中定义 （%s %d是占位符 可以改成c++格式？）
