%option noyywrap
%{
#include <string>
#include "ast.h"
#include "parser.hpp"

#define YY_USER_ACTION    	yylloc.first_line=yylloc.last_line=yylineno; \
	yylloc.first_column=yycolumn;	yylloc.last_column=yycolumn+yyleng-1; yycolumn+=yyleng;
%}
%option yylineno


ID [a-z_A-Z][a-z_A-Z0-9]*

INT ([1-9][0-9]*|0[0-7]*|(0x|0X)[0-9a-fA-F]+)
FLOAT_SUFFIX [fFlL]

DEC_EXP ([Ee][-+]?[0-9]+)
DEC_FLOAT_LIT_HAVE_POINT_HAVE_RIGHT ([0-9]*\.[0-9]+{DEC_EXP}?{FLOAT_SUFFIX}?)
DEC_FLOAT_LIT_HAVE_POINT_NO_RIGHT ([0-9]+\.{DEC_EXP}?{FLOAT_SUFFIX}?)
DEC_FLOAT_LIT_NO_POINT ([0-9]+{DEC_EXP}{FLOAT_SUFFIX}?)
DEC_FLOAT_LIT ({DEC_FLOAT_LIT_HAVE_POINT_HAVE_RIGHT}|{DEC_FLOAT_LIT_HAVE_POINT_NO_RIGHT}|{DEC_FLOAT_LIT_NO_POINT})

HEX_EXP ([Pp][-+]?[0-9]+)
HEX_FLOAT_LIT_HAVE_POINT_HAVE_RIGHT ((0[xX])[a-fA-F0-9]*\.[a-fA-F0-9]+{HEX_EXP}{FLOAT_SUFFIX}?)
HEX_FLOAT_LIT_HAVE_POINT_NO_RIGHT ((0[xX])[a-fA-F0-9]+\.{HEX_EXP}{FLOAT_SUFFIX}?)
HEX_FLOAT_LIT_NO_POINT ((0[xX])[a-fA-F0-9]+{HEX_EXP}{FLOAT_SUFFIX}?)
HEX_FLOAT_LIT ({HEX_FLOAT_LIT_HAVE_POINT_HAVE_RIGHT}|{HEX_FLOAT_LIT_HAVE_POINT_NO_RIGHT}|{HEX_FLOAT_LIT_NO_POINT})

FLOAT_LIT ({DEC_FLOAT_LIT}|{HEX_FLOAT_LIT})


SingleLineComment "//".*
MultilineComment "/*"([^\*]|(\*)*[^\*/])*(\*)*"*/"

%%

{INT}        {yylval.int_val = strtol(yytext,nullptr,0); return INT;}
{FLOAT_LIT}      {yylval.float_val = strtof(yytext,nullptr); return FLOAT;}

"int"        {return INTNUM;}
"float"      {return FLOATNUM;}
"void"       {return VOIDTYPE;}
"const"      {return CONSTANT;}
"return"     {return RETURNstmt;}
"if"         {return IF;}
"else"       {return ELSE;}
"while"      {return wHiLe;}
"break"      {return BreAk;}
"continue"   {return ConTiNue;}

{ID}    	{yylval.token = new string(yytext); return ID;}/*由于关键字的形式也符合表示符的规则，所以把关键字的处理全部放在标识符的前面，优先识别*/

">="        {return greaterequal;}
"<="        {return lessequal;}
">"         {return Greater;}
"<"         {return Less;}
"=="    	{return equaL;}
"!="        {return NOTequal;}

"("			{return LParenthesis;}
")"			{return RParenthesis;}
"["			{return LBracket;}
"]"			{return RBracket;}
"{"			{return Lcurlybrace;}
"}"			{return Rcurlybrace;}
","			{return coMMa;}
";"			{return semiCOLON;}

"!"			{return Not;}
"="			{return Assign;}
"-"			{return Minus;}
"+"			{return PLUS;}
"*"			{return Mul;}
"/"			{return Div;}
"%"			{return Mod;}
"&&"    	{return And;}
"||"    	{return Or;}

[\n]    	{yycolumn=1;}
[ \r\t] 	{}
{SingleLineComment} {}
{MultilineComment}	{}
.			{printf("Error type A :Mysterious character \"%s\"\n\t at Line %d\n",yytext,yylineno);}
%%



