#pragma once

#include "scope.h"
#include "irbuilder.h"
#include "../AST/ast.h"
#include <map>


class GenIR : public Visitor {
public:

    void visit(DECLDefAST &ast) override;
    void visit(DECLAST &ast) override;
    void visit(DEFAST &ast) override;
    void visit(INITValAST &ast) override;
    void visit(FUNCDefAST &ast) override;
    void visit(FUNCFParamAST &ast) override;
    void visit(COMPUnitAST &ast) override;
    void visit(BLOCKAST &ast) override;
    void visit(BLOCKItemAST &ast) override;
    void visit(STMTAST &ast) override;
    void visit(RETURNStmtAST &ast) override;
    void visit(SELECTStmtAST &ast) override;
    void visit(ITERATIONStmtAST &ast) override;
    void visit(ADDExpAST &ast) override;
    void visit(LValAST &ast) override;
    void visit(MULExpAST &ast) override;
    void visit(UNARYExpAST &ast) override;
    void visit(PRIMARYExpAST &ast) override;
    void visit(CALLAST&ast) override;
    void visit(NUMBERAST &ast) override;
    void visit(RELExpAST &ast) override;
    void visit(EQExpAST &ast) override;
    void visit(LANDExpAST &ast) override;
    void visit(LORExpAST &ast) override;

    
  IRBuilder *builder;
  Scope scope;
  std::unique_ptr<Module> module;

  GenIR() {
    module = std::unique_ptr<Module>(new Module());
    builder = new IRBuilder(nullptr, module.get());
    scope.enter();
    auto TyVoid = module->void_ty_;
    auto TyInt32 = module->int32_ty_;
    auto TyInt32Ptr = module->get_pointer_type(module->int32_ty_);
    auto TyFloat = module->float32_ty_;
    auto TyFloatPtr = module->get_pointer_type(module->float32_ty_);

    auto input_type = new FunctionType(TyInt32, {});
    auto get_int = new Function(input_type, "getint", module.get());
    scope.push("getint", get_int);
    input_type = new FunctionType(TyFloat, {});
    auto get_float = new Function(input_type, "getfloat", module.get());
   scope.push("getfloat", get_float);
    input_type = new FunctionType(TyInt32, {});
    auto get_char = new Function(input_type, "getch", module.get());
    scope.push("getch", get_char);
    std::vector<Type *> input_params;
    std::vector<Type *>().swap(input_params);
    input_params.push_back(TyInt32Ptr);
    input_type = new FunctionType(TyInt32, input_params);
    auto get_int_array = new Function(input_type, "getarray", module.get());
    scope.push("getarray", get_int_array);
    std::vector<Type *>().swap(input_params);
    input_params.push_back(TyFloatPtr);
    input_type = new FunctionType(TyInt32, input_params);
    auto get_float_array = new Function(input_type, "getfarray", module.get());
    scope.push("getfarray", get_float_array);
    std::vector<Type *> output_params;
    std::vector<Type *>().swap(output_params);
    output_params.push_back(TyInt32);
    auto output_type = new FunctionType(TyVoid, output_params);
    auto put_int = new Function(output_type, "putint", module.get());
    scope.push("putint", put_int);
    std::vector<Type *>().swap(output_params);
    output_params.push_back(TyFloat);
    output_type = new FunctionType(TyVoid, output_params);
    auto put_float = new Function(output_type, "putfloat", module.get());
    scope.push("putfloat", put_float);
    std::vector<Type *>().swap(output_params);
    output_params.push_back(TyInt32);
    output_type = new FunctionType(TyVoid, output_params);
    auto put_char = new Function(output_type, "putch", module.get());
    scope.push("putch", put_char);
    std::vector<Type *>().swap(output_params);
    output_params.push_back(TyInt32);
    output_params.push_back(TyInt32Ptr);
    output_type = new FunctionType(TyVoid, output_params);
    auto put_int_array = new Function(output_type, "putarray", module.get());
    scope.push("putarray", put_int_array);
    std::vector<Type *>().swap(output_params);
    output_params.push_back(TyInt32);
    output_params.push_back(TyFloatPtr);
    output_type = new FunctionType(TyVoid, output_params);
    auto put_float_array = new Function(output_type, "putfarray", module.get());
    scope.push("putfarray", put_float_array);
    output_params.clear();
    auto time_type = new FunctionType(TyVoid, output_params);
    auto sysy_start_time = new Function(time_type, "_sysy_starttime", module.get());
    auto sysy_stop_time = new Function(time_type, "_sysy_stoptime", module.get());
    scope.push("starttime", sysy_start_time);
    scope.push("stoptime", sysy_stop_time);
    output_params.clear();
    output_params.push_back(TyInt32Ptr);
    output_params.push_back(TyInt32Ptr);
    output_params.push_back(TyInt32);
    output_type = new FunctionType(TyVoid, output_params);
    auto memcpy = new Function(output_type, "__aeabi_memcpy4", module.get());
    scope.push("memcpy", memcpy);
    output_params.clear();
    output_params.push_back(TyInt32Ptr);
    output_params.push_back(TyInt32);
    output_type = new FunctionType(TyVoid, output_params);
    auto memclr = new Function(output_type, "__aeabi_memclr4", module.get());
    scope.push("memclr", memclr);
    output_params.push_back(TyInt32);
    output_type = new FunctionType(TyVoid, output_params);
    auto memset = new Function(output_type, "__aeabi_memset4", module.get());
    scope.push("memset", memset);
    output_params.clear();
    output_type = new FunctionType(TyVoid, output_params);
    auto llvm_memset = new Function(output_type, "llvm.memset.p0.i32", module.get());
    scope.push("llvm.memset.p0.i32", llvm_memset);
  }
  std::unique_ptr<Module> getModule() { return std::move(module); }

  void checkInitType() const;

  static int getNextDim(vector<int> &dimensionsCnt, int up, int cnt);

  void localInit(Value *ptr, vector<unique_ptr<INITValAST>> &list,
                 vector<int> &dimensionsCnt, int up);

  static int getNextDim(vector<int> &elementsCnts, int up);

  ConstantArray *globalInit(vector<int> &dimensions, vector<ArrayType *> &arrayTys, int up, vector<unique_ptr<INITValAST>> &list);

  static void mergeElements(vector<int> &dimensions,
                            vector<ArrayType *> &arrayTys, int up, int dimAdd, vector<Constant *> &elements, vector<int> &elementsCnts);

  void finalMerge(vector<int> &dimensions, vector<ArrayType *> &arrayTys, int up, vector<Constant *> &elements, vector<int> &elementsCnts) const;

  bool checkCalType(Value **val, int *intVal, float *floatVal);

  void checkCalType(Value **val);
};
