#include "irbuilder.h"

BinaryInst* IRBuilder::create_binary(Instruction::OpID op, Value* v1, Value* v2) {
    return new BinaryInst(this->m_->int32_ty_, op, v1, v2, this->BB_);
}


ICmpInst* IRBuilder::create_icmp(ICmpInst::ICmpOp op, Value* v1, Value* v2) {
    return new ICmpInst(op, v1, v2, this->BB_);
}

BinaryInst* IRBuilder::create_fbinary(Instruction::OpID op, Value* v1, Value* v2) {
    return new BinaryInst(this->m_->float32_ty_, op, v1, v2, this->BB_);
}

FCmpInst* IRBuilder::create_fcmp(FCmpInst::FCmpOp op, Value* v1, Value* v2) {
    return new FCmpInst(op, v1, v2, this->BB_);
}

  CallInst *IRBuilder::create_call(Value *func, std::vector<Value *> args) {

    return new CallInst(static_cast<Function *>(func), args, this->BB_);
  }

  BranchInst *IRBuilder::create_br(BasicBlock *if_true) {
    return new BranchInst(if_true, this->BB_);
  }
  BranchInst *IRBuilder::create_cond_br(Value *cond, BasicBlock *if_true,
                             BasicBlock *if_false) {
    return new BranchInst(cond, if_true, if_false, this->BB_);
  }

  ReturnInst *IRBuilder::create_ret(Value *val) { return new ReturnInst(val, this->BB_); }
  ReturnInst *IRBuilder::create_void_ret() { return new ReturnInst(this->BB_); }

  GetElementPtrInst *IRBuilder::create_gep(Value *ptr, std::vector<Value *> idxs) {
    return new GetElementPtrInst(ptr, idxs, this->BB_);
  }

  StoreInst *IRBuilder::create_store(Value *val, Value *ptr) {
    return new StoreInst(val, ptr, this->BB_);
  }
  LoadInst *IRBuilder::create_load(Type *ty, Value *ptr) {
    return new LoadInst(ptr, this->BB_);
  }
  LoadInst *IRBuilder::create_load(Value *ptr) {

    return new LoadInst(ptr, this->BB_);
  }

  AllocaInst *IRBuilder::create_alloca(Type *ty) { return new AllocaInst(ty, this->BB_); }
  ZextInst *IRBuilder::create_zext(Value *val, Type *ty) {
    return new ZextInst(Instruction::ZExt, val, ty, this->BB_);
  }
  FpToSiInst *IRBuilder::create_fptosi(Value *val, Type *ty) {
    return new FpToSiInst(Instruction::FPtoSI, val, ty, this->BB_);
  }
  SiToFpInst *IRBuilder::create_sitofp(Value *val, Type *ty) {
    return new SiToFpInst(Instruction::SItoFP, val, ty, this->BB_);
  }
  Bitcast *IRBuilder::create_bitcast(Value *val, Type *ty) {
    return new Bitcast(Instruction::BitCast, val, ty, this->BB_);
  }