#ifndef IRBUILDER_H
#define IRBUILDER_H
#include "llvm.h"


class IRBuilder {
public:
  BasicBlock *BB_;
  Module *m_;

  IRBuilder(BasicBlock *bb, Module *m) : BB_(bb), m_(m){};
  ~IRBuilder() = default;
  Module *get_module() { return m_; }
  BasicBlock *get_insert_block() { return this->BB_; }
  void set_insert_point(BasicBlock *bb) {
    this->BB_ = bb;
  }
  BinaryInst* create_binary(Instruction::OpID op, Value* v1, Value* v2) ;
  ICmpInst* create_icmp(ICmpInst::ICmpOp op, Value* v1, Value* v2) ;
  BinaryInst* create_fbinary(Instruction::OpID OP, Value* v1, Value* v2) ;
  FCmpInst* create_fcmp(FCmpInst::FCmpOp op, Value* v1, Value* v2) ;
  CallInst *create_call(Value *func, std::vector<Value *> args) ;

  BranchInst *create_br(BasicBlock *if_true) ;
  BranchInst *create_cond_br(Value *cond, BasicBlock *if_true, BasicBlock *if_false) ;

  ReturnInst *create_ret(Value *val) ;
  ReturnInst *create_void_ret() ;

  GetElementPtrInst *create_gep(Value *ptr, std::vector<Value *> idxs) ;

  StoreInst *create_store(Value *val, Value *ptr) ;
  LoadInst *create_load(Type *ty, Value *ptr) ;
  LoadInst *create_load(Value *ptr) ;

  AllocaInst *create_alloca(Type *ty) ;
  ZextInst *create_zext(Value *val, Type *ty) ;
  FpToSiInst *create_fptosi(Value *val, Type *ty) ;
  SiToFpInst *create_sitofp(Value *val, Type *ty) ;
  Bitcast *create_bitcast(Value *val, Type *ty) ;
};

#endif
