#include "llvm.h"


std::map<Instruction::OpID, std::string> instr_id2string_ = {
    {Instruction::Ret, "ret"},{Instruction::Br, "br"},{Instruction::FNeg, "fneg"},{Instruction::Add, "add"},{Instruction::Sub, "sub"},{Instruction::Mul, "mul"},{Instruction::SDiv, "sdiv"},
    {Instruction::SRem, "srem"},{Instruction::UDiv, "udiv"},{Instruction::URem, "urem"},{Instruction::FAdd, "fadd"},{Instruction::FSub, "fsub"},{Instruction::FMul, "fmul"},{Instruction::FDiv, "fdiv"},
    {Instruction::Shl, "shl"},{Instruction::LShr, "lshr"},{Instruction::AShr, "ashr"},{Instruction::And, "and"},{Instruction::Or, "or"},{Instruction::Xor, "xor"},{Instruction::Alloca, "alloca"},
    {Instruction::Load, "load"},{Instruction::Store, "store"},{Instruction::GetElementPtr, "getelementptr"},{Instruction::ZExt, "zext"},{Instruction::FPtoSI, "fptosi"},{Instruction::SItoFP, "sitofp"},
    {Instruction::BitCast, "bitcast"},{Instruction::ICmp, "icmp"},{Instruction::FCmp, "fcmp"},{Instruction::PHI, "phi"},{Instruction::Call, "call"}}; 

std::string print_as_op( Value *v, bool print_ty ) 
{
    std::string op_ir;
    std::stringstream ss;
    if( print_ty ) {
        ss<< v->type->print()<<" ";
       
    }

    if (dynamic_cast<GlobalVariable *>(v)) {
     
       ss<<  "@"<< v->name_;
    }
    else if (dynamic_cast<Function *>(v) ) {
      
        ss<< "@"<<v->name_;
    }
    else if (dynamic_cast<Constant *>(v)) {
        ss<< v->print();;
    }
    else {
        ss<<"%"<<v->name_;
        
    }
    op_ir=ss.str();
    return op_ir;
}

std::string print_fcmp_type( FCmpInst::FCmpOp op ) 
{
    switch (op)
    {
        case FCmpInst::FCMP_UGE:
            return "uge";
            break;
        case FCmpInst::FCMP_UGT:
            return "ugt";
            break;
        case FCmpInst::FCMP_ULE:
            return "ule";
            break;
        case FCmpInst::FCMP_ULT:
            return "ult";
            break;
        case FCmpInst::FCMP_UEQ:
            return "ueq";
            break;
        case FCmpInst::FCMP_UNE:
            return "une";
            break;
        default:
            break;
    }
    return "";
}

std::string print_cmp_type( ICmpInst::ICmpOp op ) 
{
    switch (op)
    {
        case ICmpInst::ICMP_SGE:
            return "sge";
            break;
        case ICmpInst::ICMP_SGT:
            return "sgt";
            break;
        case ICmpInst::ICMP_SLE:
            return "sle";
            break;
        case ICmpInst::ICMP_SLT:
            return "slt";
            break;
        case ICmpInst::ICMP_EQ:
            return "eq";
            break;
        case ICmpInst::ICMP_NE:
            return "ne";
            break;
        default:
            break;
    }
    return "";
}

std::string fcmpname (FCmpInst::FCmpOp id) {
    switch(id) {
        case FCmpInst::FCmpOp::FCMP_FALSE:
            return "nv";
            break;
        case FCmpInst::FCmpOp::FCMP_OEQ:
            return "eq";
            break;
        case FCmpInst::FCmpOp::FCMP_OGT:
            return "gt";
            break;
        case FCmpInst::FCmpOp::FCMP_OGE:
            return "ge";
            break;
        case FCmpInst::FCmpOp::FCMP_OLT:
            return "cc";
            break;
        case FCmpInst::FCmpOp::FCMP_OLE:
            return "ls";
            break;
        case FCmpInst::FCmpOp::FCMP_ONE:
            return "ne";
            break;
        case FCmpInst::FCmpOp::FCMP_ORD:
            return "vc";
            break;
        case FCmpInst::FCmpOp::FCMP_UNO:
            return "vs";
            break;
        case FCmpInst::FCmpOp::FCMP_UEQ:
            return "eq";
            break;
        case FCmpInst::FCmpOp::FCMP_UGT:
            return "hi";
            break;
        case FCmpInst::FCmpOp::FCMP_UGE:
            return "cs";
            break;
        case FCmpInst::FCmpOp::FCMP_ULT:
            return "lt";
            break;
        case FCmpInst::FCmpOp::FCMP_ULE:
            return "le";
            break;
        case FCmpInst::FCmpOp::FCMP_UNE:
            return "ne";
            break;
        case FCmpInst::FCmpOp::FCMP_TRUE:
            return "al";
            break;
        default:
            break;
    }
    return "";
}

std::string icmpname (ICmpInst::ICmpOp id) {
    switch(id) {
        case ICmpInst::ICmpOp::ICMP_EQ:
            return "eq";
            break;
        case ICmpInst::ICmpOp::ICMP_UGE:
            return "cs";
            break;

         case ICmpInst::ICmpOp::ICMP_NE:
            return "ne";
            break;
        case ICmpInst::ICmpOp::ICMP_UGT:
            return "hi";
            break;
        case ICmpInst::ICmpOp::ICMP_ULT:
            return "cc";
            break;
        case ICmpInst::ICmpOp::ICMP_ULE:
            return "ls";
            break;
        case ICmpInst::ICmpOp::ICMP_SGT:
            return "gt";
            break;
        case ICmpInst::ICmpOp::ICMP_SGE:
            return "ge";
            break;
        case ICmpInst::ICmpOp::ICMP_SLT:
            return "lt";
            break;
        case ICmpInst::ICmpOp::ICMP_SLE:
            return "le";
            break;
        default:
            break;
    }
    return "";
}

std::string GlobalVariable::print()
{
    std::string global_val_ir;
    std::stringstream ss;
    ss<<print_as_op(this, false)<<" = "<<(this->is_const_ ? "constant " : "global ")<< static_cast<PointerType *>(this->type)->contained_->print()<<" "<< this->init_val_->print();
   global_val_ir=ss.str();
    return global_val_ir;
}

std::string ConstantInt::print()
{
    std::string const_ir;
    std::stringstream ss;
    if (this->type->tid_ == Type::IntegerTyID && static_cast<IntegerType *>(this->type)->num_bits_ == 1 ) {
       
        ss<<((this->value_ == 0) ? "0" : "1");
    }
    else 
        ss<< std::to_string(this->value_);
        const_ir=ss.str();
    return const_ir;
}

std::string ConstantFloat::print()
{
    std::stringstream ss;
    std::string fp_ir;
    double val = this->value_;
    ss << "0x"<< std::hex << *(uint64_t *)&val << std::endl;
    ss >> fp_ir;
    return fp_ir;
}

std::string ConstantArray::print()
{
    std::string const_ir;
    std::stringstream ss;
    ss<<"["<<static_cast<ArrayType *>(this->type)->contained_->print()<<" "<<const_array[0]->print();
  
    for (size_t i = 1 ; i < this->const_array.size() ; i++ ){
ss<<", "<<static_cast<ArrayType *>(this->type)->contained_->print()<<" "<<const_array[i]->print();
     
    }
   ss<< "]";
   const_ir=ss.str();;
    return const_ir;
}

std::string ConstantZero::print()
{
    return "zeroinitializer";
}

std::string BasicBlock::print()
{
    std::string bb_ir;
    std::stringstream ss;
    ss<<this->name_<<":";
   
  
    if(!this->pre_bbs_.empty()) {
        ss<< "                                                ; preds = ";
    }
    for (auto bb : this->pre_bbs_) {
        if( bb != *this->pre_bbs_.begin() )
            ss<< ", ";
        ss<<print_as_op(bb, false);
        
    }
   
    if ( !this->parent_) {
        ss<<"\n"<< "; Error: Block without parent!";
       
    }
    ss<< "\n";
    for (auto instr : this->instr_list_) {
        ss<< "  "<< instr->print()<<"\n";
        
    }
    bb_ir=ss.str();;
    return bb_ir;
}

Instruction *BasicBlock::get_terminator() {
    if (instr_list_.empty())
        return nullptr;
    switch (instr_list_.back()->op_id_) {
        case Instruction::Ret:
        case Instruction::Br:
            return instr_list_.back();
        default:
            return nullptr;
    }
}

bool BasicBlock::delete_instr(Instruction *instr) {
    if((!instr)||instr->pos_in_bb.size()!=1||instr->parent_!= this)
        return false;
    this->instr_list_.erase(instr->pos_in_bb.back());
    instr->remove_use_of_ops();
    instr->pos_in_bb.clear();
    instr->parent_ = nullptr;
    return true;
}

bool BasicBlock::add_instruction(Instruction *instr) {
    if(instr->pos_in_bb.size()!=0){ 
        return false;
    }else{
        instr_list_.push_back(instr);
        std::list<Instruction*>::iterator tail = instr_list_.end();
        instr->pos_in_bb.emplace_back(--tail);
        instr->parent_ = this;
        return true;
    }
}
bool BasicBlock::add_instruction_front(Instruction *instr) {
    if(instr->pos_in_bb.size()!=0){ 
        return false;
    }else{
        instr_list_.push_front(instr);
        std::list<Instruction*>::iterator head = instr_list_.begin();
        instr->pos_in_bb.emplace_back(head);
        instr->parent_ = this;
        return true;
    }
}

bool BasicBlock::add_instruction_before_terminator(Instruction *instr) {
    if(instr->pos_in_bb.size()!=0){ 
        return false;
    }else if(instr_list_.empty()){ 
        return false;
    }else{
        auto it = std::end(instr_list_);
        instr_list_.emplace(--it, instr);
        instr->pos_in_bb.emplace_back(--it);
        instr->parent_ = this;
        return true;
    }
}
bool BasicBlock::add_instruction_before_inst(Instruction* new_instr,Instruction* instr) {
    if((!instr)||instr->pos_in_bb.size()!=1||instr->parent_!= this)
        return false;
    if(new_instr->pos_in_bb.size()!=0) 
        return false;
    else if(instr_list_.empty()) 
        return false;
    else{
        auto it = instr->pos_in_bb.back();
        instr_list_.emplace( it, new_instr); 
        new_instr->pos_in_bb.emplace_back(--it); 
        new_instr->parent_ = this;
        return true;
    }
}
bool BasicBlock::remove_instr(Instruction *instr) {
    
    if((!instr)||instr->pos_in_bb.size()!=1||instr->parent_!= this)
        return false;
    this->instr_list_.erase(instr->pos_in_bb.back());
    instr->pos_in_bb.clear(); 
    instr->parent_ = nullptr;
    return true;
}
std::string BinaryInst::print()
{
    std::stringstream ss;
    std::string instr_ir;
    ss<<"%" <<this->name_ << " = " << instr_id2string_[this->op_id_] << " " << this->operands_[0]->type->print() << " ";
    ss<< print_as_op(this->get_operand(0), false) << ", ";
    ss<< print_as_op(this->get_operand(1), false);
    instr_ir=ss.str();
    return instr_ir;
}
std::string UnaryInst::print()
{
    std::stringstream ss;
    std::string instr_ir;
    ss<< "%"<< this->name_<<" = "<<instr_id2string_[this->op_id_]<<" ";
   ss<< this->operands_[0]->type->print()<<" "<< print_as_op(this->get_operand(0), false);
    
    switch (this->op_id_) {
        case Instruction::ZExt:
            instr_ir += " to i32";
            break;
        case Instruction::FPtoSI:
            instr_ir += " to i32";
            break;
        case Instruction::SItoFP:
            instr_ir += " to float";
            break;
        default:
            break;
    }
    instr_ir=ss.str();;
    return instr_ir;
}

std::string ICmpInst::print()
{
    std::string instr_ir;
    std::stringstream ss;
    ss<<"%"<<this->name_<< " = "<<instr_id2string_[this->op_id_] <<" "<<print_cmp_type(this->icmp_op_)<<" ";
    ss<<this->get_operand(0)->type->print()<<" "<<print_as_op(this->get_operand(0), false)<<", ";
    if (this->get_operand(0)->type->tid_ == this->get_operand(1)->type->tid_) {
       
       ss<<print_as_op(this->get_operand(1), false);
    }
    else {
       ss<<print_as_op(this->get_operand(1), true);
    }
    instr_ir=ss.str();
    return instr_ir;
}

std::string FCmpInst::print()
{
    std::string instr_ir;
    std::stringstream ss;
    ss<< "%" << this->name_ << " = " << instr_id2string_[this->op_id_] + " ";
    ss<< print_fcmp_type(this->fcmp_op_) <<" " << this->get_operand(0)->type->print() << " ";
    ss<< print_as_op(this->get_operand(0), false) << ", ";
    if (this->get_operand(0)->type->tid_ == this->get_operand(1)->type->tid_)
    {
        ss<< print_as_op(this->get_operand(1), false);
    }
    else
    {
        ss<< print_as_op(this->get_operand(1), true);
    }
    instr_ir=ss.str();;
    return instr_ir;
}

std::string CallInst::print()
{
    std::string instr_ir;
    std::stringstream ss;
    if(!(this->type->tid_ == Type::VoidTyID))
    {
        ss<< "%"<< this->name_<<" = ";
    }
    ss<< instr_id2string_[this->op_id_]<< " ";
    unsigned int numops = this->num_ops_;
    ss<< static_cast<FunctionType *>(this->get_operand(numops-1)->type)->result_->print()<< " ";
    
    if (dynamic_cast<Function *>(this->get_operand(numops - 1))->name_ == "__aeabi_memclr4") {
        ss<< "@llvm.memset.p0.i32(";
       
        ss<< this->get_operand(0)->type->print()<< " "<<print_as_op(this->get_operand(0), false);
       
        ss<< ", i8 0, ";
      
        ss<<this->get_operand(1)->type->print()<< " "<<print_as_op(this->get_operand(1), false);
        
        ss<< ", i1 false)";
        instr_ir=ss.str();;
        return instr_ir;
    }

    ss << print_as_op(this->get_operand(numops - 1), false)<<"(";
    for (unsigned int i = 0; i < numops - 1; i++) {
        if( i > 0 )
        ss<< ", "<< this->get_operand(i)->type->print()<<" "<< print_as_op(this->get_operand(i), false);
    }
    ss<< ")";
    instr_ir=ss.str();;
    return instr_ir;
}

std::string BranchInst::print()
{
    std::string instr_ir;
    std::stringstream ss;
    ss<< instr_id2string_[this->op_id_]<< " "<< print_as_op(this->get_operand(0), true);
    if( this->num_ops_ == 3 ) {
        ss<< ", "<<print_as_op(this->get_operand(1), true);
        ss<< ", "<< print_as_op(this->get_operand(2), true);
    }
    instr_ir=ss.str();
    return instr_ir;
}

std::string ReturnInst::print()
{
    std::stringstream ss;
    std::string instr_ir;
    ss<< instr_id2string_[this->op_id_]<< " ";
    if ( this->num_ops_ != 0 ) {
        ss<< this->get_operand(0)->type->print()<< " ";
        ss<< print_as_op(this->get_operand(0), false);
    }
    else {
        ss<< "void";
    }
    instr_ir=ss.str();
    return instr_ir;
}

std::string GetElementPtrInst::print()
{
    std::string instr_ir;
    std::stringstream ss;
    ss<< "%"<< this->name_<< " = "<< instr_id2string_[this->op_id_]<<" ";
    ss<<static_cast<PointerType *>(this->get_operand(0)->type)->contained_->print()<<", ";
    for (unsigned int i = 0; i < this->num_ops_; i++)
    {
        if( i > 0 )
        ss<< ", "<< this->get_operand(i)->type->print();
        ss<< " "<<print_as_op(this->get_operand(i), false);
    }
    instr_ir=ss.str();
    return instr_ir;
}

std::string StoreInst::print()
{
    std::string instr_ir;
    std::stringstream ss;
    ss<<instr_id2string_[this->op_id_]<<" ";
    ss<<this->get_operand(0)->type->print()<<" ";
    ss<<print_as_op(this->get_operand(0), false)<<", ";
    ss<<print_as_op(this->get_operand(1), true);
    instr_ir=ss.str();
    return instr_ir;
}

std::string LoadInst::print()
{
    std::string instr_ir;
    std::stringstream ss;
    ss<<"%"<< this->name_<< " = "<<instr_id2string_[this->op_id_]<<" ";
    ss<<static_cast<PointerType *>(this->get_operand(0)->type)->contained_->print();
    ss<< ","<<" "<< print_as_op(this->get_operand(0), true);
    instr_ir=ss.str();
    return instr_ir;
}

std::string AllocaInst::print()
{
    std::string instr_ir;
    std::stringstream ss;
    ss<< "%"<< this->name_<<" = "<< instr_id2string_[this->op_id_];
    ss<< " "<<alloca_ty_->print();
    instr_ir=ss.str();
    return instr_ir;
}

std::string ZextInst::print()
{
    std::string instr_ir;
    std::stringstream ss;
    ss<<"%"<< this->name_<< " = "<<instr_id2string_[this->op_id_];
    ss<<" "<<this->get_operand(0)->type->print()<<" ";
    ss<< print_as_op(this->get_operand(0), false)<< " to "<<this->dest_ty_->print();
    instr_ir=ss.str();
    return instr_ir;
}

std::string FpToSiInst::print()
{
    std::string instr_ir;
    std::stringstream ss;
    ss<< "%"<< this->name_<< " = "<< instr_id2string_[this->op_id_]<< " ";
    ss<<this->get_operand(0)->type->print()<<" "<<print_as_op(this->get_operand(0), false);
    ss<< " to "<< this->dest_ty_->print();
    instr_ir=ss.str();
    return instr_ir;
}

std::string SiToFpInst::print()
{
    std::string instr_ir;
    std::stringstream ss;
    ss<< "%"<<this->name_;
    ss<< " = "<< instr_id2string_[this->op_id_]<< " ";
    ss<<this->get_operand(0)->type->print()<< " ";
    ss<<print_as_op(this->get_operand(0), false)<< " to ";
    ss<< this->dest_ty_->print();
    instr_ir=ss.str();
    return instr_ir;
}

std::string Bitcast::print() {
    std::string instr_ir;
    std::stringstream ss;
    ss<<"%"<<this->name_<<" = "<< instr_id2string_[this->op_id_];
    ss<<" "<< this->get_operand(0)->type->print()<< " ";
    ss<< print_as_op(this->get_operand(0), false)<<" to ";
    ss<<this->dest_ty_->print();
    instr_ir=ss.str();
    return instr_ir;
}

std::string PhiInst::print()
{
    std::string instr_ir;
    std::stringstream ss;
    ss<<"%"<<this->name_<<" = "<< instr_id2string_[this->op_id_]<< " ";
   ss<< this->get_operand(0)->type->print()<< " ";
    for (int i = 0; i < this->num_ops_ / 2; i++) {
        if( i > 0 )
            ss<< ", "<<"[ "<< print_as_op(this->get_operand(2*i), false);
        ss<<", "<<print_as_op(this->get_operand(2*i+1), false)<<" ]";
    }
    if ( this->num_ops_ / 2 < this->parent_->pre_bbs_.size() ) {
        for ( auto pre_bb : this->parent_->pre_bbs_ ) {
            if (std::find(this->operands_.begin(), this->operands_.end(), static_cast<Value *>(pre_bb)) == this->operands_.end()) {
              
                ss<< ", [ undef, " <<print_as_op(pre_bb, false)<<" ]";
            }
        }
    }
    instr_ir=ss.str();
    return instr_ir; 
}

void Function::set_instr_name() {
    std::map<Value*, int> seq;
    for (auto arg : this->arguments_) {
        if (!seq.count(arg)) {
            auto seq_num = seq.size() + seq_cnt_;
            if (arg->name_ == "") {
                arg->name_ = "arg_" + std::to_string(seq_num);
                seq.insert({ arg, seq_num });
            }
        }
    }
    for (auto bb : basic_blocks_) {
        if (!seq.count(bb)) {
            auto seq_num = seq.size() + seq_cnt_;
            if (bb->name_.length() <= 6 || bb->name_.substr(0, 6) != "label_") {
                bb->name_ = "label_" + std::to_string(seq_num);
                seq.insert({ bb, seq_num });
            }
        }
        for (auto instr : bb->instr_list_) {
            if (instr->type->tid_ != Type::VoidTyID && !seq.count(instr)) {
                auto seq_num = seq.size() + seq_cnt_;
                if (instr->name_ == "") {
                    instr->name_ = "v" + std::to_string(seq_num);
                    seq.insert({ instr, seq_num });
                }
            }
        }
    }
    seq_cnt_ += seq.size();
}
std::string Function::print()
{
    if (this->name_ == "llvm.memset.p0.i32") {
        std::string func_ir = "declare void @llvm.memset.p0.i32(i32*, i8, i32, i1)";
        return func_ir;
    }
    set_instr_name();
    std::string func_ir;
    std::stringstream func_ss;
    if ( this->is_declaration() )
        func_ss<<"declare ";
    else
        func_ss<<"define ";

    func_ss<< this->get_return_type()->print()<<" ";
    func_ss<< print_as_op(this, false)<< "(";

   
    if (this->is_declaration()) {
        for (size_t i = 0 ; i < this->arguments_.size() ; i ++ ) {
            if(i)
              {  func_ss<< ", ";}
            func_ss<< static_cast<FunctionType *>(this->type)->args_[i]->print();
        }
    }
    else {
        for (auto arg = this->arguments_.begin(); arg != arguments_.end() ; arg++ ) {
            if( arg != this->arguments_.begin()) {
                func_ss << ", ";
            }
            func_ss<< static_cast<Argument *>(*arg)->print();
        }
    }
   func_ss<<")";
    
    if(!this->is_declaration()) {
        func_ss<< " {"<< "\n";
        for ( auto bb : this->basic_blocks_) {
            func_ss<< bb->print();
        }
        func_ss<<"}";
    }
    func_ir=func_ss.str();
    return func_ir;
}

std::string Argument::print()
{
    std::string arg_ir;
    std::stringstream ss;
    ss<< this->type->print()<< " %"<<this->name_;
    arg_ir=ss.str();
    return arg_ir;
}

void Function::remove_bb(BasicBlock* bb)
{ 
    basic_blocks_.erase(std::remove(basic_blocks_.begin(), basic_blocks_.end(), bb) , basic_blocks_.end());
    for (auto pre : bb->pre_bbs_) {
        pre->remove_succ_basic_block(bb);
    }
    for (auto succ : bb->succ_bbs_) {
        succ->remove_pre_basic_block(bb);
    }
}

BasicBlock *Function::getRetBB() {
    for (auto bb : basic_blocks_) {
        if (bb->get_terminator()->op_id_ == Instruction::Ret) {
            return bb;
        }
    }
    return nullptr;
}
std::string Module::print()
{
    std::string module_ir;
    std::stringstream ss;
    for ( auto global_val : this->global_list_) {
        ss<< global_val->print()<<"\n";
    }
    for ( auto func : this->function_list_) {
        ss<< func->print()<<"\n";
    }
    module_ir=ss.str();
    return module_ir;
}

Function *Module::getMainFunc() {
    for (auto f : function_list_) {
        if (f->name_ == "main") {
            return f;
        }
    }
    return nullptr;
}


