#ifndef TYPE_H
#define TYPE_H
#include <iostream>
#include <string>
#include <vector>
#include <list>


class Type {
public:
  enum TypeID {
    VoidTyID,     
    LabelTyID,    
    IntegerTyID,  
    FloatTyID,    
    FunctionTyID, 
    ArrayTyID, 
    PointerTyID, 
  };
  TypeID tid_;
   bool is_const = false;
  explicit Type(TypeID tid) : tid_(tid) {}
  ~Type() = default;
  bool is_Void() const { return tid_ == VoidTyID; };
    bool is_Var_Int() const { return tid_ == IntegerTyID; };
    bool is_Var_Float() const { return tid_ == FloatTyID; };
    bool is_Function() const { return tid_ == FunctionTyID; };
    bool is_Pointer() const { return tid_ == PointerTyID; };
    bool is_Array() const { return tid_ == ArrayTyID; };
    virtual bool isConst() const { return is_const; }
  virtual std::string print();
  
};

class IntegerType : public Type {
public:
    unsigned num_bits_;
    explicit IntegerType(unsigned num_bits) : Type(Type::IntegerTyID), num_bits_(num_bits) {}
    
};

class ArrayType : public Type {
public:
    Type* contained_;        
    unsigned num_elements_;  
    ArrayType(Type* contained, unsigned num_elements) : Type(Type::ArrayTyID), num_elements_(num_elements), contained_(contained) {}
    
};


class PointerType : public Type {
public:
    Type* contained_;  
    PointerType(Type* contained) : Type(Type::PointerTyID), contained_(contained) {}
    
};

class FunctionType : public Type {
public:

  Type *result_;
  std::vector<Type *> args_;
  FunctionType(Type *result, std::vector<Type *> params)
      : Type(Type::FunctionTyID) {
    result_ = result;
    for (Type *p : params) {
      args_.push_back(p);
    }
  }
  
};

#endif