
#include "./ARM/build.h"
#include "./IR/genIR.h"
#include "./AST/ast.h"

#include "./IR/type.h"
#include "./ARM/OPT/opt.h"
#include <fstream>
#include <iostream>
#include <ostream>
#include <unistd.h>

extern unique_ptr<COMPUnitAST> root;
extern int yyparse();
extern FILE *yyin;

bool optimize (int o2) {
  if(o2)
    return true;
  else
    return false;
}

bool id_ = false;


int main(int argc, char **argv) {
  // Assert the number of arguments
  assert(argc >= 2);

  // TODO: advanced argument parser
  char *filename = nullptr;
  int print_ir = false;
  int print_asm = false;

  std::string output = "-";

  int opt;
  bool isO2 = false;
  while ((opt = getopt(argc, argv, "Sco:O::")) != -1) {
    switch (opt) {
    case 'S':
      print_asm = true;
      print_ir = false;
      break;
    case 'c':
      print_ir = true;
      print_asm = false;
      break;
    case 'o':
      output = optarg;
      break;
    case 'O':
      isO2 = true;
      break;
    default:
      break;
    }
  }
  filename = argv[optind];

  yyin = fopen(filename, "r");
  if (yyin == nullptr) {
    std::cout << "yyin open" << filename << "failed" << std::endl;
    return -1;
  }

  // Frontend parser
  yyparse();

  // Generate IR from AST
  GenIR genIR;
  root->ACcept(genIR);
  std::unique_ptr<Module> m = genIR.getModule();

  // Run IR optimization
  if (optimize(isO2)) {
          DCodeSolve* o1 = new DCodeSolve(m.get());
          CInstr* o2 = new CInstr(m.get());
         DTree* o3 = new DTree(m.get());
          SJump* o4 = new SJump(m.get());
         LInvariant* o5 = new LInvariant(m.get());
          SJump* o6 = new SJump(m.get());
          o1->implement();
          o2->implement();o3->implement();o4->implement();
          o5->implement();o6->implement();
    }


  // Open output file
  std::ofstream fout;
  std::ostream *out;
  if (output == "-") {
    out = &std::cout;
  } else {
    fout.open(output);
    out = &fout;
  }

  
  const std::string IR = m->print();
  if (print_ir) {
    *out << IR << std::endl;
  }

  
  if (print_asm) {
    CreateStmt* lirbuilder;
    lirbuilder = new CreateStmt(m.get(), false, isO2);
    lirbuilder->clir();
    lirbuilder->carm();
    std::string ASM = lirbuilder->m_lir->print();
    *out << ASM << std::endl;
  }
  return 0;
}
